// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** @file */

#include "nvsgcommon.h"

#include "nvmath/Vecnf.h"

namespace nvmath
{
  //! Template for nxn-matrices
  template<unsigned char n> class Matnnf
  {
    public:
      //! Default constructor.
      /** For performance reasons no initialization is performed. */
      Matnnf<n>( void );

      //! Non-constant subscript operator.
      /** \returns reference to \a i th row of the matrix  */
      Vecnf<n> & operator[]( size_t i   //!<  index to row to use (i=0,..,n-1)
                           );

      //! Constant subscript operator.
      /** \returns constant reference to \a i th row of the matrix */
      const Vecnf<n> & operator[]( size_t i   //!<  index to row to use (i=0,..,n-1)
                                 ) const;

      //! Transpose operator.
      /** returns transposed matrix */
      Matnnf<n> operator~( void ) const;

      //! Negation operator.
      /** \returns negated matrix  */
      Matnnf<n> operator-( void ) const;

      //! Matrix multiplication with a scalar operator.
      /** \returns multiplied matrix */
      Matnnf<n> operator*( float f                //!<  scalar to multiply with
                         ) const;

      //! Matrix multiplication with a vector operator.
      /** \returns matrix time \a v */
      Vecnf<n> operator*( const Vecnf<n>& v       //!<  vector to multiply with
                        ) const;

      //! Matrix division by a scalar operator.
      /** \returns divided matrix */
      Matnnf<n> operator/( float f                //!<  scalar to divide through
                         ) const;

      //! Matrix addition operator.
      /** \returns sum of the two matrices  */
      Matnnf<n> operator+( const Matnnf<n> &v     //!<  matrix to add
                         ) const;

      //! Matrix subtraction operator.
      /** \returns difference of the two matrices  */
      Matnnf<n> operator-( const Matnnf<n> &v     //!<  matrix to add
                         ) const;

      //! Equality operator.
      /** \returns true if the matrix is equal to \a m, otherwise false  */
      bool operator==( const Matnnf<n> &m         //!<  matrix to compare with
                     ) const;

      //! Inequality operator.
      /** \returns true if the matrix is not equal to \a m, otherwise false  */
      bool operator!=( const Matnnf<n> &m         //!< matrix to compare with
                     ) const;

      //! Assignment operator.
      /** \returns reference to the assigned matrix */
      Matnnf<n> & operator=( const Matnnf<n> &m   //!<  matrix to copy
                           );

      //! Matrix multiplication with a scalar and assignment operator.
      /** \returns reference to the assigned matrix after multiplication */
      Matnnf<n> & operator*=( float f    //!<  scalar to multiply with
                            );

      //! Matrix multiplication with a matrix (i.e. concatenation) and assignment operator.
      /** \returns reference to the assigned matrix after multiplication (concatenation) */
      Matnnf<n> & operator*=( const Matnnf<n> &m    //!<  matrix to multiply
                           );

      //! Matrix division by a scalar and assignment operator.
      /** \returns reference to the assigned matrix after division */
      Matnnf<n> & operator/=( float f    //!<  scalar to divide through
                            );

      //! Matrix addition and assignment operator.
      /** \returns  reference to the assigned matrix after addition */
      Matnnf<n> & operator+=( const Matnnf<n> &m    //!<  matrix to add
                           );

      //! Matrix subtraction and assignment operator.
      /** \returns  reference to the assigned matrix after addition */
      Matnnf<n> & operator-=( const Matnnf<n> &m    //!<  matrix to subtract
                           );

      //! Get a constant pointer to the nxn floats of the matrix.
      /** \returns constant pointer to matrix elements */
      const float* getPtr( void ) const;

    private:
      Vecnf<n>  m_mat[n];
  };

  // - - - - -  - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
  // non-member functions
  // - - - - -  - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
  /*! \relates nvmath::Matnnf
   *  Test if a matrix is the identity
   *  \returns true if the matrix is the identity, otherwise false */
  template<unsigned char n> bool isIdentity( const Matnnf<n> &m   //!<  matrix to test
                                           )
  {
    bool identity = true;
    for ( size_t i=0 ; identity && i<n ; i++ )
    {
      for ( size_t j=0 ; identity && j<i ; j++ )
      {
        identity = ( fabsf( m[i][j] ) <= FLT_EPSILON );
      }
      if ( identity )
      {
        identity = ( fabsf( m[i][i] - 1.0f ) <= FLT_EPSILON );
      }
      for ( size_t j=i+1 ; identity && j<n ; j++ )
      {
        identity = ( fabsf( m[i][j] ) <= FLT_EPSILON );
      }
    }
    return( identity );
  }

  /*! \relates nvmath::Matnnf
   *  Test if a matrix is normalized.
   *  A matrix is normalized when both all rows and all columns are normalized.
   * \returns true if the matrix is normalized, otherwise false */
  template<unsigned char n> bool isNormalized( const Matnnf<n> &mat   //!<  matrix to test
                                             )
  {
    bool normalized = true;
    for ( size_t i=0 ; normalized && i<n ; i++ )
    {
      normalized = isNormalized( mat[i] );
    }
    for ( size_t i=0 ; normalized && i<n ; i++ )
    {
      Vecnf<n> v;
      for ( size_t j=0 ; j<n ; j++ )
      {
        v[j] = mat[j][i];
      }
      normalized = isNormalized( v );
    }
    return( normalized );
  }

  /*! \relates nvmath::Matnnf
   *  Test if a matrix is orthogonal.
   *  A matrix is orthogonal when every row is orthogonal to each other row and every column is orthogonal to each
   *  other column.
   *  \returns true if the matrix is orthogonal, otherwise false */
  template<unsigned char n> bool isOrthogonal( const Matnnf<n> &mat   //!<  matrix to test
                                             )
  {
    bool orthogonal = true;
    for ( size_t i=0 ; orthogonal && i<n-1 ; i++ )
    {
      for ( size_t j=i+1 ; orthogonal && j<n ; j++ )
      {
        orthogonal = areOrthogonal( mat[i], mat[j] );
      }
    }
    if ( orthogonal )
    {
      Matnnf<n> tmp = ~mat;
      for ( size_t i=0 ; orthogonal && i<n-1 ; i++ )
      {
        for ( size_t j=i+1 ; orthogonal && j<n ; j++ )
        {
          orthogonal = areOrthogonal( tmp[i], tmp[j] );
        }
      }
    }
    return( orthogonal );
  }

  /*! \relates nvmath::Matnnf
   *  Matrix multiplication with a matrix (i.e. concatenation) operator.
   *  \returns multiplied (concatenated) matrix  */
  template<unsigned char n> Matnnf<n> operator*( const Matnnf<n> &m0, const Matnnf<n> &m1 )
  {
    Matnnf<n> ret;
    for ( size_t i=0 ; i<n ; i++ )
    {
      for ( size_t j=0 ; j<n ; j++ )
      {
        float f = 0.0f;
        for ( size_t k=0 ; k<n ; k++ )
        {
          f += m0[i][k] * m1[k][j];
        }
        ret[i][j] = f;
      }
    }
    return( ret );
  }

  // - - - - -  - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
  // "inlines"
  // - - - - -  - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
  template<unsigned char n> Matnnf<n>::Matnnf( void )
  {
  }

  template<unsigned char n> Vecnf<n> & Matnnf<n>::operator[]( size_t i ) 
  { 
    __ASSERT( 0 <= i && i < n );
    return( m_mat[i] );
  }

  template<unsigned char n> const Vecnf<n> & Matnnf<n>::operator[]( size_t i ) const
  { 
    __ASSERT( 0 <= i && i < n );
    return( m_mat[i] );
  }

  template<unsigned char n> Matnnf<n> Matnnf<n>::operator~( void ) const
  {
    Matnnf<n> ret;
    for ( size_t i=0 ; i<n ; i++ )
    {
      for ( size_t j=0 ; j<n ; j++ )
      {
        ret[i][j] = m_mat[j][i];
      }
    }
    return( ret );
  }

  template<unsigned char n> Matnnf<n> Matnnf<n>::operator-( void ) const
  {
    Matnnf<n> ret;
    for ( size_t i=0 ; i<n ; i++ )
    {
      ret[i] = - m_mat[i];
    }
  }

  template<unsigned char n> Matnnf<n> Matnnf<n>::operator*( float f ) const
  {
    Matnnf<n> ret( *this );
    return( ret *= f );
  }

  template<unsigned char n> Vecnf<n> Matnnf<n>::operator*( const Vecnf<n> &v ) const
  {
    Vecnf<n>  ret;
    for ( size_t i=0 ; i<n ; i++ )
    {
      ret[i] = m_mat[i] * v;
    }
    return( ret );
  }

  template<unsigned char n> Matnnf<n> Matnnf<n>::operator/( float f ) const
  {
    __ASSERT( fabs( f ) > FLT_EPSILON );
    Matnnf<n> ret( *this );
    return( ret /= f );
  }

  template<unsigned char n> Matnnf<n> Matnnf<n>::operator+( const Matnnf<n> &m ) const
  {
    Matnnf<n> ret( *this );
    return( ret += m );
  }

  template<unsigned char n> Matnnf<n> Matnnf<n>::operator-( const Matnnf<n> &m ) const
  {
    Matnnf<n> ret( *this );
    return( ret -= m );
  }

  template<unsigned char n> bool Matnnf<n>::operator==( const Matnnf<n> &m ) const
  {
    bool equal = true;
    for ( unsigned char i=0 ; equal && i<n ; i++ )
    {
      equal = ( m_mat[i] == m[i] );
    }
    return( equal );
  }

  template<unsigned char n> bool Matnnf<n>::operator!=( const Matnnf<n> &m ) const
  {
    return( ! ( *this == m ) );
  }

  template<unsigned char n> Matnnf<n> & Matnnf<n>::operator=( const Matnnf<n> &m )
  {
    for ( size_t i=0 ; i<n ; i++ )
    {
      m_mat[i] = m[i];
    }
    return( *this );
  }

  template<unsigned char n> Matnnf<n> & Matnnf<n>::operator*=( float f )
  {
    for ( size_t i=0 ; i<n ; i++ )
    {
      m_mat[i] *= f;
    }
    return( *this );
  }

  template<unsigned char n> Matnnf<n> & Matnnf<n>::operator*=( const Matnnf<n> &m )
  {
    *this = *this * m;
    return( *this );
  }

  template<unsigned char n> Matnnf<n> & Matnnf<n>::operator/=( float f )
  {
    __ASSERT( fabs( f ) > FLT_EPSILON );
    for ( size_t i=0 ; i<n ; i++ )
    {
      m_mat[i] /= f;
    }
    return( *this );
  }

  template<unsigned char n> Matnnf<n> & Matnnf<n>::operator+=( const Matnnf<n> &m )
  {
    for ( size_t i=0 ; i<n ; i++ )
    {
      m_mat[i] += m[i];
    }
    return( *this );
  }

  template<unsigned char n> Matnnf<n> & Matnnf<n>::operator-=( const Matnnf<n> &m )
  {
    for ( size_t i=0 ; i<n ; i++ )
    {
      m_mat[i] -= m[i];
    }
    return( *this );
  }

  template<unsigned char n> const float * Matnnf<n>::getPtr( void ) const
  {
    return( m_mat[0].getPtr() );
  }

} // end namspace nvmath
