// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** @file */

#include "nvsgcommon.h"

#include "nvmath/nvmath.h"
#include "nvmath/Vec4f.h"
#include "nvmath/Vec3f.h"
#include "nvmath/Mat33f.h"

namespace nvmath
{
  //! Quaternion class.
  /** Quaternions add a fourth element to the [x,y,z] values that define a three-component vector. Quaternions are an
    * alternative to the matrix methods that are typically used for 3-D rotations. A quaternion represents an axis in 3-D
    * space and a rotation around that axis. For example, a quaternion might represent a (1,1,2) axis and a rotation of 1
    * radian.
    * \note Only unit quaternions represent a rotation. We always expect unit quaternions.
    */
  class Quatf
  {
    public:
      //! Default constructor.
      /** For performance reasons no initialization is performed. */
      NVSG_API Quatf( void );

      //! Constructor using a Vec4f.
      /** A quaternion with the same values as \a v is constructed. */
      NVSG_API Quatf( const Vec4f& v   //!<  vector to construct from
                    );

      //! Constructor using four floats.
      /** A quaternion with the given values is constructed.
        * \note f0, f1, and f2 are \b not the x,y,z-component of the rotation axis, and f3 is \b not the rotation angle.
        * If you have such values handy, use the constructor that takes the axis as a Vec3f and an angle. */
      NVSG_API Quatf( float f0    //!<  x component of the new quaternion
                    , float f1    //!<  y component of the new quaternion
                    , float f2    //!<  z component of the new quaternion
                    , float f3    //!<  w component of the new quaternion
                    );

      //! Constructor using an axis and an angle
      /** A quaternion is constructed through a normalized axis and an angle. */
      NVSG_API Quatf( const Vec3f& axis    //!<  axis to use for quaternion
                    , float angle   //!<  rotation angle in radians
                    );

      //! Constructor by two vectors.
      /** A quaternion is constructed that is the rotation that maps \a vec0 to \a vec1. */
      NVSG_API Quatf( const Vec3f& vec0    //!<  start vector
                    , const Vec3f& vec1    //!<  end vector
                    );

      //! Non-constant subscript operator.
      /** \returns reference to \a i th element of the quaternion  */
      NVSG_API float& operator[]( size_t i   //!<  index to quaternion to use (i=0,1,2,3)
                                );

      //! Constant subscript operator.
      /** \returns constant reference to \a i th element of the quaternion  */
      NVSG_API const float& operator[]( size_t i   //!<  index to quaternion to use (i=0,1,2,3)
                                      ) const;

      //! Assignment operator.
      /** \returns reference to the assigned quaternion  */
      NVSG_API Quatf& operator=( const Quatf& q   //!<  copy \a q to the quaternion
                               );

      //! Equality operator.
      /** \returns \c true if the quaternion is equal to \a q, otherwise \c false  */
      NVSG_API bool operator==( const Quatf& q    //!<  quaternion to compare
                              ) const;

      //! Inequality operator.
      /** \returns true if the quaternion is not equal to \a q, otherwise false  */
      NVSG_API bool operator!=( const Quatf& q    //!<  quaternion to compare
                              ) const;

      //! Quaternion addition and assignment operator.
      /** \returns reference to the assigned quaternion after addition */
      NVSG_API Quatf& operator+=( const Quatf& q    //!<  quaternion to add
                                );

      //! Quaternion subtraction and assignment operator.
      /** \returns reference to the assigned quaternion after subtraction  */
      NVSG_API Quatf& operator-=( const Quatf& q    //!<  quaternion to subtract
                                );

      //! Quaternion subtraction operator.
      /** \returns difference of the two quaternions */
      NVSG_API Quatf operator-( const Quatf& q    //!<  quaternion to subtract
                              ) const;

      //! Negation operator.
      /** \returns negated quaternion  */
      NVSG_API Quatf operator-( void ) const;

      //! Quaternion multiplication with a scalar and assignment operator.
      /** \returns reference to the assigned quaternion after multiplication */
      NVSG_API Quatf& operator*=( float f   //!<  scalar to use in the multiplication
                                );

      //! Quaternion multiplication with a scalar operator.
      /** \returns multiplied quaternion */
      NVSG_API Quatf operator*( float f   //!<  scalar to multiply with
                              ) const;

      //! Quaternion multiplication with a quaternion (i.e. concatenation) and assignment operator.
      /** \returns reference to the assigned quaternion after multiplication (concatenation) */
      NVSG_API Quatf& operator*=( const Quatf& q    //!<  quaternion to use in the multiplication
                                );

      //! Quaternion multiplication with a quaternion (i.e. concatenation) operator.
      /** \returns multiplied (concatenated) quaternion  */
      NVSG_API Quatf operator*( const Quatf& q    //!<  quaternion to use in the multiplication
                              ) const;

      //! Quaternion division by a scalar and assignment operator.
      /** \returns reference to the assigned quaternion after division */
      NVSG_API Quatf& operator/=( float f   //!<  scalar to use in the division
                                );

      //! Quaternion division by a scalar operator.
      /** \returns divided quaternion  */
      NVSG_API Quatf operator/( float  f    //!<  scalar to use in the division
                              ) const;

      //! Quaternion division by a quaternion and assignment operator.
      /** \returns reference to the assigned quaternion after division */
      NVSG_API Quatf& operator/=( const Quatf& q //!<  quaternion to use in the division
                                );

      //! Quaternion division by a quaternion operator.
      /** \returns divided quaternion */
      NVSG_API Quatf operator/( const Quatf& q    //!<  quaternion to use in the division
                              ) const;

      //! Quaternion conjugation operator.
      /** The conjugation of a quaternion is a rotation of the same size around the negated axis.
        * \returns conjugated quaternion */
      NVSG_API Quatf operator~( void ) const;

      //! Invert the quaternion.
      /** Inversion means take the conjugated quaternion and divide it by it's magnitude. */
      NVSG_API void invert( void );

      //! Normalize the quaternion.
      NVSG_API Quatf& normalize( void );

    private:
      NVSG_API void _byAxisAndAngle( const Vec3f& axis, float angle );

    private:
      float m_quat[4];
  };

  // - - - - -  - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
  // non-member functions
  // - - - - -  - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
  /*! \relates nvmath::Quatf
   *  Decompose a quaternion into axis and angle. */
  inline  void  decompose( const Quatf &q   //!<  Quaternion to decompose
                         , Vec3f &axis      //!<  Resulting axis
                         , float &angle     //!<  Resulting angle (in radians)
                         )
  {
    angle = 2.0f * acosf( q[3] );
    if ( angle < FLT_EPSILON )
    {
      //  no angle to rotate about => take just any one
      axis[0] = 0.0f;
      axis[1] = 0.0f;
      axis[2] = 1.0f;
    }
    else
    {
      float dummy = 1.0f / sinf( 0.5f * angle );
      axis[0] = q[0] * dummy;
      axis[1] = q[1] * dummy;
      axis[2] = q[2] * dummy;
      axis.normalize();
    }
  }

  /*! \relates nvmath::Quatf
   *  Determine the distance between two quaternions.
   *  \returns The distance between \a q0 and \a q1 */
  inline float distance( const Quatf &q0    //!<  first quaternion
                       , const Quatf &q1    //!<  second quaternion
                       )
  {
    return( sqrtf( square( q0[0] - q1[0] ) + square( q0[1] - q1[1] ) + square( q0[2] - q1[2] ) + square( q0[3] - q1[3] ) ) );
  }

  /*! \relates nvmath::Quatf
   *  Calculate the magnitude of a quaternion.
   *  \returns magnitude of the quaternion */
  inline float magnitude( const Quatf &q   //!< Quaternion to determine magnitude of
                        )
  {
    return( sqrtf( square( q[0] ) + square( q[1] ) + square( q[2] ) + square( q[3] ) ) );
  }

  /*! \relates nvmath::Quatf
   *  Test if a quaternion is normalized.
   *  A quaternion is normalized if it's magnitude is one.
   *  \returns true if the quaternion is normalized, otherwise false */
  inline bool isNormalized( const Quatf &q  //!< Quaternion to test
                          )
  {
    return( fabs( magnitude( q ) - 1.0f ) <= FLT_EPSILON );
  }

  /*! \relates nvmath::Quatf
   *  Test if a quaternion is null.
   *  \returns true if the quaternion is null, otherwise false  */
  inline bool isNull( const Quatf &q      //!< Quaternion to test
                    )
  {
    return( magnitude( q ) <= FLT_EPSILON );
  }

  /*! \relates nvmath::Quatf
   *  Multiply a scalar \a f with a quaternion \a q.
   *  \returns product of \a f with \a q */
  inline  Quatf operator*( float f          //!<  scalar to multiply with
                         , const Quatf &q   //!<  quaternion to multiply
                         )
  {
    return( q * f );
  }

  /*! \relates nvmath::Quatf
   *  Multiply a quaternion \a q with an vector \a v. 
   *  Rotation of vector \a v by quaternion \a q.
   *  \returns product of \a q with \a v */
  inline Vec3f operator*( const Quatf &q, const Vec3f &v )
  {
    return Mat33f(q) * v;
  }

  /*! \relates nvmath::Quatf
   *  Spherical linear interpolation between two quaternions \a q0 and \a q1
   */
  template<>
  NVSG_API Quatf lerp<Quatf>( float alpha          //!<  interpolation parameter
                            , const Quatf &q0      //!<  starting value
                            , const Quatf &q1      //!<  ending value
                            );

  // - - - - -  - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
  // inlines
  // - - - - -  - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
  inline Quatf::Quatf( void )
  {
  }

  inline Quatf::Quatf( float f0, float f1, float f2, float f3 )
  {
    m_quat[0] = f0;
    m_quat[1] = f1;
    m_quat[2] = f2;
    m_quat[3] = f3;
  }

  inline Quatf::Quatf( const Vec4f& v )
  {
    m_quat[0] = v[0];
    m_quat[1] = v[1];
    m_quat[2] = v[2];
    m_quat[3] = v[3];
    __ASSERT( isNormalized( *this ) );
  }

  inline void Quatf::_byAxisAndAngle( const Vec3f& axis, float angle )
  {
    __ASSERT( isNormalized( axis ) );
    float dummy = sinf( 0.5f * angle );
    m_quat[0] = axis[0] * dummy;
    m_quat[1] = axis[1] * dummy;
    m_quat[2] = axis[2] * dummy;
    m_quat[3] = cosf( 0.5f * angle );
    __ASSERT( isNormalized( *this ) );
  }

  inline Quatf::Quatf( const Vec3f& axis, float angle )
  {
    _byAxisAndAngle( axis, angle );
  }

  inline Quatf::Quatf( const Vec3f& vec0, const Vec3f& vec1 )
  {
    __ASSERT( isNormalized( vec0 ) && isNormalized( vec1 ) );
    if ( areCollinear( vec0, vec1 ) )
    {
      (*this) = Quatf( 0.0f, 0.0f, 0.0f, 1.0f );
    }
    else
    {
      _byAxisAndAngle( nvmath::normalize( vec0 ^ vec1 ), acosf( vec0 * vec1 ) );
    }
  }

  inline float& Quatf::operator[]( size_t i ) 
  { 
    __ASSERT( 0 <= i && i <= 3 );
    return( m_quat[i] );
  }

  inline const float& Quatf::operator[]( size_t i ) const 
  { 
    __ASSERT( 0 <= i && i <= 3 );
    return( m_quat[i] );
  }

  inline Quatf& Quatf::operator=( const Quatf& q )
  {
    m_quat[0] = q[0];
    m_quat[1] = q[1];
    m_quat[2] = q[2];
    m_quat[3] = q[3];
    return( *this );
  }

  inline bool Quatf::operator==( const Quatf& q ) const 
  { 
    return(     m_quat[0] == q[0]
            &&  m_quat[1] == q[1]
            &&  m_quat[2] == q[2]
            &&  m_quat[3] == q[3]
          ); 
  }

  inline bool Quatf::operator!=( const Quatf& q ) const 
  { 
    return(     m_quat[0] != q[0]
            ||  m_quat[1] != q[1]
            ||  m_quat[2] != q[2]
            ||  m_quat[3] != q[3]
          );
  }

  inline Quatf Quatf::operator-( void ) const
  {
    return( Quatf( m_quat[0], m_quat[1], m_quat[2], -m_quat[3] ) );
  }

  inline Quatf& Quatf::operator*=( const Quatf & q )
  {
    __ASSERT( isNormalized( *this ) && isNormalized( q ) );
    *this = Quatf( m_quat[1]*q[2] - m_quat[2]*q[1] + m_quat[3]*q[0] + m_quat[0]*q[3],
                   m_quat[2]*q[0] - m_quat[0]*q[2] + m_quat[3]*q[1] + m_quat[1]*q[3],
                   m_quat[0]*q[1] - m_quat[1]*q[0] + m_quat[3]*q[2] + m_quat[2]*q[3],
                   m_quat[3]*q[3] - m_quat[0]*q[0] - m_quat[1]*q[1] - m_quat[2]*q[2] );
    normalize();
    return( *this );
  }

  inline Quatf Quatf::operator*( const Quatf & q ) const
  {
    __ASSERT( isNormalized( *this ) && isNormalized( q ) );
    Quatf r(*this);
    r *= q;
    __ASSERT( isNormalized( r ) );
    return( r );
  }

  inline Quatf& Quatf::operator/=( const Quatf & q )
  {
    Quatf tmp = *this / q;
    *this = tmp;
    return( *this );
  }

  inline Quatf Quatf::operator/( const Quatf & q ) const
  {
    __ASSERT( fabsf( magnitude( q ) - 1.0f ) < FLT_EPSILON );
    return( *this * ~q /*/ magnitude( q )*/ );
  }

  inline Quatf Quatf::operator~( void ) const
  {
    return( Quatf( -m_quat[0], -m_quat[1], -m_quat[2], m_quat[3] ) );
  }

  inline void Quatf::invert( void )
  {
    __ASSERT( fabsf( magnitude( *this ) - 1.0f ) < FLT_EPSILON );
    *this = ~(*this)/* / getMagnitude()*/;
  }

  /** Quaternions always obey:  a^2 + b^2 + c^2 + d^2 = 1.0
  *  If they don't add up to 1.0, dividing by their magnitude will
  *  renormalize them.
  */
  inline Quatf& Quatf::normalize( void )
  {
    float mag = magnitude( *this );
    __ASSERT( mag > FLT_EPSILON );
    for (int i = 0; i < 4; i++)
    {
      m_quat[i] /= mag;
    }
    return( *this );
  }
}
