// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** @file */

#include "nvsgcommon.h"

#include "nvmath/Vec3f.h"

namespace nvmath
{
  //! Sphere class
  class Sphere3f
  {
    public:
      //! Default constructor.
      /** For performance reasons no initialization is performed. */
      NVSG_API Sphere3f( void );

      //! Constructor by center and radius
      NVSG_API Sphere3f( const Vec3f &center  //!<  center
                       , float radius         //!<  radius
                       );

      //! Assignment operator
      /** \returns reference to the assigned plane */
      NVSG_API Sphere3f & operator=( const Sphere3f &sphere //!<  sphere to assign
                                   );

      //! Equality operator.
      /** \returns true if the sphere is equal to \a pl, otherwise false  */
      NVSG_API bool operator==( const Sphere3f& sphere      //!<  sphere to compare with
                              ) const;

      //! Inequality operator.
      /** \returns true if the sphere is not equal to \a pl, otherwise false  */
      NVSG_API bool operator!=( const Sphere3f& sphere      //!<  sphere to compare with
                              ) const;

      //! Calling operator
      /** Calculates the distance of a point to the sphere. When the point is inside the sphere, the distance is
        * negative; when the point is outside, the distance is positive.
        * \returns the signed distance of \a p to the sphere. */
      NVSG_API float operator()( const Vec3f &p             //!<  point to determine distance from sphere
                               ) const;

      //! Get the center of the sphere.
      /** \returns the center of the sphere. */
      NVSG_API const Vec3f & getCenter( void ) const;

      //! Set the center of the sphere.
      NVSG_API void setCenter( const Vec3f &center    //!<  point to set as center
                             );

      //! Get the radius of the sphere.
      /** \returns the radius of the sphere. */
      NVSG_API float getRadius( void ) const;

      //! Set the radius of the sphere.
      NVSG_API void setRadius( float radius           //!<  radius to set
                             );

    private:
      Vec3f m_center;
      float m_radius;
  };

  // - - - - -  - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
  // non-member functions
  // - - - - -  - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
  /*! \relates nvmath::Sphere3f
   *  Determine the bounding sphere a number of points
   *  \note The determined sphere is not neccessarily the smallest.
   *  \returns A reasonable small sphere around the given points */
  Sphere3f boundingSphere( const Vec3f *points      //!<  points to to include
                         , size_t numberOfPoints    //!<  number of points
                         );

  /*! \relates nvmath::Sphere3f
   *  Determine the bounding sphere around a sphere \a s and a point \a p.
   *  \returns The bounding sphere around \a s and \a p. */
  Sphere3f boundingSphere( const Sphere3f &s        //!<  sphere to include
                         , const Vec3f &p           //!<  point to include
                         );

  /*! \relates nvmath::Sphere3f
   *  Determine the bounding sphere around two given spheres.
   *  \returns the smallest sphere that includes both given spheres \a s0 and \a s1. */
  Sphere3f boundingSphere( const Sphere3f &s0    //!<  first sphere to include
                         , const Sphere3f &s1    //!<  second sphere to include
                         );

  // - - - - -  - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
  // inlines
  // - - - - -  - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
  inline Sphere3f::Sphere3f( void )
  {
  }

  inline Sphere3f::Sphere3f( const Vec3f &center, float radius )
    : m_center(center)
    , m_radius(radius)
  {
  }

  inline Sphere3f & Sphere3f::operator=( const Sphere3f &sphere )
  {
    m_center = sphere.m_center;
    m_radius = sphere.m_radius;
    return( *this );
  }

  inline bool Sphere3f::operator==( const Sphere3f &sphere ) const
  {
    return( ( m_center == sphere.m_center ) && ( fabsf( m_radius - sphere.m_radius ) < FLT_EPSILON ) );
  }

  inline bool Sphere3f::operator!=( const Sphere3f &sphere ) const
  {
    return( ! (this->operator==( sphere )) );
  }

  inline float Sphere3f::operator()( const Vec3f &p ) const
  {
    return( distance( m_center, p ) - m_radius );
  }

  inline const Vec3f & Sphere3f::getCenter( void ) const
  {
    return( m_center );
  }

  inline void Sphere3f::setCenter( const Vec3f &center )
  {
    m_center = center;
  }

  inline float Sphere3f::getRadius( void ) const
  {
    return( m_radius );
  }

  inline void Sphere3f::setRadius( float radius )
  {
    m_radius = radius;
  }

  //! global unit sphere.
  extern NVSG_API const Sphere3f cUnitSphere;

} // end namspace nvmath
