// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** @file */

#include "nvsgcommon.h"

#include  "nvsg/Animation.h"
#include  "nvsg/Transform.h"

namespace nvsg
{
  //! Class representing an animated transformation.
  /** An AnimatedTransform is a Transform that also holds an Animation. The transformation is determined from
    * the Animation with every frame.
    */
  class AnimatedTransform : public Transform
  {
    public:
      //! Create an AnimatedTransform.
      /** There is an optional argument \a pAnimation to use a given Animation<Trafo>.
        * \returns a constant pointer to an AnimatedQuads. */
      NVSG_API static const AnimatedTransform * create( const Animation<nvmath::Trafo> *pAnimation = NULL  //!< optional animation to use
                                                      );

      //! Create an AnimatedTransform by copying from a Transform.
      /** There is an optional argument \a pAnimation to use a given Animation<Trafo>.
        * \returns a constant pointer to an AnimatedTransform. */
      NVSG_API static const AnimatedTransform * createFromBase( const Transform &rhs                              //!< Transform to copy from
                                                              , const Animation<nvmath::Trafo> *pAnimation = NULL //!< optional animation to use
                                                              );

      //! Create a new AnimatedTransforms by copying from this.
      /** \returns a constant pointer to an identical AnimatedTransform. */
      NVSG_API virtual AnimatedTransform * clone( void ) const;

      //! Get a constant pointer to the animation.
      /** \returns  constant pointer to the animation */
      NVSG_API const Animation<nvmath::Trafo> * getAnimation( void ) const;

      //! Set the Animation.
      /** A previous Animation is dereferenced, the new one is referenced, and the animation frame is reset to zero.  */
      NVSG_API void setAnimation( const Animation<nvmath::Trafo> *pAnimation    //!<  new animation to use
                                );

      //! Set the animation frame.
      /** The transformation for animation step \a frame is calculated and used as the current. */
      NVSG_API void setAnimationFrame( size_t frame   //!<  current animation step
                                     );

      //--  Functions reimplemented from Transform  --
      NVSG_API virtual void setCenter( const nvmath::Vec3f &center );
      NVSG_API virtual void setIdentity( void );
      NVSG_API virtual void setOrientation( const nvmath::Quatf &orientation );
      NVSG_API virtual void setScaling( const nvmath::Vec3f &scaling );
      NVSG_API virtual void setTranslation( const nvmath::Vec3f &translation );

      //--  Functions reimplemented from Node --
      NVSG_API virtual size_t getNumberOfFrames( void ) const;

      //--  Functions reimplemented from Object  --
      NVSG_API virtual DataID getDataID( void ) const;
      NVSG_API virtual bool isDataShared( void ) const;

    protected:
      //! Constructor with an Animation.
      NVSG_API AnimatedTransform( const Animation<nvmath::Trafo> *pAnimation );

      //! Constructur with a Transform and an Animation
      NVSG_API AnimatedTransform( const Transform &rhs, const Animation<nvmath::Trafo> *pAnimation );

      //! Copy Constructor
      NVSG_API AnimatedTransform( const AnimatedTransform& rhs );

      //! Protected destructor to prevent explicit creation on stack.
      NVSG_API virtual ~AnimatedTransform(void);

    private:
      AnimatedTransform();    //  private default constructor to prevent use

    private:
      const Animation<nvmath::Trafo>  * m_animation;
      size_t                            m_currentFrame;
  };

  // - - - - - - - - - - - - - - - - - - - - - - - - - 
  // inlines
  // - - - - - - - - - - - - - - - - - - - - - - - - - 

  inline const Animation<nvmath::Trafo> * AnimatedTransform::getAnimation( void ) const
  {
    __TRACE();
    return( m_animation );
  }

} //  namespace nvsg
