// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** @file */

#include "nvsgcommon.h"

#include "CgFX/ICgFXEffect.h"
#include "nvsg/StateAttribute.h"
#include <string>
#include <vector>

namespace nvsg
{
  class  CgFx;
  class  CgFxEntity;
  class  CgFxParameter;
  class  CgFxAnnotation;
  class  CgFxPass;
  class  CgFxTechnique;

  typedef const CgFxEntity     * CgFxEntityHandle;      //!<  Handle to a \c CgFxEntity
  typedef const CgFxAnnotation * CgFxAnnotationHandle;  //!<  Handle to a \c CgFxAnnotation
  typedef const CgFxParameter  * CgFxParameterHandle;   //!<  Handle to a \c CgFxParameter
  typedef const CgFxTechnique  * CgFxTechniqueHandle;   //!<  Handle to a \c CgFxTechnique
  typedef const CgFxPass       * CgFxPassHandle;        //!<  Handle to a \c CgFxPass

  typedef enum
  { // known semantics
    CGFX_SEMANTIC_UNKNOWN = 0,
    CGFX_SEMANTIC_WORLD,                       //!< Object to world transform
	  CGFX_SEMANTIC_FIRST_TRANSFORM = CGFX_SEMANTIC_WORLD,
	  CGFX_SEMANTIC_VIEW,                        //!< World to eyespace transform
	  CGFX_SEMANTIC_PROJECTION,                  //!< Projection matrix
	  CGFX_SEMANTIC_WORLDT,                      //!< Transposed world transform
	  CGFX_SEMANTIC_VIEWT,                       //!< Transposed view transform
	  CGFX_SEMANTIC_PROJECTIONT,                 //!< Transposed projection matrix
	  CGFX_SEMANTIC_WORLDI,                      //!< World to object transform
	  CGFX_SEMANTIC_VIEWI,                       //!< Eyespace to world transform
	  CGFX_SEMANTIC_PROJECTIONI,                 //!< Inverse projection matrix
	  CGFX_SEMANTIC_WORLDIT,                     //!< Inverse transposed world transform
	  CGFX_SEMANTIC_VIEWIT,                      //!< Inverse transposed view transform
	  CGFX_SEMANTIC_PROJECTIONIT,                //!< Inverse transposed projection matrix
	  CGFX_SEMANTIC_WORLDVIEW,                   //!< Concatenated modelview matrix
	  CGFX_SEMANTIC_WORLDVIEWPROJECTION,         //!< Concatenated modelview projection matrix
	  CGFX_SEMANTIC_WORLDVIEWT,                  //!< Transposed modelview matrix
	  CGFX_SEMANTIC_WORLDVIEWPROJECTIONT,        //!< Transposed modelview projection matrix
	  CGFX_SEMANTIC_WORLDVIEWI,                  //!< Inverse modelview matrix
	  CGFX_SEMANTIC_WORLDVIEWPROJECTIONI,        //!< Inverse modelview projection matrix
	  CGFX_SEMANTIC_WORLDVIEWIT,                 //!< Inverse transposed modelview matrix  
	  CGFX_SEMANTIC_WORLDVIEWPROJECTIONIT,       //!< Inverse transposed modelview projection matrix
    CGFX_SEMANTIC_LAST_TRANSFORM = CGFX_SEMANTIC_WORLDVIEWPROJECTIONIT 
  } eCgFxSemantic;

  //! A pair of a \c eCgFxSemantic and a \c CgFxParameterHandle.
  typedef std::pair<eCgFxSemantic, CgFxParameterHandle> CgFxTransform;

#if !defined(DOXYGEN_IGNORE) // The following are implementation details and need no public documentation.
                             // Exclude them from doxygen build

  // holds data common to all 'CgFx entities' like CgFxParameter, CgFxAnnotation, CgFxTechnique, and CgFxPass
  // note: this cannot be instantiated by general clients
  class CgFxEntity : public nvutil::RCObject
  {
    friend class CgFx; // private to CgFx implementation!
  protected:
    CgFxEntity();
    virtual ~CgFxEntity();
    CGFXHANDLE hdl;                     // opaque handle to talk to the corresponding effect
    std::vector<CgFxAnnotation*> annos; // collection of optional annotations 
  };

  // represents a CgFx parameter in general and acts as a helper used with the CgFx implementation
  // note: this cannot be instantiated by general clients
  class CgFxParameter : public CgFxEntity
  {
    friend class CgFx; // private to CgFx implementation ...
    
    // ... and the following queries
    friend const char *      getParameterName          (CgFxParameterHandle hdl);
    friend const char *      getParameterSemantic      (CgFxParameterHandle hdl);
    friend CgFXPARAMETERTYPE getParameterType          (CgFxParameterHandle hdl);
    friend bool              isParameterScalarType     (CgFxParameterHandle hdl);
    friend bool              isParameterVector2Type    (CgFxParameterHandle hdl);
    friend bool              isParameterVector3Type    (CgFxParameterHandle hdl);
    friend bool              isParameterVector4Type    (CgFxParameterHandle hdl);
    friend bool              isParameterMatrix33Type   (CgFxParameterHandle hdl);
    friend bool              isParameterMatrix34Type   (CgFxParameterHandle hdl);
    friend bool              isParameterMatrix43Type   (CgFxParameterHandle hdl);
    friend bool              isParameterMatrix44Type   (CgFxParameterHandle hdl);
    friend bool              isParameterTextureType    (CgFxParameterHandle hdl);
    friend bool              isParameterSamplerType    (CgFxParameterHandle hdl);

  protected:
    CgFxParameter() {}
    virtual ~CgFxParameter() {}
    CgFXPARAMETER_DESC desc;        // detailed description for this parameter
  };

  // represents an annotation for a general entity
  // note: as defined by the CgFX spec, an annotation itself is a parameter
  // note: this cannot be instantiated by general clients
  class CgFxAnnotation : public CgFxParameter
  {
    friend class CgFx; // private to CgFx implementation
  protected:
    CgFxAnnotation() {}
    ~CgFxAnnotation() {}
  };

  // represents a CgFx pass description and acts as a helper used with the CgFx implementation
  // note: this cannot be instantiated by general clients
  class CgFxPass : public CgFxEntity
  {
    friend class CgFx; // private to CgFx implementation ... 
    friend const char * getPassName (CgFxPassHandle hdl); // ... and this query
  protected:
    CgFxPass() {}
    ~CgFxPass() {}
    CgFXPASS_DESC desc;             //!< detailed pass description
  };

  // represents a CgFx technique description and acts as a helper used with the CgFx implementation
  // note: this cannot be instantiated by general clients
  class CgFxTechnique : public CgFxEntity
  {
    friend class CgFx; // private to CgFx implementation ...
    // ... and the following queries
    friend const char * getTechniqueName (CgFxTechniqueHandle hdl);
    friend bool         isTechniqueValid (CgFxTechniqueHandle hdl);
  protected:
    CgFxTechnique();                // initialization
    ~CgFxTechnique();               // termination
    CgFXTECHNIQUE_DESC desc;        // detailed technique description
    std::vector<CgFxPass*> passes;  // required render passes for this technique
    bool valid;                     // assigns whether this is a valid technique for the current environment
  };

#endif // DOXYGEN_IGNORE

  //! Effect attribute
  /** A special state attribute describing a CgFX effect.
    */
  class CgFx : public StateAttribute
  {
    public: // public interface

      //! Create a CgFx.
      /** \returns a constant pointer to a CgFx. */
      NVSG_API static const CgFx * create( void );

      //! Create a CgFx by copying from a StateAttribute.
      /** \returns a constant pointer to a CgFx. */
      NVSG_API static const CgFx * createFromBase( const StateAttribute &rhs    //!< StateAttribute to copy from
                                                 );

      //! Get a clone of this CgFx attribute.
      /** \returns A constant pointer to the new created clone. */
      NVSG_API virtual const CgFx * clone( void ) const;

      //! Create effect from file.
      /** This function creates an effect from an effect file given by \a file. 
        * \returns \c true if creation was successfull, \c false otherwise. */
      NVSG_API bool createEffectFromFile( const std::string& file    //!< Name of the effect file.
                                        );

      //! Detect whether the embedded effect is valid
      /** \returns \c true if the embedded effect is valid, \c false otherwise. */
      NVSG_API bool isEffectValid( void ) const;

      //! Get effect file.
      /** \returns The filename of the effect file that was taken to create the effect. */
      NVSG_API const std::string& getEffectFile( void ) const;

      //! Get the number of transforms.
      /** \returns The number of transform parameters for this effect. */
      NVSG_API size_t getNumberOfTransforms( void ) const;

      //! Get transform by index.
      /** \returns A reference to the CgFxTransform object at index \a index. */
      NVSG_API const CgFxTransform& getTransform( size_t index    //!< Zero based index.
                                                ) const;

      //! Get the number of tweakables.
      /** \returns The number of tweakable parameters for this effect. */
      NVSG_API size_t getNumberOfTweakables( void ) const;
      
      //! Get tweakable by index.
      /** \returns A handle for the tweakable parameter at index \a index. */
      NVSG_API CgFxParameterHandle getTweakable( size_t index   //!< Zero based index.
                                               ) const;

      //! Get tweakable by name.
      /** \returns A handle for the tweakable parameter if found, NULL otherwise. */
      NVSG_API CgFxParameterHandle getTweakable( const std::string& name   //!< name of the tweakable parameter.
                                               ) const;

      //! Get the number of Annotations
      /** \returns The number of annotations for an entity identified by \a hdl  */
      NVSG_API size_t getNumberOfAnnotations( CgFxEntityHandle hdl    //! Handle for the object holding annotations. 
                                            ) const;

      //! Get annotation by index.
      /** \returns A handle for the annotation if successful, NULL otherwise. */
      NVSG_API CgFxAnnotationHandle getAnnotation( CgFxEntityHandle hdl   //!< Handle for the object holding annotations.
                                                 , size_t index           //!< Zero based index of the annotation.
                                                 ) const;

      //! Get annotation by name.
      /** \returns A handle for the annotation if successful, a NULL otherwise. */
      NVSG_API CgFxAnnotationHandle getAnnotation( CgFxEntityHandle hdl   //!< Handle for the object holding annotations.
                                                 , const std::string& name     //!< Name of the annotation.
                                                 ) const;

      //! Get a CgFxParameter's value.
      /** A function of this function family obtains the current value for a CgFxParameter identified by by \a hdl. 
        * \note \a ValueType must be convertible to the embedded parameter's type. The embedded parameter's type
        * arises out of the parameter's base type and the parameter's dimension. */
      template<typename ValueType>
      void getValue( CgFxParameterHandle hdl    //!< Handle for the target parameter object.
                  , ValueType& value            //!< Reference to an object of type \a ValueType to receive the value.
                  ) const;

      //! \c getValue specialization for general pointer types.
      template<typename ValueType>
      void getValue( CgFxParameterHandle hdl    //!< Handle for the target parameter object.
                  , ValueType * value           //!< Pointer to an object of type \a ValueType to receive the value.
                  ) const;

      //! \c getValue specialization for 'pointer to char pointer' types.    
      NVSG_API void getValue( CgFxParameterHandle hdl   //!< Handle for the target parameter object.
                            , const char ** value       //!< Pointer to a 'C'-string to receive the string.
                            ) const;

      //! \c getValue specialization for string types.    
      NVSG_API void getValue( CgFxParameterHandle hdl   //!< Handle for the target parameter object.
                            , std::string& value             //!< Reference to a string object to receive the string.
                            ) const;

      //! Set a CgFxParameter's value.
      /** A function of this family updates the current value for a CgFxParameter identified by \a hdl. 
        * \note \a ValueType must be convertible to the embedded parameter's type. The embedded parameter's type
        * arises out of the parameter's base type and the parameter's dimension. */
      template<typename ValueType>
      void setValue( CgFxParameterHandle hdl    //!< Handle for the target parameter object.
                  , const ValueType& value      //!< Reference to an object of type \a ValueType which holds the value to set.
                  );

      template<typename ValueType>
      void setValue( CgFxParameterHandle hdl    //!< Handle for the target parameter object.
                  , const ValueType& value      //!< Reference to an object of type \a ValueType which holds the value to set.
                  ) const ;

      //! \c setValue specialization for general pointer types.
      template<typename ValueType>
      void setValue( CgFxParameterHandle hdl    //!< Handle for the target parameter object.
                  , const ValueType * value     //!< Pointer to an object of type \a ValueType which holds the value to set.
                  );

      template<typename ValueType>
      void setValue( CgFxParameterHandle hdl    //!< Handle for the target parameter object.
                  , const ValueType * value     //!< Pointer to an object of type \a ValueType which holds the value to set.
                  ) const;

      //! \c setValue specialization for 'pointer to char' types.    
      NVSG_API void setValue( CgFxParameterHandle hdl   //!< Handle for the target parameter object.
                            , const char * value        //!< Pointer to a 'C'-string which holds the string to set.
                            );
      
      NVSG_API void setValue( CgFxParameterHandle hdl   //!< Handle for the target parameter object.
                            , const char * value        //!< Pointer to a 'C'-string which holds the string to set.
                            ) const;

      //! \c setValue specialization for string types.    
      NVSG_API void setValue( CgFxParameterHandle hdl   //!< Handle for the target parameter object.
                            , const std::string& value       //!< Pointer to a string object which holds the string to set.
                            );
      
      NVSG_API void setValue( CgFxParameterHandle hdl   //!< Handle for the target parameter object.
                            , const std::string& value       //!< Pointer to a string object which holds the string to set.
                            ) const;

      //! Get the number of techniques.
      /** \returns The number of techniques available for this effect. */
      NVSG_API size_t getNumberOfTechniques( void ) const;

      //! Get the technique by index.
      /** \returns A handle for the CgFxTechnique object at index \a index. */
      NVSG_API CgFxTechniqueHandle getTechnique( size_t index   //!< Zero based index of the technique.
                                               ) const;

      // Get the current technique.
      /** \returns The zero based index of the technique currently used for rendering. */
      NVSG_API size_t getCurrentTechnique( void ) const;

      // Set the current technique.
      /** Makes the technique at index \a index the current one for subsequent rendering. */
      NVSG_API void setCurrentTechnique( size_t index   //!< Zero based index of the technique.
                                       );
      
      //! Validate the current technique
      /** Validates the current technique for subsequent rendering.
        * \returns \c true if validation was successfull, \c false otherwise. */
      NVSG_API bool validateCurrentTechnique( void ) const;
      
      //! Determine if the current technique is valid.
      /** \returns \c true if the current technique is valid, \c false otherwise. */ 
      NVSG_API bool isCurrentTechniqueValid( void ) const;

      //! Detect whether validation of the current technique is pending.
      /** \returns \c true if the current technique was not validated before, \c false otherwise. */
      NVSG_API bool isCurrentTechniqueValidationPending( void ) const;

      //! Get the number of render passes.
      /** \returns The number of render passes required for the technique identified by by \a tech. */ 
      NVSG_API size_t getNumberOfPasses( CgFxTechniqueHandle hdl    //! Handle for the technique to query.
                                       ) const;

      //! Get the render pass by index.
      /** \returns A handle for the CgFxPass object at index \a index of the technique identified by by \a tech. */
      NVSG_API CgFxPassHandle getPass( CgFxTechniqueHandle hdl    //! Handle for the technique to query.
                                     , size_t index               //! Zero base index of the render pass.
                                     ) const;

      //! Initiate the first rendering pass
      /** Initiates the first render pass for the current technique. Before the client code sends vertex data 
        * to the hardware for rendering, this call is needed to a.) set up the current technique for 
        * subsequent application on a per-pass basis, and b.) to apply the state for the first rendering pass.
        * \returns \c true if successful, \c false otherwise. */
      NVSG_API bool passFirstPass( void ) const;

      //! Introduce subsequent rendering passes
      /** Applies the state for the very next rendering pass if available. If no subsequent rendering pass 
        * is available, this function properly ends the application of the current technique.
        * \returns \c true if supsequent rendering passes are available, \c false otherwise. */
      NVSG_API bool passNextPass( void ) const;

      //--  Functions reimplemented from Object  --
      //! Shared data identifier
      /** Data that is shared between different CgFx attributes will be unambiguously identified by its data ID. 
        * \returns The data ID for the underlying data. */
      NVSG_API virtual DataID getDataID( void ) const;
      
      //! Data sharing
      /** A CgFX attribute can share its data with annother CgFx attribute to save memory.
        * \returns Whether the underlying data is shared. */
      NVSG_API virtual bool isDataShared( void ) const;

    protected:
      //! Default Constructor
      NVSG_API CgFx( void );

      //! Partial Constructor
      NVSG_API CgFx( const StateAttribute &rhs );

      //! Copy Constructor
      NVSG_API CgFx( const CgFx& rhs );
      
      /* prohibit explicit creation on stack by making the destructor protected */
      NVSG_API virtual ~CgFx(void);

    private: // private interface
      bool initEffect( void );
      void initAnnotations( CgFxEntity * p, size_t count );

    private: // private data
      ICgFXEffect            * m_effect;    //!< effect interface
      // 
      std::string                   m_file;      //!< copy of the effect's source file name
      //
      std::vector<CgFxTransform>    m_trafos;    //!< transforms required by the effect
      std::vector<CgFxParameter*>   m_tweaks;    //!< effect's tweakables
      //
      std::vector<CgFxTechnique*>   m_techs;     //!< effect's techniques
      mutable size_t           m_currTech;  //!< current technique used with rendering
      mutable size_t           m_lastCurrTech; //!< last technique that could be successfully validated
      mutable size_t           m_currPass;  //!< current rendering pass for current technique
      mutable size_t           m_numPasses; //!< number of passes required for current technique
  };


  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  // inline implementations can go here
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  //! Get texture file name
  /** If a texture filename is available for a texture type parameter identified by \a hdl, 
    * this function obtains the texture file name in \a fname.
    * \returns \c true if a texture file name could be found, \c false otherwise.
    * \note It doesn't make sense to call this function for non-texture type parameters. 
    * It will not crash, though. */
  NVSG_API bool CgFx_getTextureFileName( const CgFx * cgfx, CgFxParameterHandle hdl, std::string& file );

  //! Set texture file name
  /** For a texture type parameter identified by \a hdl, this function sets the 
    * texture file name to \a file.
    * \note Calling this function for non-texture type parameters yields undefind behaviour!  */
  NVSG_API void CgFx_setTextureFileName( CgFx * cgfx, CgFxParameterHandle hdl, const std::string& file );

  //! Get the name for a certain parameter
  /** \returns The name of the parameter identified by \a hdl. */
  inline const char * getParameterName( CgFxParameterHandle hdl )
  {
    __ASSERT(hdl);
    return hdl->desc.Name;
  }

  //! Get the name for a certain render pass
  /** \returns The name of the pass identified by \a hdl. */
  inline const char * getPassName( CgFxPassHandle hdl )
  {
    __ASSERT(hdl);
    return hdl->desc.Name;
  }

  //! Get the name for a certain render technique
  /** \returns The name of the technique identified by \a hdl. */
  inline const char * getTechniqueName( CgFxTechniqueHandle hdl )
  {
    __ASSERT(hdl);
    return hdl->desc.Name;
  }

  //! Get the semantic for a certain parameter
  /** \returns The semantic string for the parameter identified by \a hdl. */
  inline const char * getParameterSemantic( CgFxParameterHandle hdl )
  {
    __ASSERT(hdl);
    return hdl->desc.Semantic;
  }

  //! Get the type of a certain parameter
  /** \returns The enum type for the parameter identified by \a hdl. */
  inline CgFXPARAMETERTYPE getParameterType( CgFxParameterHandle hdl )
  {
    __ASSERT(hdl);
    return hdl->desc.Type;
  }

  //! Detect if the parameter identified by \a hdl is a scalar type.
  /** \returns \c true if the parameter is a scalar type, \c false otherwise. */
  inline bool isParameterScalarType( CgFxParameterHandle hdl )
  {
    __ASSERT(hdl);
    return hdl->desc.Dimension[1]==0 && hdl->desc.Dimension[0]==1; 
  }

  //! Detect if the parameter identified by \a hdl is a two-dimensional vector type.
  /** \returns \c true if the parameter is a two-dimensional vector type, \c false otherwise. */
  inline bool isParameterVector2Type( CgFxParameterHandle hdl )
  {
    __ASSERT(hdl);
    return hdl->desc.Dimension[1]==0 && hdl->desc.Dimension[0]==2;
  }

  //! Detect if the parameter identified by \a hdl is a three-dimensional vector type.
  /** \returns \c true if the parameter is a three-dimensional vector type, \c false otherwise. */
  inline bool isParameterVector3Type( CgFxParameterHandle hdl )
  {
    __ASSERT(hdl);
    return hdl->desc.Dimension[1]==0 && hdl->desc.Dimension[0]==3;
  }

  //! Detect if the parameter identified by \a hdl is a four-dimensional vector type.
  /** \returns \c true if the parameter is a four-dimensional vector type, \c false otherwise. */
  inline bool isParameterVector4Type( CgFxParameterHandle hdl )
  {
    __ASSERT(hdl);
    return hdl->desc.Dimension[1]==0 && hdl->desc.Dimension[0]==4;
  }

  //! Detect if the parameter identified by \a hdl is a 3x3 matrix type.
  /** \returns \c true if the parameter is a 3x3 matrix type, \c false otherwise. */
  inline bool isParameterMatrix33Type( CgFxParameterHandle hdl )
  {
    __ASSERT(hdl);
    return hdl->desc.Dimension[1]==3 && hdl->desc.Dimension[0]==3; 
  }

  //! Detect if the parameter identified by \a hdl is a 3x4 matrix type.
  /** \returns \c true if the parameter is a 3x4 matrix type, \c false otherwise. */
  inline bool isParameterMatrix34Type( CgFxParameterHandle hdl )
  {
    __ASSERT(hdl);
    return hdl->desc.Dimension[1]==4 && hdl->desc.Dimension[0]==3; 
  }

  //! Detect if the parameter identified by \a hdl is a 4x3 matrix type.
  /** \returns \c true if the parameter is a 4x3 matrix type, \c false otherwise. */
  inline bool isParameterMatrix43Type( CgFxParameterHandle hdl )
  {
    __ASSERT(hdl);
    return hdl->desc.Dimension[1]==3 && hdl->desc.Dimension[0]==4; 
  }

  //! Detect if the parameter identified by \a hdl is a 4x4 matrix type.
  /** \returns \c true if the parameter is a 4x4 matrix type, \c false otherwise. */
  inline bool isParameterMatrix44Type( CgFxParameterHandle hdl )
  {
    __ASSERT(hdl);
    return hdl->desc.Dimension[1]==4 && hdl->desc.Dimension[0]==4; 
  }

  //! Detect if the parameter identified by \a hdl is a texture type.
  /** \returns \c true if the parameter is a texture type, \c false otherwise. */
  inline bool isParameterTextureType( CgFxParameterHandle hdl )
  {
    __ASSERT(hdl);
    return hdl->desc.Type>=CgFXPT_TEXTURE1D && hdl->desc.Type<=CgFXPT_TEXTURE3D; 
  }

  //! Detect if the parameter identified by \a hdl is a sampler type.
  /** \returns \c true if the parameter is a sampler type, \c false otherwise. */
  inline bool isParameterSamplerType( CgFxParameterHandle hdl )
  {
    __ASSERT(hdl);
    return hdl->desc.Type>=CgFXPT_SAMPLER1D && hdl->desc.Type<=CgFXPT_SAMPLER3D; 
  }

  //! Detect whether the technique identified by \a hdl is valid.
  /** \returns \c true if the technique is valid, \c false otherwise. */
  inline bool isTechniqueValid( CgFxTechniqueHandle hdl )
  {
    __ASSERT(hdl);
    return hdl->valid;
  }

  inline bool CgFx::isEffectValid( void ) const
  {
    __TRACE();
    return m_effect!=NULL;
  }

  inline const std::string& CgFx::getEffectFile( void ) const 
  { 
    __TRACE();
    return m_file; 
  }

  inline size_t CgFx::getNumberOfTransforms( void ) const
  {
    __TRACE();
    return m_trafos.size();
  }

  inline const CgFxTransform& CgFx::getTransform( size_t index ) const
  {
    __TRACE();
    __ASSERT(index<m_trafos.size());
    return m_trafos[index];
  }

  inline size_t CgFx::getNumberOfTweakables( void ) const
  {
    __TRACE();
    return m_tweaks.size();
  }
      
  inline CgFxParameterHandle CgFx::getTweakable( size_t index ) const
  {
    __TRACE();
    __ASSERT(index<m_tweaks.size());
    return m_tweaks[index];
  }

  inline size_t CgFx::getNumberOfAnnotations( CgFxEntityHandle hdl ) const
  {
    __TRACE();
    __ASSERT(hdl);
    return hdl->annos.size();
  }

  inline CgFxAnnotationHandle CgFx::getAnnotation( CgFxEntityHandle hdl, size_t index ) const
  {
    __TRACE();
    __ASSERT(hdl && index<hdl->annos.size());
    return hdl->annos[index];
  }

  template<typename ValueType>
  inline void CgFx::getValue( CgFxParameterHandle hdl, ValueType& value ) const
  {
    //  __TRACE();  produces an internal compiler error
#if !defined(CGFX_NOT_SUPPORTED)
    __ASSERT(m_effect); // need the embedded effect successfully created beforehand
    __ASSERT(hdl->desc.Type!=CgFXPT_STRING); // must not use this for string parameters
    __ASSERT(hdl->desc.Bytes==sizeof(ValueType)); // sizes must match
    __VERIFY(SUCCEEDED(m_effect->GetValue(hdl->hdl, &value, hdl->desc.Bytes)));
#endif
  }

  template<typename ValueType>
  inline void CgFx::getValue( CgFxParameterHandle hdl, ValueType * value ) const
  {
    //  __TRACE();  produces an internal compiler error
#if !defined(CGFX_NOT_SUPPORTED)
    __ASSERT(m_effect); // need the embedded effect successfully created beforehand
    __ASSERT(hdl->desc.Type!=CgFXPT_STRING); // must not use this for string parameters
    __ASSERT(!(hdl->desc.Bytes%sizeof(ValueType))); // sizes must match
    __VERIFY(SUCCEEDED(m_effect->GetValue(hdl->hdl, value, hdl->desc.Bytes)));
#endif
  }

  inline void CgFx::getValue( CgFxParameterHandle hdl, const char ** value ) const
  {
    __TRACE();
#if !defined(CGFX_NOT_SUPPORTED)
    __ASSERT(m_effect); // need the embedded effect successfully created beforehand
    __ASSERT(hdl->desc.Type==CgFXPT_STRING); // must not use this for other than string parameters
    __VERIFY(SUCCEEDED(m_effect->GetString(hdl->hdl, (LPCSTR*)value)));
#endif
  }

  inline void CgFx::getValue( CgFxParameterHandle hdl, std::string& value ) const
  {
    __TRACE();
#if !defined(CGFX_NOT_SUPPORTED)
    __ASSERT(m_effect); // need the embedded effect successfully created beforehand
    __ASSERT(hdl->desc.Type==CgFXPT_STRING); // must not use this for other than string parameters
    LPCSTR s;
    __VERIFY(SUCCEEDED(m_effect->GetString(hdl->hdl, &s)));
    value=s;
#endif
  }

  template<typename ValueType>
  inline void CgFx::setValue( CgFxParameterHandle hdl, const ValueType& value )
  {
    //  __TRACE();  produces an internal compiler error
#if !defined(CGFX_NOT_SUPPORTED)
    __ASSERT(m_effect); // need the embedded effect successfully created beforehand
    __ASSERT(hdl->desc.Type!=CgFXPT_STRING); // must not use this for string parameters
    __ASSERT(hdl->desc.Bytes==sizeof(ValueType)); // sizes must match
    __VERIFY(SUCCEEDED(m_effect->SetValue(hdl->hdl, &value, hdl->desc.Bytes)));
    ++m_incarnation; // force new validation 
#endif
  }

  template<typename ValueType>
  inline void CgFx::setValue( CgFxParameterHandle hdl, const ValueType * value )
  {
    //  __TRACE();  produces an internal compiler error
#if !defined(CGFX_NOT_SUPPORTED)
    __ASSERT(m_effect); // need the embedded effect successfully created beforehand
    __ASSERT(hdl->desc.Type!=CgFXPT_STRING); // must not use this for string parameters
    __ASSERT(!(hdl->desc.Bytes%sizeof(ValueType))); // sizes must match
    __VERIFY(SUCCEEDED(m_effect->SetValue(hdl->hdl, value, hdl->desc.Bytes)));
    ++m_incarnation; // force new validation 
#endif
  }

  inline void CgFx::setValue( CgFxParameterHandle hdl, const char * value )
  {
    __TRACE();
#if !defined(CGFX_NOT_SUPPORTED)
    __ASSERT(m_effect); // need the embedded effect successfully created beforehand
    __ASSERT(hdl->desc.Type==CgFXPT_STRING); // must not use this for other than string parameters
    __VERIFY(SUCCEEDED(m_effect->SetString(hdl->hdl, (LPCSTR)value)));
    ++m_incarnation; // force new validation 
#endif
  }

  inline void CgFx::setValue( CgFxParameterHandle hdl, const std::string& value )
  {
    __TRACE();
#if !defined(CGFX_NOT_SUPPORTED)
    __ASSERT(m_effect); // need the embedded effect successfully created beforehand
    __ASSERT(hdl->desc.Type==CgFXPT_STRING); // must not use this for other than string parameters
    __VERIFY(SUCCEEDED(m_effect->SetString(hdl->hdl, value.c_str())));
    ++m_incarnation; // force new validation 
#endif
  }

  template<typename ValueType>
  inline void CgFx::setValue( CgFxParameterHandle hdl, const ValueType& value ) const
  {
    __TRACE();
#if !defined(CGFX_NOT_SUPPORTED)
    __ASSERT(m_effect); // need the embedded effect successfully created beforehand
    __ASSERT(hdl->desc.Type!=CgFXPT_STRING); // must not use this for string parameters
    __ASSERT(hdl->desc.Bytes==sizeof(ValueType)); // sizes must match
    __VERIFY(SUCCEEDED(m_effect->SetValue(hdl->hdl, &value, hdl->desc.Bytes)));
#endif
  }

  template<typename ValueType>
  inline void CgFx::setValue( CgFxParameterHandle hdl, const ValueType * value ) const
  {
    __TRACE();
#if !defined(CGFX_NOT_SUPPORTED)
    __ASSERT(m_effect); // need the embedded effect successfully created beforehand
    __ASSERT(hdl->desc.Type!=CgFXPT_STRING); // must not use this for string parameters
    __ASSERT(!(hdl->desc.Bytes%sizeof(ValueType))); // sizes must match
    __VERIFY(SUCCEEDED(m_effect->SetValue(hdl->hdl, value, hdl->desc.Bytes)));
#endif
  }

  inline void CgFx::setValue( CgFxParameterHandle hdl, const char * value ) const
  {
    __TRACE();
#if !defined(CGFX_NOT_SUPPORTED)
    __ASSERT(m_effect); // need the embedded effect successfully created beforehand
    __ASSERT(hdl->desc.Type==CgFXPT_STRING); // must not use this for other than string parameters
    __VERIFY(SUCCEEDED(m_effect->SetString(hdl->hdl, (LPCSTR)value)));
#endif
  }

  inline void CgFx::setValue( CgFxParameterHandle hdl, const std::string& value ) const
  {
#if !defined(CGFX_NOT_SUPPORTED)
    __TRACE();
    __ASSERT(m_effect); // need the embedded effect created beforehand
    __ASSERT(hdl->desc.Type==CgFXPT_STRING); // must not use this for other than string parameters
    __VERIFY(SUCCEEDED(m_effect->SetString(hdl->hdl, value.c_str())));
#endif
  }

  inline size_t CgFx::getNumberOfTechniques( void ) const
  {
    __TRACE();
    return m_techs.size();
  }

  inline CgFxTechniqueHandle CgFx::getTechnique( size_t index ) const
  {
    __TRACE();
    __ASSERT(index<m_techs.size());
    return m_techs[index];
  }

  inline size_t CgFx::getCurrentTechnique( void ) const
  {
    __TRACE();
    return m_currTech;
  }

  inline void CgFx::setCurrentTechnique( size_t index )
  {
    __TRACE();
    __ASSERT(index<m_techs.size());
    m_currTech = index;
    ++m_incarnation; // force new validation 
  }
      
  inline bool CgFx::isCurrentTechniqueValid( void ) const
  {
    __TRACE();
    return (m_currTech<m_techs.size()) ? m_techs[m_currTech]->valid : false;
  }

  inline bool CgFx::isCurrentTechniqueValidationPending( void ) const
  {
    __TRACE();
    return m_lastCurrTech != m_currTech;
  }

  inline size_t CgFx::getNumberOfPasses( CgFxTechniqueHandle hdl ) const
  {
    __TRACE();
    __ASSERT(hdl);
    return hdl->passes.size();
  }

  inline CgFxPassHandle CgFx::getPass( CgFxTechniqueHandle hdl, size_t index ) const
  {
    __TRACE();
    __ASSERT(hdl);
    __ASSERT(index<hdl->passes.size());
    return hdl->passes[index];
  }

} // namespace nvsg
