// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include "nvsgcommon.h"

#include <vector>
#include <map>
#include "nvutil/Singleton.h"

namespace nvsg
{
class DALServer;
class DALData;

typedef unsigned long HDAL; // handle to a Device Abstraction Link
extern const HDAL HDAL_INVALID; 

//! Device Abstraction Link (DAL) Host
/** A DALHost can host arbitrary DALData. DALData can be used to abstract device dependent data
  * that will be linked to device independent data via a Device Abstraction Link (DAL). */
class DALHost
{
public:
  //! Stores Device Abstraction Link Data.
  /** Attaches the data identified by \a dataID and pointed to by \a data 
    * to the host's internal Device Abstraction Link (DAL).
    *
    * \note It is an error to store the same DALData object twice, even with a different host. 
    * Doing so will result in undefined behavior!
    *
    * \returns A handle to the Device Abstraction Link the data was attached to, 
    * or DAL_INVALID if the function fails. */
  NVSG_API HDAL storeDeviceAbstractionLinkData(unsigned int dataID, DALData * data) const;
  
  //! Releases Device Abstraction Link Data.
  /** Releases the host's internal Device Abstraction link with all DALData atached to it. */
  NVSG_API void releaseDeviceAbstractionLinkData();
  
  //! Passes back Device Abstraction Link Data.
  /** Fills the vector \a data with pointers to DALData objects that are identified by \a dataID,
    * and that are attached to the host's internal Device Abstraction Link (DAL).
    * \returns \c true if any requested data was passed back, \c false otherwise. */
  NVSG_API bool getDeviceAbstractionLinkData(unsigned int dataID, std::vector<DALData*>& data) const;

  //! Returns the last recently used Device Abstraction Link Data.
  /** \returns A pointer to the DALData object that was last set through 
    * setLastRecentlyUsedDeviceAbstractionLinkData for this host, 
    * or NULL if that data has become invalid in between. */
  NVSG_API DALData * getLastRecentlyUsedDeviceAbstractionLinkData() const;

  //! Sets the last recently used Device Abstraction Link Data.
  /** Sets the the last recently used Device Abstraction Link Data attached to this host. This
    * is intended to perform faster DALData lookups. */
  NVSG_API void setLastRecentlyUsedDeviceAbstractionLinkData(DALData * lastRecentlyUsed) const;

#if !defined(DOXYGEN_IGNORE)
public:
  DALHost();
  DALHost(const DALHost&);
  ~DALHost();
private:
  mutable HDAL m_hDAL;
  mutable DALData * m_lastRecentlyUsed;
#endif
};

//! Creating DALData to be exchanged via a Device Abstraction Link requires this particular interface. 
class DALDataCreator
{
public:
#if !defined(DOXYGEN_IGNORE)
  virtual ~DALDataCreator() {}
#endif
  //! Called from the framework immediately before the Device Abstraktion Link (DAL) becomes invalid.
  /** Called from the framework to inform creators that the DAL identified by \a hDAL immediately becomes invalid.
    * This is for example the case when the object that host the DAL will be deleted. 
    *
    * A concrete DALDataCreator ususally overrides this function to dismiss all temporary stored handles
    * to the DALs that creator has attached DALData to. Note that the creator must not explicitely delete
    * the attached data! The deletetion of the attached data will be performed by the framework, 
    * immediately after this function was called. */  
  virtual void onReleaseDAL(HDAL hDAL) {}

protected:
  //! DALDataCreator serves as base class only. 
  DALDataCreator() {}

  //! Releases Device Abstraction Link Data.
  /** Releases the data identified by \a dataID that was attached to the 
    * Device Abstraction Link identified by \a hDAL.
    *
    * This function should be called immediately befor this creator becomes invalid, to free
    * all the data that was created by this creator and attached to the DAL identified by \a hDAL. 
    * The function deletes the data and removes the according entry from the DAL. */
  NVSG_API void releaseDeviceAbstractionLinkData(HDAL hDAL, unsigned int dataID);
};

//! Device Abstraction Link Data.
/** DALData can be used to abstract device dependent data that will be linked to device independent
  * data via a Device Abstraction Link (DAL). 
  * 
  * Note that only a DALDataCreator is permitted to create DALData. DALData can be attached to a DAL
  * using the procedural interface for either a DALServer or a DALHost. */
class DALData
{
#if !defined(DOXYGEN_IGNORE)
  friend class DALServer;
  friend class DALDataCreator;
protected:
  virtual ~DALData();
#endif

protected:
  //! Initializes a DALData object.
  DALData(
    DALDataCreator * creator //! The creator of this DALData.
  , const DALHost * host //! The host of this DALData. 
  );

  //! Induces the deletion of this DALData object.
  /** This function will be called from the DALServer framework when this DALData object
    * is going to be released from a Device Abstraction Link.
    *
    * The default implementation calls delete on the this pointer. 
    * If a custom memory manager is used, in particular a memory manager that does not
    * issue new and delete for creating and deleting a DALData object, it is recommended
    * to override this function, so the deletion of the DALData object will be performed
    * in the correct way. */
  virtual void deleteThis() { delete this; }

#if !defined(DOXYGEN_IGNORE)
private:
  DALDataCreator * m_creator;
  const DALHost * m_host;
#endif
};

//! Device Abstraction Link Server
class DALServer
{
  friend class DALDataCreator;
  friend class DALHost;

#if !defined(DOXYGEN_IGNORE)
public:
  // Initializes the DALServer at instantiation
  DALServer();
  // Cleanup after deletion
  ~DALServer();
private:
  // Attaches the data identified by \a dataID and pointed to by \a data 
  // to the Device Abstraction Link (DAL) identified by \a hDAL.
  // If HDAL_INVALID is passed for \a hDAL, a new DAL will be created first.
  // \returns A handle to the Device Abstraction Link or DAL_INVALID if the function fails.
  HDAL storeDALData(HDAL hDAL, unsigned int dataID, DALData* data);
  // Releases the data identified by \a dataID that was attached to the DAL identified by \a hDAL 
  // by the creator pointed to by \a creator.
  // If 0xFFFFFFFF is passed for \a dataID, all data that was attached to the DAL identified by \a hDAL
  // by the creator pointed to by \a creator will be released.
  // If NULL is passed for \a creator, all data attached to the DAL identified by \a hDAL will be released
  // and the DAL will be deleted afterwards.
  void releaseDALData(HDAL hDAL, DALDataCreator * creator=NULL, unsigned int dataID=0xFFFFFFFF);
  // Provides all addresses of the data objects identified by \a dataID that are currently stored 
  // with the DAL identified by \a hDAL. 
  // The addresses of the data objects will be copied to the vector referenced by \a data.
  bool getDALData(HDAL hDAL, unsigned int dataID, std::vector<DALData*>& data) const;
private:
  std::map<HDAL, std::multimap<unsigned int, DALData*> > m_DALMap; // map of DALs
  std::vector<HDAL> m_freeHDALs; // free DAL handles can be re-used
  HDAL m_hDALNext; // use this if no handle is available for re-use
#endif
};

#if !defined(DOXYGEN_IGNORE)
typedef nvutil::Singleton<DALServer> DALS;
#endif

inline HDAL DALHost::storeDeviceAbstractionLinkData(unsigned int dataID, DALData * data) const
{
  m_hDAL = DALS::instance()->storeDALData(m_hDAL, dataID, data);
  return m_hDAL;
}

inline void DALHost::releaseDeviceAbstractionLinkData()
{
  if ( m_hDAL != HDAL_INVALID )
  {
    DALS::instance()->releaseDALData(m_hDAL);
    m_hDAL = HDAL_INVALID;
    m_lastRecentlyUsed = NULL;
  }
}

inline bool DALHost::getDeviceAbstractionLinkData(unsigned int dataID, std::vector<DALData*>& data) const
{
  return DALS::instance()->getDALData(m_hDAL, dataID, data);
}

inline DALData * DALHost::getLastRecentlyUsedDeviceAbstractionLinkData() const
{
  return m_lastRecentlyUsed;
}

inline void DALHost::setLastRecentlyUsedDeviceAbstractionLinkData(DALData * lastRecentlyUsed) const
{
  m_lastRecentlyUsed = lastRecentlyUsed;
}

inline void DALDataCreator::releaseDeviceAbstractionLinkData(HDAL hDAL, unsigned int dataID)
{
  __ASSERT(hDAL != HDAL_INVALID);
  DALS::instance()->releaseDALData(hDAL, this, dataID);
}

} // namespace nvsg