// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** @file */

#include "nvsgcommon.h"

#include "nvsg/Object.h"

//  forward declarations
namespace nvmath
{
  class Sphere3f;
}
namespace nvsg
{
  class GeoNode;
}

namespace nvsg
{
  //! Base class of all drawable objects.
  /** Defines the interface that every drawable object (e.g. GeoSet) has to resolve.  */
  class Drawable : public Object
  {
    public:
      // ! Resolve construction of the right drawable type through late binding.
      /** \returns A pointer to the newly created Drawable object. */
      NVSG_API virtual const Drawable * clone( void ) const = 0;

      //! Ask if the bounding sphere of this Node is valid.
      /** If the bounding sphere isn't valid, it is tried to validate it.
        * \returns The valid state of the bounding sphere. */
      NVSG_API virtual bool isBoundingSphereValid( void ) const = 0;

      //! Get the bounding sphere of this Node.
      /** The bounding sphere of a Node is a sphere that completely contains the Node and all its children.
        * \note It is assumed that the bounding sphere of this Node is valid.
        * \returns A constant reference to the bounding sphere. */
                               
      NVSG_API virtual const nvmath::Sphere3f & getBoundingSphere( void ) const = 0;

            //! Invalidate the number of frames information.
      NVSG_API void invalidateNumberOfFrames( void ) const;

      //! Get the number of animation frames.
      /** \returns 0  */
      NVSG_API virtual size_t getNumberOfFrames( void ) const;

      //! Get the number of owners of this Drawable.
      /** \returns  The number of owners if this Drawable */
      NVSG_API size_t getNumberOfOwners( void ) const;

      //! Get a constant pointer to the owner node at position \a index.
      /** \returns  The constant pointer to the owner node at position \a index */
      NVSG_API const GeoNode * getOwner( size_t index    //!<  The position to get the owner at
                                       ) const;

      //--  Functions reimplemented from Object  --
      NVSG_API virtual DataID getDataID( void ) const;
      NVSG_API virtual bool isDataShared( void ) const;

    protected:
      //! Protected constructor
      NVSG_API Drawable( void );
      // Copy constructor
      NVSG_API Drawable( const Drawable& );

      //! Protected destructor to prevent explicit creation on stack.
      NVSG_API virtual ~Drawable( void );

      //! Invalidate bounding sphere information
      NVSG_API virtual void invalidateBoundingSphere( void ) const;

    private:
      // A Drawable can be the child of more than one parent. The child-to-parent connection 
      // is used, for example, when invalidating the owner GeoNode's BoundingSphere while 
      // invalidating the BoundingSphere of a Drawable.
      friend class GeoNode;
      void addOwner( const GeoNode * node );
      void removeOwner( const GeoNode * node );

    private:
      std::vector<const GeoNode*>  m_owners;   //!<  vector of parent geo nodes
  };

  inline size_t Drawable::getNumberOfOwners( void ) const
  {
     __TRACE();
    return( m_owners.size() );
  }

  inline const GeoNode * Drawable::getOwner( size_t index ) const
  {
    __TRACE();
    __ASSERT( index < m_owners.size() );   
    return( m_owners[index] );
  }
} // namespace nvsg

