// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** @file */

#include "nvsgcommon.h"

#include  "nvsg/Node.h"

namespace nvsg
{
  //! Group node.
  /** A Group is a base class for grouping nodes that contain children, such as  
    * Transform or Switch - in other words, nodes that are defined relative to
    * this node.
    */
  class Group : public Node
  {
    public:
      //! Create a Group.
      /** \returns a constant pointer to a Group. */
      NVSG_API static const Group * create( void );

      //! Create a Group by copying from a Node.
      /** \returns a constant pointer to a Group. */
      NVSG_API static const Group * createFromBase( const Node &rhs  //!<  Node to copy from
                                                  );

      //! Get a clone of this \c Group object.
      /** \returns A constant pointer to the new created clone. */
      NVSG_API virtual const Group * clone( void ) const;

      //! Determine whether the data of this object is shared.
      /** \returns  \c true if the object's data is shared, otherwise \c false.  */
      NVSG_API virtual bool isDataShared( void ) const;

      //! Get the DataID of the data of this object.
      /** \returns  DataID of the object's data.  */
      NVSG_API virtual DataID getDataID( void ) const;

      //! Returns the number of children in this group.
      /** \returns  The number of children in this group.  */
      NVSG_API size_t getNumberOfChildren( void ) const;

      //! Add a child to this group.
      /** Adds the node \a pChild at position \a position. 
        * If \a position is larger than the group size, the child is
        * added to the end. By default, the child is added to the end.
        * \returns  The actual position of added child
        * \note The reference count of the new child will be incremented,
        * while the reference count of the former child will be decremented. */
      NVSG_API size_t addChild( const Node * pChild                   //!<  child to add to this group
                              , size_t position = 0xFFFFFFFF    //!<  position to put \a pChild in the list
                              );
      
      //! Returns a pointer to the const child Node at position \a index.
      /** \returns  A pointer to the const Node at position \a index.
        * \note the behaviour is undefined if called with an invalid index. */
      NVSG_API const Node * getChild( size_t index    //!< Zero-based index to get child at.
                                    ) const;

      //! Remove a child from this group
      /** \returns \c true if the child \a pChild 
        * was removed, otherwise \c false  
        * \note The reference count of the removed 
        * child will be decremented.*/
      NVSG_API bool removeChild( const Node * pChild   //!<  child node to remove from this group
                               );

      //! Remove the child at given position.
      /** If the position is not valid in this group, nothing is removed.
        * \returns  \c true on successful removal, otherwise \c false 
        * \note The reference count of the removed 
        * child will be decremented.*/
      NVSG_API bool removeChild( size_t position    //!<  position of child to remove from this group
                               );

      //! Replace a child of this Group by an other.
      /** Replaces the \a oldChild by the \a newChild, keeping the position in the Group constant. If oldChild is not a
        * child of this Group, nothing is changed and \c false is returned.
        * \returns \c true on successful replacement, otherwise \c false */
      NVSG_API bool replaceChild( const Node * newChild   //!<  new child
                                , const Node * oldChild   //!<  child to be replaced
                                );

      //! Replace a child of this Group by an other
      /** Replaces the child at \a index by \a newChild. If the index is not valid, nothing is changed and \c false is
        * returned.
        * \returns \c true on successful replacement, otherwise \c false */
      NVSG_API bool replaceChild( const Node * newChild   //!<  new child
                                , size_t position         //!<  position of child to be replaced
                                );
      
      //! Check if node is a child of this group.
       /** \return \c true - The node is in this group; \c false - The node is not in this group.*/
      NVSG_API bool hasChild( const Node * node    //!< Node pointer. 
                            ) const;

      // Retrieve the index of the given node.
      /** \return \c true - The node is a child of this group; 
        * \c false - The node is not a child of this group.\n
        * The result of \a index is not defined when the 
        * return value is false.
        * \note This method only retrieves the first 
        * appearance of the node in this group! */
      NVSG_API bool getIndex( const Node * node   //!< Node pointer.
                            , size_t & index      //!< Resulting index.
                            ) const;

      //! Ask if this Group contains any CgFx attributes.
      /** \returns \c true if the Group contains a child with a CgFx attribute, otherwise \c false.  */
      NVSG_API virtual bool containsCgFx( void ) const;

      //! Invalidate the CgFx containment cache.
      /** The cache is invalidated and the invalidation of a Node is called.  */
      NVSG_API virtual void invalidateCgFxContainment( void );

      //! Ask if this Group contains any LightSource.
      /** \returns \c true if the Group contains a LightSource as a child, otherwise \c false.  */
      NVSG_API virtual bool containsLight( void ) const;

      //! Invalidate the LightSource containment cache.
      /** The cache is invalidated and the invalidation of a Node is called.  */
      NVSG_API virtual void invalidateLightContainment( void );

      //! Ask if this Group contains any transparent Material attributes.
      /** \returns \c true if the Group contains a child with a transparent 
        * Material attribute, otherwise \c false.  */
      NVSG_API virtual bool containsTransparentMaterial( void ) const;

      //! Invalidate the transparent Material containment cache.
      /** The cache is invalidated and the invalidation of a Node is called.  */
      NVSG_API virtual void invalidateTransparentMaterialContainment( void );

      //! Ask if this Group contains any transparent Texture.
      /** \returns \c true if the Group contains a child with a transparent 
        * Texture, otherwise \c false.  */
      NVSG_API virtual bool containsTransparentTexture( void ) const;

      //! Invalidate the transparent Texture containment cache.
      /** The cache is invalidated and the invalidation of a Node is called.  */
      NVSG_API virtual void invalidateTransparentTextureContainment( void );

      //! Get the number of animation frames that are part of this Node.
      /** \returns The number of animation frames of at least one child.  */
      NVSG_API virtual size_t getNumberOfFrames( void ) const;

      //! Invalidate the number of animation frames cache.
      /** The cache is invalidated and the invalidation of a Node is called.  */
      NVSG_API virtual void invalidateNumberOfFrames( void ) const;

    protected:
      //! Protected constructor
      NVSG_API Group( void );

      //! Partial Constructor
      NVSG_API Group( const Node &rhs );

      //! Protected copy constructor
      NVSG_API Group( const Group& rhs );

      //! Protected destructor to prevent explicit creation on stack.
      NVSG_API virtual ~Group(void);

      //! Calculate the bounding sphere of this group.
      /** If this group doesn't contain any real geometry (for example only light sources without a bounding sphere),
        * it doesn't have a valid bounding sphere.
        * \returns \c true if the calculated bounding sphere is valid, otherwise \c false. */
      NVSG_API virtual bool calcBoundingSphere( void ) const;

      //! Invalidate all caches of this Group, possibly depending on the added or removed Node.
      /** Invalidates the bounding sphere of this Group. If the CgFx containment or the LightSource containment
        * changes by adding/removing \a pChild, the corresponding caches are invalidated. */
      void  invalidateCaches( const Node *pChild    //!<  Node that has been added or removed
                            , bool insert           //!<  \c true, if the Node has been added, otherwise \c false
                            );

    private:
      std::vector<const Node*> m_children;       //!<  Array of children

      //  State caches
      mutable bool    m_containsCgFx;
      mutable bool    m_containsCgFxValid;
      mutable bool    m_containsLight;
      mutable bool    m_containsLightValid;
      mutable bool    m_containsTransparentMaterial;
      mutable bool    m_containsTransparentMaterialValid;
      mutable bool    m_containsTransparentTexture;
      mutable bool    m_containsTransparentTextureValid;
      mutable size_t  m_numberOfFrames;
      mutable bool    m_numberOfFramesValid;
  };
  
  inline size_t Group::getNumberOfChildren( void ) const
  {
    __TRACE();
    return( m_children.size() );
  }

  inline const Node * Group::getChild( size_t index ) const
  {
    __TRACE();
    __ASSERT(index<m_children.size()); // undefined behaviour for invalid index
    return m_children[index];
  }

  inline  void  Group::invalidateCaches( const Node *pChild, bool insert )
  {
    __TRACE();
    invalidateBoundingSphere();
    if ( pChild->containsCgFx() && ! ( insert && containsCgFx() ) )
    {
      invalidateCgFxContainment();
    }
    if ( pChild->containsLight() && ! ( insert && containsLight() ) )
    {
      invalidateLightContainment();
    }
    if ( pChild->containsTransparentMaterial() && ! ( insert && containsTransparentMaterial() ) )
    {
      invalidateTransparentMaterialContainment();
    }
    if ( pChild->containsTransparentTexture() && ! ( insert && containsTransparentTexture() ) )
    {
      invalidateTransparentTextureContainment();
    }
  }
} //  namespace nvsg
