// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** @file */

#include "nvsgcommon.h"

#include  "nvsg/SimpleCamera.h"

namespace nvsg
{
  //! A jittering camera.
  /** A JitterCamera is a MonoCamera that produces an average of a number of slightly jittered projections. This results
    * in full-scene antialiased images.
    * \par  Jittering
    * If you need to take nine or sixteen samples to antialias an image, you might think that the best choice of points is
    * an equally spaced grid across the pixel. Surprisingly, this is not necessarily true. In fact, sometimes it's a
    * good idea to take points that lie in adjacent pixels. You might want a uniform distribution or a normalized
    * distribution, clustering toward the center of the pixel. Following is a short example code showing how to set jitters
    * for a JitterCamera:
    * \code
    *   vector<Vec2f> jitters;
    *   jitters.resize(2);
    *   jitters[0] = Vec2f( 0.25f, 0.75f );
    *   jitters[1] = Vec2f( 0.75f, 0.25f );
    *   jitterCamera->setJitters( jitters );
    * \endcode
    * Following are a few sets of reasonable jitter values to use for selected sample counts. All of these
    * examples are uniformly distributed in the pixel and all lie within the pixel.
    * \code
    *   Count   Values
    *     2     { 0.25,  0.75 }, { 0.75,  0.25 }
    *     4     { 0.375, 0.25 }, { 0.125, 0.75 }, { 0.875, 0.25 }, { 0.625, 075 }
    *     8     { 0.5625, 0.4375 }, { 0.0625, 0.9375 }, { 0.3125, 0.6875 }, { 0.6875, 0.8125 },
    *           { 0.8125, 0.1875 }, { 0.9375, 0.5625 }, { 0.4375, 0.0625 }, { 0.1875, 0.3125 }
    * \endcode
    */
  class JitterCamera : public MonoCamera
  {
    public:
      //! Create a JitterCamera.
      /** \returns a constant pointer to a JitterCamera. */
      NVSG_API static const JitterCamera * create( const SimpleCamera &base                   //!< base camera to use
                                                 , const std::vector<nvmath::Vec2f> &jitters  //!< vectors of jitter
                                                 );

      NVSG_API static const JitterCamera * createFromBase( const MonoCamera &rhs                      //! MonoCamera to copy from
                                                         , const SimpleCamera &base                   //! base camera to use
                                                         , const std::vector<nvmath::Vec2f> &jitters  //! vectors of jitter
                                                         );

      //! Create a clone of this JitterCamera.
      /** \returns  const pointer to the new JitterCamera as a pointer to a MonoCamera */
      NVSG_API virtual const JitterCamera * clone( void ) const;

      //! Determine whether the data of this object is shared.
      /** \returns  c\ true if the object's data is shared, otherwise \c false.  */
      NVSG_API virtual bool isDataShared()  const;

      //! Get the DataID of the data of this object.
      /** \returns The DataID of the object's data.  */
      NVSG_API virtual DataID getDataID() const;

      //
      // Jitter interface
      //

      //! Get the jitters.
      /** The jitter vector of Vec2f defines the number of passes and the jitter 
        * of each pass. The jitter units are pixels. */
      NVSG_API void getJitters( std::vector<nvmath::Vec2f> &jitters    //!<  vector of Vec2f to fill with jitters
                              ) const;

      //! Set the jitters.
      /** The size of \a jitters gives the number of passes to make, 
        * and the Vec2f elements determine the amount to jitter
        * in each pass. The jitter units are pixels. */
      NVSG_API void setJitters( const std::vector<nvmath::Vec2f> &jitters    //!<  vector of jitters
                              );

      //! Get the number of passes.
      /** \returns The number of jittering passes (actually the size of the jitter vector)*/
      NVSG_API size_t getNumberOfPasses() const;

      //! Update the simple base cameras with the current camera settings and the viewport size.
      /** The viewport width and height are used to scale the jitter values, 
        * allowing the jitters to be in pixel units. */
      NVSG_API void updateSimpleCameras( size_t vpWidth   //!<  viewport width
                                       , size_t vpHeight  //!<  viewport height
      ) const;

      //! Get the base camera.
      /** The base camera of a jitter camera is a SimpleCamera that 
        * is used in multiple passes, each one slightly jittered.
        * \returns  A const pointer to the base camera */
      NVSG_API const SimpleCamera * getBaseCamera() const;

      //! Get camera. 
      /** \returns Camera at index \a index. */
      NVSG_API const SimpleCamera * getCamera( size_t index //!< Zero based index.
                                             ) const;

      // 
      // MonoCamera interface
      //

      //! Get the projection matrix
      /** \returns  The projection matrix (unjittered) */
      NVSG_API virtual nvmath::Mat44f getProjection() const;

      //! Get the projection translation (used with stereo only).
      /** \returns  The projection translation in x (unjittered) */
      NVSG_API virtual float getProjectionDx() const;

      //! Get the inverse projection matrix
      /** \returns The inverse projection matrix (unjittered) */
      NVSG_API virtual nvmath::Mat44f getInverseProjection()  const;

      //! Set the projection translation (used with stereo only).
      NVSG_API virtual void setProjectionDx( float dx   //!<  shift of projection in x-direction
                                           );
 
      //
      // head light handling interfaces
      //

      //! Interface for adding a headlight to this camera.
      NVSG_API virtual void addHeadLight(
        const LightSource * pLight                                                //!<  A pointer to the LightSource to add.
      , const nvmath::Vec3f &position = nvmath::Vec3f( 0.0f, 0.0f, 0.0f )         //!<  The position relative to the camera.
      , const nvmath::Quatf &quat     = nvmath::Quatf( 0.0f, 0.0f, 0.0f, 1.0f )   //!<  The rotation relative to the camera.
      );

      //! Interface for removing a headlight.
      /** \returns \c true if headlight was removed, otherwise \c false */
      NVSG_API virtual bool removeHeadLight( size_t index = 0   //!<  index of head light to remove
                                           );

      //! Interface for getting the number of headlights attached to the camera.
      /** \returns  The number of headlights attached to this camera. */
      NVSG_API virtual size_t getNumberOfHeadLights() const;

      //! Interface for getting a headlight of this camera.
      /** \returns A constant pointer to the headlight at index \a index */
      NVSG_API virtual const LightSource* getHeadLight( size_t index   //!<  Index of headlight to get
                                                      ) const;

      //! Interface for getting the position/orientation of a camera's headlight.
      NVSG_API virtual void getHeadLightTrafo( nvmath::Vec3f &trans  //!<  The translation relative to the camera will be received here.
                                             , nvmath::Quatf &quat   //!<  The rotation relative to the camera will be received here.
                                             , size_t index = 0      //!<  Index of the headlight to get transform information from.
                                             ) const;

      //! Interface for setting the headlight position and orientation relative to the camera.
      /** \returns \c true if a head light with this index exists, \c false otherwise. */
      NVSG_API virtual bool setHeadLightTrafo( const nvmath::Vec3f &trans   //!<  position of headlight - default is (0,0,0)
                                             , const nvmath::Quatf &quat    //!<  orientation of headlight - default is (0,0,0,1) 
                                             , size_t index = 0             //!<  index of headlight
                                             );


      //
      // basic camera interface
      //

      //! Interface for getting the viewing direction.
      /** \returns The viewing direction  */
      NVSG_API virtual nvmath::Vec3f getDirection() const;

      //! Interface for setting the viewing direction.
      NVSG_API virtual void setDirection( const nvmath::Vec3f& dir    //!<  new viewing direction (might change upVector)
                                        );

      //! Interface for getting the distance to the target point.
      /** \returns The distance to the target point */
      NVSG_API virtual float getTargetDistance() const;

      //! Interface for setting the distance to the target point.
      NVSG_API virtual void setTargetDistance( float td   //!<  distance to target point (or viewing plane)
                                             );

      //! Interface for getting the camera position.
      /** \returns The camera position */
      NVSG_API virtual const nvmath::Vec3f& getPosition() const;

      //! Interface for setting the camera position.
      NVSG_API virtual void setPosition( const nvmath::Vec3f& pos   //!<  new camera position
                                       );

      //! Interface for setting the up direction.
      NVSG_API virtual void setUpVector( const nvmath::Vec3f &up    //!<  up direction (might change viewing direction)
                                       );

      //! Interface for getting the up direction.
      /** \returns The camera up direction */
      NVSG_API virtual nvmath::Vec3f getUpVector() const;

      //! Interface for setting the camera orientation.
      NVSG_API virtual void setOrientation( const nvmath::Quatf &quat   //!<  quaternion that determines direction and up vector
                                          );

      //! Interface for setting the camera orientation.
      NVSG_API virtual void setOrientation( const nvmath::Vec3f &dir    //!<  direction to look at
                                          , const nvmath::Vec3f &up     //!<  up vector
                                          );

      //! Interface for setting the camera orientation.
      NVSG_API virtual void setOrientation( const nvmath::Vec3f &axis   //!<  axis to rotate about
                                          , float angle         //!<  angle to rotate
      );

      //! Get the current orientation.
      /** \returns  quaternion that describes the orientation */
      NVSG_API virtual nvmath::Quatf getOrientation( void ) const;

      //! Interface for getting the world-to-view transformation.
      /** \returns world-to-camera transformation */
      NVSG_API virtual nvmath::Mat44f getWorldToViewMatrix() const;

      //! Interface for getting the inverse world-to-view transformation.
      /** \returns inverse of the world-to-camera transformation (i.e. camera to world) */
      NVSG_API virtual nvmath::Mat44f getViewToWorldMatrix() const;

      //! Interface for getting the window offset.
      /** \returns The window offset  */
      NVSG_API virtual const nvmath::Vec2f& getWindowOffset() const;

      //! Interface for setting the window offset.
      NVSG_API virtual void setWindowOffset( const nvmath::Vec2f &offset    //!<  window ofset to viewing reference point
                                           );

      //! Interface for getting the relative window region to view
      NVSG_API virtual void getWindowRegion( nvmath::Vec2f &lowerLeft     //!< relative lower left position
                                           , nvmath::Vec2f &upperRight    //!< relative upper right position
                                           ) const;

      //! Interface for setting the relative window region to view
      NVSG_API virtual void setWindowRegion( const nvmath::Vec2f &lowerLeft     //!< relative lower left position
                                           , const nvmath::Vec2f &upperRight    //!< relative upper right position
                                           );

      //! Interface for getting the window size.
      /** \returns The window size  */
      NVSG_API virtual const nvmath::Vec2f& getWindowSize() const;

      //! Interface for setting the window size.
      NVSG_API virtual void setWindowSize( const nvmath::Vec2f &size    //!<  window size (world coordinates)
                                         );

      //! Interface for getting the current aspect ratio.
      /** \returns The width verus height of the camera window */
      NVSG_API virtual float getAspectRatio() const;

      //! Interface for setting the aspect ratio.
      NVSG_API virtual void setAspectRatio( float ar                //!<  new aspect ratio
                                          , bool keepWidth = false  //!<  keep width or height constant
                                          );

      //! Interface for determining if auto clipping is on.
      /** If auto clip is on, the camera determines the optimal near 
        * and front clipping distances out of the scene. If
        * the user has additional information about the scene that might 
        * help clipping, it should be turned off.
        * \returns \c true if auto clip is on, otherwise c\ false */
      NVSG_API virtual bool isAutoClipPlanes() const;

      //! Interface for enabling/disabling automatic clip plane determination.
      NVSG_API virtual void setAutoClipPlanes( bool on    //!< \c true to enable, \c false to disable
                                             );

      //! Interface for calculating the near and far clipping distances.
      NVSG_API virtual void calcNearFarDistances( const nvmath::Sphere3f &sphere  //!<  bounding sphere of the world to look at
                                                );

      //! Interface for getting the far clipping distance.
      /** \returns The distance to the far clipping plane */
      NVSG_API virtual float getFarDistance() const;

      //! Interface for setting the far clipping distance.
      NVSG_API virtual void setFarDistance( float fd    //!< far clipping distance
                                          );

      //! Interface for getting the near clipping distance.
      /** \returns The distance to the near clipping plane  */
      NVSG_API virtual float getNearDistance() const;

      //! Interface for setting the near clipping distance.
      NVSG_API virtual void setNearDistance( float nd   //!<  near clipping distance
                                           );

      //
      // operations to perform on a camera
      //

      //! Interface for moving the camera.
      NVSG_API virtual void move( const nvmath::Vec3f& delta    //!<  delta to move the camera (camera coordinates)
                                );

      //! Interface for orbiting the camera.
      NVSG_API virtual void orbit( const nvmath::Vec3f& axis    //!<  axis to orbit around (camera coordinates)
                                 , float rad                    //!<  angle to orbit (radians)
                                 );

      //! Interface for rotating the camera.
      NVSG_API virtual void rotate( const nvmath::Vec3f& axis     //!<  axis to rotate around 
                                  , float rad                     //!<  angle to rotate (radians)
                                  , bool  cameraRelative = true   //!<  flag to indicate if the axis is camera or world relative
                                  );

      //! Interface for zooming by a factor.
      NVSG_API virtual void zoom( float factor    //!< zoom in/out by this factor; changes windowSize
                                );

      //! Interface for zooming to fit view to a sphere.
      NVSG_API virtual void zoom( const nvmath::Sphere3f &sphere  //!<  sphere to zoom on
                                , float fovy = -1.0f              //!<  field of view; if negative, keep the current
                                );

 
    protected:
      //! Constructor
      /** Constructs a JitterCamera with the SimpleCamera \a base as it's base. The size of \a jitters gives the numbers
        * of passes to make, and the Vec2f elements determine the amount to jitter in each pass. The units of the jitters
        * are pixels.
        */
      NVSG_API JitterCamera( const SimpleCamera& base       //!<  base camera to use
                           , const std::vector<nvmath::Vec2f> &jitters   //!<  vector of jitters
                           );

      //! Partial Constructur
      NVSG_API JitterCamera( const MonoCamera &rhs
                           , const SimpleCamera &base
                           , const std::vector<nvmath::Vec2f> &jitters );

      //! Copy constructor
      NVSG_API JitterCamera( const JitterCamera& rhs //!< source camera object
                           );

      //! Protected destructor to prevent explicit creation on stack.
      NVSG_API virtual ~JitterCamera(void);

    private:
      JitterCamera();   //  private default constructor to prevent usage
      JitterCamera& operator=(const JitterCamera& rhs) {} // assignment not permitted

    private:
      const SimpleCamera                                * m_baseCamera;
      mutable std::vector< std::pair<const SimpleCamera*,nvmath::Vec2f> >   m_cameras;
  };

  // - - - - - - - - - - - - - - - - - - - - -
  // inlines
  // - - - - - - - - - - - - - - - - - - - - -
  inline const SimpleCamera * JitterCamera::getBaseCamera() const
  {
    __TRACE();
    return( m_baseCamera );
  }

  inline const SimpleCamera * JitterCamera::getCamera(size_t index) const
  {
    __TRACE();
    __ASSERT(index<m_cameras.size());
    return m_cameras[index].first;
  }

  inline void JitterCamera::getJitters( std::vector<nvmath::Vec2f> &jitters ) const
  {
    __TRACE();
    jitters.resize( m_cameras.size() );
    for ( size_t i=0 ; i<jitters.size() ; i++ )
    {
      jitters[i] = m_cameras[i].second;
    }
  }

  inline size_t JitterCamera::getNumberOfPasses() const
  {
    __TRACE();
    return( m_cameras.size() );
  }
} // namespace
