// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include "nvsgcommon.h"

#include "nvsg/Group.h"


namespace nvsg
{
  //! Level Of Detail class
  /** This class is a special kind of a group containing different 
    * representations of the same object as well as some range information. 
    * The representations include different levels of detail, meaning 
    * that the first child is the best and most precise representation 
    * of the object, the second child is of lower resolution, and so on.
    * The current implementation determines the distance between the 
    * camera and the given center point of the object (dist). The distance 
    * is compared against the ranges and the corresponding range defines 
    * the LOD representation to use.
    *
    *
    * It works like this:
    * \code
    * IF (dist < ranges[0])                    THEN activeChild = 0
    * ELSE IF (ranges[i-1] < dist < ranges[i]) THEN activeChild = i
    * ELSE IF (dist > ranges[LASTRANGE])       THEN activeChild = LASTRANGE+1
    * \endcode
    *
    * \note You should provide <b>n</b> ranges and <b>n+1</b> children. 
    * If you specify too few ranges, the extra children will never 
    * be used. If you instead provide to few children, the last child will be 
    * used for all the extra ranges.
    *
    * \note In some cases you do not want to draw something on the last (infinite)
    * level. In this case simply add an empty Group node at the lowest resolution level.
    * Same for the highest resolution level if you want geometry to disappear if it 
    * is too near to the camera.*/
  class LOD : public Group
  {
    public:
      //! Create a LOD.
      /** \returns a constant pointer to a LOD. */
      NVSG_API static const LOD * create( void );

      //! Create a LOD by copying from a Group.
      /** \returns a constant pointer to a LOD. */
      NVSG_API static const LOD * createFromBase( const Group &rhs  //!< Group to copy from
                                                );

      //! Get a clone of this \c LOD object.
      /** \returns A pointer to the new created clone. */
      NVSG_API virtual const LOD * clone( void ) const;

      //! Set all range values.
      NVSG_API void setRanges( const float * ranges   //!< Pass NULL to clear the ranges.
                             , size_t count           //!< Number of range values.
                             );

      //! Get all ranges.
      /** \return A pointer to a float array. */
      NVSG_API const float * getRanges( void ) const;

      //! Get the number of ranges.
      /** \return The number of ranges. */
      NVSG_API size_t getNumberOfRanges( void ) const;
      
      //! Set the center
      /** The center is the point used for finding  
        * the exact level of detail by calculating 
        * its distance to the camera. */
      NVSG_API void setCenter( const nvmath::Vec3f & center   //!< Center point.
                             );

      //! Get the senter point
      /** \return The point that is used for the distance calculation. */
      NVSG_API const nvmath::Vec3f & getCenter( void ) const;

      //!< Determine the LOD to use.
      /** This method is called by the ModelViewTraverser. It calculates
        * the distance between the center point and the camera.
        *  \par We calculate the distance by transforming the center into 
        * object space - while the camera is located at the origin in 
        * object space the calculation trivial.
        * \return The index of the LOD level that this node will use. 
        * 0xffffffff will be returned when there is no LOD level available.*/
      NVSG_API size_t getLODToUse(const nvmath::Mat44f & modelViewMatrix   //!< currentWorldMatrix * currentViewingMatrix
                                 ) const;

    protected:
      //! Default Constructor
      NVSG_API LOD( void );

      //! Partial Constructor
      NVSG_API LOD( const Group &rhs );

      //! Copy Constructor
      NVSG_API LOD( const LOD &rhs );
      
      //! Prohibit explicit creation on stack by making the destructor protected.
      NVSG_API virtual ~LOD( void );

    private:
      nvmath::Vec3f        m_center;
      std::vector<float>   m_ranges;
  };

}
