// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include "nvsgcommon.h"

#include "nvsg/Animation.h"
#include "nvsg/Node.h"
#include "nvmath/Quatf.h"
#include "nvmath/Trafo.h"

namespace nvsg
{
  //! Base class for representing all light sources.
  /** A LightSource contains data and functionality common to all light sources 
    * (DirectedLight, SpotLight, ...).\n   
    * \note NVSG LightSources do not have a method to turn them on or off. 
    * The concept is slightly different to other scene graph APIs. Since we have a 
    * Switch node that allows to turn multiple Nodes on / off light sources should be handeled via this switch.
    *
    * \note Light sources only influence objects under the same Switch/Group or in a subtree under the same Switch/Group.\n
    * Here a short example:\n
    * \code
                        |
     -----------------------------------------
     |                  |                    |
     Ls0               Sw                    NB
                        |
            -----------------------
            |      |       |      |
            Ls1    NA      Ls2    Gr
                                  |
                                 ...
    *  \endcode
    * Light source Ls0 lits Switch Sw and in turn all active children of Sw and of course Node NB.\n
    * Light source Ls1 lits Node NA and Group Gr and in turn all children of Gr.\n
    * And at least light source Ls2 lits Group Gr and in turn all children of Gr.\n
    */

  class LightSource : public Node
  {
    public:
      //! Get the ambient color of this light source.
      /** The ambient color of a light simply contributes to the global ambient term. It is not dependent on any other
        * lighting factors such as vertex normals, light direction, light position, range, or attenuation. The default is
        * (0.0,0.0,0.0).
        * \returns  The ambient color of this light source */
      NVSG_API const nvmath::Vec3f& getAmbientColor( void ) const;

      //! Set the ambient color of this light source.
      /** The ambient color of a light simply contributes to the global ambient term. It is not dependent on any other
        * lighting factors such as vertex normals, light direction, light position, range, or attenuation. The default is
        * (0.0,0.0,0.0). */
      NVSG_API void setAmbientColor( const nvmath::Vec3f &c   //!<  ambient color for this light source
                                   );

      //! Get the diffuse color of this light source.
      /** The diffuse color of a light is used to calculate lighting contribution that might depend on vertex normals,
        * light direction, light position, range, or attenutation, depending on the type of light source. The default is
        * (1.0,1.0,1.0).
        * \returns  The diffuse color of this light source */
      NVSG_API const nvmath::Vec3f& getDiffuseColor( void ) const;

            //! Set the diffuse color of this light source.
      /** The diffuse color of a light is used to calculate lighting contribution from vertex normals,
        * light direction, light position, range, or attenutation, depending on the type of light source. The default is
        * (1.0,1.0,1.0). */
      NVSG_API void setDiffuseColor( const nvmath::Vec3f &c   //!<  diffuse color for this light source
                                   );

      //! Get the specular color of this light source.
      /** The specular color of a light is used to calculate any lighting contribution from vertex normals,
        * vertex position, light direction, light position, range, or attenuation, depending on the type of light source.
        * The default is (1.0,1.0,1.0).
        * \returns  The specular color of this light source */
      NVSG_API const nvmath::Vec3f& getSpecularColor( void ) const;

      //! Set the specular color of this light source.
      /** The specular color of a light is used to calculate lighting contribution from vertex normals,
        * vertex position, light direction, light position, range, or attenuation, depending on the type of light source.
        * The default is (1.0,1.0,1.0). */
      NVSG_API void setSpecularColor( const nvmath::Vec3f &c    //!<  specular color for this light source
                                    );

      //! Get the intensity of this light source.
      /** The intensity of a light is a multiplier for all three color types of this light.
        * \returns  color intensity of this light source */
      NVSG_API float getIntensity( void ) const;

      //! Set the intensity of this light source.
      /** The intensity of a light is a multiplier for all three color types of this light. */
      NVSG_API void setIntensity( float i   //!<  color intensity for this light source
                                );

      //! Get the animation of this light.
      /** \returns  The constant pointer to the animation of this light, or NULL if there isn't any.  */
      NVSG_API const Animation<nvmath::Trafo> * getAnimation( void ) const;

      //! Set an animation for this light.
      NVSG_API void setAnimation( const Animation<nvmath::Trafo> * pAnimation   //!<  animation to add
                                );

      //! Set the animation frame.
      /** The transformation for animation step \a frame is calculated and used as the current. */
      NVSG_API void setAnimationFrame( size_t frame   //!<  current animation step
                                     );

      //! Get the number of animation frames that are part of this LightSource.
      /** \returns The number of animation frames if there is an animation attached, otherwise zero.  */
      NVSG_API virtual size_t getNumberOfFrames( void ) const;

      //! Set the position of the light source.
      NVSG_API void setPosition( const nvmath::Vec3f & position    //!< The position of the light source.
                               );

      //! Set the position of the light source.
      /** \returns The position of the light source. */ 
      NVSG_API const nvmath::Vec3f & getPosition( void ) const;

      //! Get the direction.
      /** The direction is the main direction of the light's cone - Default direction: (0, 0, -1).
        * \returns  The vector pointing into the direction of the light source.*/
      NVSG_API nvmath::Vec3f getDirection( void ) const;

      //! Get the direction.
      /** The direction is the main direction of the light's cone - Default direction: (0, 0, -1)
        * \note The direction is expected to be normalized. */
      NVSG_API void setDirection( const nvmath::Vec3f & direction    //!<  lighting direction
                                );

      //! Get the current rotational transformation.
      /** \returns  The quaternion that describes the rotational part */
      NVSG_API const nvmath::Quatf & getOrientation( void ) const;

      //! Set the rotational part of the light transformation.
      NVSG_API void setOrientation( const nvmath::Quatf & orientation  //!<  rotational part of transformation  
                                  );

      //! Get the current transformation.
      /** \returns  The transformation as a Mat44f  */
      NVSG_API nvmath::Mat44f getTransformationMatrix( void ) const;

      //! Get the inverse of the current transformation.
      /** \returns The inverse transformation as a Mat44f  */
      NVSG_API nvmath::Mat44f getInverse( void ) const;

      //! Determine whether the data of this object is shared.
      /** \returns  \c true if the object's data is shared, otherwise \c false.  */
      NVSG_API virtual bool isDataShared( void )  const;

      //! Get the DataID of the data of this object.
      /** \returns  DataID of the object's data.  */
      NVSG_API virtual DataID getDataID( void )     const;
      
      //! Ask if this LightSource is a LightSource.
      /** \returns  \c true  */
      NVSG_API virtual bool containsLight( void ) const;

    protected:
      //! Constructor
      NVSG_API LightSource( void );

      //! Copy constructor
      NVSG_API LightSource( const LightSource& rhs );

      //! Protected destructor to prevent explicit creation on stack.
      NVSG_API virtual ~LightSource(void);

      //! Calculate the bounding sphere of this light source.
      /** LightSource doesn't have a valid bounding sphere.
        * \returns \c false. */
      NVSG_API virtual bool calcBoundingSphere( void ) const;

    private:
      nvmath::Trafo m_trafo;
  
      float m_intensity;
      nvmath::Vec3f m_ambientColor;
      nvmath::Vec3f m_diffuseColor;
      nvmath::Vec3f m_specularColor;

      const Animation<nvmath::Trafo>  * m_animation; // animation working on a Trafo
      size_t              m_currentFrame;
  };

  inline bool LightSource::calcBoundingSphere( void ) const
  {
    __TRACE();
    return( false );
  }

  inline const nvmath::Vec3f& LightSource::getAmbientColor( void ) const
  {
    __TRACE();
    return( m_ambientColor );
  }

  inline const nvmath::Vec3f& LightSource::getDiffuseColor( void ) const
  {
    __TRACE();
    return( m_diffuseColor );
  }

  inline const nvmath::Vec3f& LightSource::getSpecularColor( void ) const
  {
    __TRACE();
    return( m_specularColor );
  }

  inline float LightSource::getIntensity( void ) const
  {
    __TRACE();
    return( m_intensity );
  }

  inline void LightSource::setAmbientColor( const nvmath::Vec3f &c )
  {
    __TRACE();
    m_ambientColor = c;
  }

  inline void LightSource::setDiffuseColor( const nvmath::Vec3f &c )
  {
    __TRACE();
    m_diffuseColor = c;
  }

  inline void LightSource::setSpecularColor( const nvmath::Vec3f &c )
  {
    __TRACE();
    m_specularColor = c;
  }

  inline void LightSource::setIntensity( float i )
  {
    __TRACE();
    m_intensity = i;
  }

  inline  const nvmath::Vec3f & LightSource::getPosition( void ) const
  {
    __TRACE();
    return m_trafo.getTranslation();
  }

  inline  void  LightSource::setPosition( const nvmath::Vec3f & position )
  {
    __TRACE();
    m_trafo.setTranslation(position);
    invalidateBoundingSphere();
  }

  inline nvmath::Vec3f LightSource::getDirection( void ) const
  {
    __TRACE();
    return m_trafo.getOrientation() * nvmath::Vec3f (0.f, 0.f, -1.f);
  }
  
  inline const nvmath::Quatf & LightSource::getOrientation( void ) const
  {
    __TRACE();
    return m_trafo.getOrientation();
  }

  inline void LightSource::setOrientation( const nvmath::Quatf & orientation )
  {
    __TRACE();
    m_trafo.setOrientation(orientation);
    invalidateBoundingSphere();
  }

  inline  nvmath::Mat44f LightSource::getTransformationMatrix( void ) const
  {
    __TRACE();
    return m_trafo.getMatrix();
  }

  inline nvmath::Mat44f LightSource::getInverse( void ) const
  {
    __TRACE();
    return m_trafo.getInverse();
  }

  inline const Animation<nvmath::Trafo> * LightSource::getAnimation( void ) const
  {
    __TRACE();
    return( m_animation );
  }

  inline size_t LightSource::getNumberOfFrames( void ) const
  {
    __TRACE();
    return( m_animation ? m_animation->getNumberOfFrames() : 0 );
  }
} //  namespace nvsg

