// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** @file */

#include "nvsgcommon.h"

#include  "nvsg/InterpolatedAnimation.h"

#ifdef _DEBUG
# ifndef new
#  define new new(__FILE__, __LINE__)
#  define _DEFINED_DBGNEW // restrict the 'new' macro to this header file only
# endif
#endif

namespace nvsg
{
  //! An \c InterpolatedAnimation that performs linear interpolations between the key frames.
  /** \note Currently the only supported types for LinearInterpolatedAnimation are \c Trafo and \c VNVector.  */
  template<typename T> class LinearInterpolatedAnimation : public InterpolatedAnimation<T>
  {
    public:
      //! Create a LinearInterpolatedAnimation.
      /** \returns a constant pointer to a LinearInterpolatedAnimation. */
      static const LinearInterpolatedAnimation<T> * create( void );

      //! Create a LinearInterpolatedAnimation by copying from an InterpolatedAnimation.
      /** \returns a constant pointer to a LinearInterpolatedAnimation. */
      static const LinearInterpolatedAnimation<T> * createFromBase( const InterpolatedAnimation<T> &rhs   //!< InterpolatedAnimation to copy from
                                                                  );

      //! Create a LinearInterpolatedAnimation by copying from this.
      /** \returns a constant pointer to a LinearInterpolatedAnimation. */
      virtual const LinearInterpolatedAnimation<T> * clone( void ) const;

      //! Interpolating random access operator
      /** If the access frame is not a key frame, the value is linearly interpolated between the neighbouring
        * key frames. */
      virtual const T & operator[]( size_t i ) const;

    protected:
      //! Default Constructor
      NVSG_API LinearInterpolatedAnimation( void );

      //! Partial Constructor
      NVSG_API LinearInterpolatedAnimation( const InterpolatedAnimation<T> &rhs );

      //! Copy Constructor
      NVSG_API LinearInterpolatedAnimation( const LinearInterpolatedAnimation &rhs );

      //! Protected destructor to prevent explicit creation on stack.
      virtual ~LinearInterpolatedAnimation( void );
  };

  template<typename T> 
  inline LinearInterpolatedAnimation<T>::~LinearInterpolatedAnimation(void)
  {
    __TRACE();
  }

  template<typename T>
  inline const LinearInterpolatedAnimation<T> * LinearInterpolatedAnimation<T>::create( void )
  {
    __TRACE();
    return( new LinearInterpolatedAnimation<T> );
  }

  template<typename T>
  inline const LinearInterpolatedAnimation<T> * LinearInterpolatedAnimation<T>::createFromBase( const InterpolatedAnimation<T> &rhs )
  {
    __TRACE();
    return( new LinearInterpolatedAnimation<T>( rhs ) );
  }

  template<typename T> 
  inline const LinearInterpolatedAnimation<T> * LinearInterpolatedAnimation<T>::clone( void ) const
  {
    __TRACE();
    return( new LinearInterpolatedAnimation<T>( *this ) );
  }

  template<typename T> 
  inline const T & LinearInterpolatedAnimation<T>::operator[]( size_t frame ) const
  {
    __TRACE();
    if ( frame != getCurrentFrame() )
    {
#if defined( _DEBUG )
      __ASSERT( getFrame(0) == 0 );
      for ( size_t i=1 ; i<getNumberOfKeys() ; i++ )
      {
        __ASSERT( getFrame(i-1) < getFrame(i) );
      }
#endif
      size_t key = getKey( (unsigned int) frame );
      if ( getFrame( key ) == frame ) //  frame is on a key => use this key frame
      {
        setCurrent( key );
      }
      else    //  frame is between it-1 and it
      {
        __ASSERT( key != 0 );
        setCurrent( (unsigned int) frame, lerp( (float)( frame - getFrame(key-1) ) / ( getFrame(key) - getFrame(key-1) ),
                                                getValue( key-1 ), getValue( key ) ) );
      }
    }
    return( getCurrentValue() );
  }
}

#ifdef _DEFINED_DBGNEW
# undef new
#endif
