// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include "nvsgcommon.h"

#include "nvmath/Vec3f.h"
#include "nvsg/StateAttribute.h"


namespace nvsg
{
  class Material;

//  ignore private member class on documentation
# if ! defined( DOXYGEN_IGNORE )
  class MaterialData : public nvutil::RCObject
  {
    friend class Material;

    public:
      //! Copy constructor
      /** Constructs a MaterialData object from another MaterialData object.
        * Public, because reference counting requires the copy constructor accessible. */
      NVSG_API MaterialData(const MaterialData& rhs);

    private:
      // instantiation on stack is not permitted
      MaterialData();
      ~MaterialData();
      // assignment also not permitted
      MaterialData& operator=(const MaterialData& rhs) {}

      nvmath::Vec3f m_ambientColor;
      nvmath::Vec3f m_diffuseColor;
      nvmath::Vec3f m_emissiveColor;
      float m_opacity;
      nvmath::Vec3f m_specularColor;
      float m_specularExponent;
  };
#endif    //  DOXYGEN_IGNORE

  //! StateAttribute that determines the material properties to use for the following geometries.
  /** \b Default \b Values:                   \n
    * ambient color     ( 0.3f, 0.3f, 0.3f )  \n
    * diffuse color     ( 0.9f, 0.9f, 0.9f )  \n
    * emissive color    ( 0.0f, 0.0f, 0.0f )  \n
    * opacity           ( 1.0f )              \n
    * specular color    ( 0.9f, 0.9f, 0.9f )  \n
    * specular exponent ( 10.0f ) */
  class Material : public StateAttribute
  {
    public:
      //! Create a Material.
      /** \returns a constant pointer to a Material. */
      NVSG_API static const Material * create( void );

      //! Create a Material by copying from a StateAttribute.
      /** \returns a constant pointer to a Material. */
      NVSG_API static const Material * createFromBase( const StateAttribute &rhs  //!< StateAttribute to copy from
                                                     );

      //! Get a clone of this material.
      /** \returns A pointer to the new created clone. */
      NVSG_API virtual const Material * clone( void ) const;

      //! Determine whether the data of this object is shared.
      /** \returns  \c true if the object's data is shared, otherwise \c false.  */
      NVSG_API virtual bool isDataShared( void ) const;

      //! Get the DataID of the data of this object.
      /** \returns  DataID of the object's data.  */
      NVSG_API virtual DataID getDataID( void ) const;

      //! Get the ambient color of this material.
      /** Ambient reflection, like ambient light, is nondirectional. Ambient reflection has a lesser impact on the
        * apparent color of a rendered object, but it does affect the overall color and is most noticeable when little or
        * no diffuse light reflects off the material. A material's ambient reflection is affected by the ambient light
        * set for a scene. The default is (0.3,0.3,0.3).
        * \returns  The ambient color of this material */
      NVSG_API const nvmath::Vec3f& getAmbientColor( void ) const;

      //! Get the diffuse color of this material.
      /** The diffuse and ambient colors of a material describe how a material reflects the ambient and diffuse light in
        * a scene. Because most scenes contain much more diffuse light than ambient light, diffuse reflection plays the
        * largest role in determining color. Additionally, because diffuse light is directional, the angle of incidence
        * for diffuse light affects the overall intensity of the reflection. Diffuse reflection is greatest when the
        * light strikes a vertex parallel to the vertex normal. As the angle increases, the effect of diffuse reflection
        * diminishes. The amount of light reflected is the cosine of the angle between the incoming light and the vertex
        * normal. The default is (0.9,0.9,0.9).
        * \returns  The diffuse color of this material */
      NVSG_API const nvmath::Vec3f& getDiffuseColor( void ) const;

      //! Get the emissive color of this material.
      /** Materials can be used to make a rendered object appear to be self-luminous. The emissive color of a material is
        * used to describe the color of the emitted light. Emission affects an object's color and can, for example, make
        * a dark material brighter and take on part of the emitted color. You can use a material's emissive color to add
        * the illusion that an object is emitting light, without incurring the computational overhead of adding a light
        * to the scene. Remember, materials with emissive color don't emit light that can be reflected by other objects
        * in a scene. To achieve this reflected light, you need to place an additional light within the scene. The
        * default is (0.0,0.0,0.0).
        * \returns  The emissive color of this material */
      NVSG_API const nvmath::Vec3f& getEmissiveColor( void ) const;

      //! Get the opacity of this material.
      /** The opacity of a material controls how much light is reflected and how much light goes through. An opacity of
        * 1.0 means the material is completely opaque - all light is reflected and no objects behind shine through. An
        * opacity of 0.0 means the material is completely transparent - no light is reflected and objects behind are seen
        * undisturbed. The default is 1.0.
        * \returns  The opacity of this material */
      NVSG_API float getOpacity( void ) const;

      //! Get the specular color of this material.
      /** Specular reflection creates highlights on objects, making them appear shiny. The material contains two members
        * that describe the specular highlight color as well as the material's overall shininess. You establish the color
        * of the specular highlights by setting the specular color to the desired RGBA color  the most common colors are
        * white or light gray. The values you set in the specular exponent control how sharp the specular effects are.
        * The default of the specular color is (0.9,0.9,0.9).
        * \returns  The specular color of this material */
      NVSG_API const nvmath::Vec3f& getSpecularColor( void ) const;

      //! Get the specular exponent of this material.
      /** Specular reflection creates highlights on objects, making them appear shiny. The material contains two members
        * that describe the specular highlight color as well as the material's overall shininess. You establish the color
        * of the specular highlights by setting the specular color to the desired RGBA color  the most common colors are
        * white or light gray. The values you set in the specular exponent control how sharp the specular effects are.
        * The default of the specular exponent is 10.0.
        * \returns  The specular exponent of this material */
      NVSG_API float getSpecularExponent( void ) const;

      //! Set the ambient color of this material.
      /** Ambient reflection, like ambient light, is nondirectional. Ambient reflection has a lesser impact on the
        * apparent color of a rendered object, but it does affect the overall color and is most noticeable when little or
        * no diffuse light reflects off the material. A material's ambient reflection is affected by the ambient light
        * set for a scene. The default is (0.3,0.3,0.3). */
      NVSG_API void setAmbientColor( const nvmath::Vec3f &ac );

      //! Set the diffuse color of this material.
      /** The diffuse and ambient colors of a material describe how a material reflects the ambient and diffuse light in
        * a scene. Because most scenes contain much more diffuse light than ambient light, diffuse reflection plays the
        * largest part in determining color. Additionally, because diffuse light is directional, the angle of incidence
        * for diffuse light affects the overall intensity of the reflection. Diffuse reflection is greatest when the
        * light strikes a vertex parallel to the vertex normal. As the angle increases, the effect of diffuse reflection
        * diminishes. The amount of light reflected is the cosine of the angle between the incoming light and the vertex
        * normal. The default is (0.9,0.9,0.9). */
      NVSG_API void setDiffuseColor( const nvmath::Vec3f &dc );

      //! Set the emissive color of this material.
      /** Materials can be used to make a rendered object appear to be self-luminous. The emissive color of a material is
        * used to describe the color of the emitted light. Emission affects an object's color and can, for example, make
        * a dark material brighter and take on part of the emitted color. You can use a material's emissive color to add
        * the illusion that an object is emitting light, without incurring the computational overhead of adding a light
        * to the scene. Remember, materials with emissive color don't emit light that can be reflected by other objects
        * in a scene. To achieve this reflected light, you need to place an additional light within the scene. The
        * default is (0.0,0.0,0.0). */
      NVSG_API void setEmissiveColor( const nvmath::Vec3f &ec );

      //! Set the opacity of this material.
      /** The opacity of a material controls how much light is reflected and how much light goes through. An opacity of
        * 1.0 means the material is completely opaque - all light is reflected and no objects behind shine through. An
        * opacity of 0.0 means the material is completely transparent - no light is reflected and objects behind are seen
        * undisturbed. The default is 1.0.
        * \returns  The opacity of this material */
      NVSG_API void setOpacity( const float o );

      //! Set the specular color of this material.
      /** Specular reflection creates highlights on objects, making them appear shiny. The material contains two members
        * that describe the specular highlight color as well as the material's overall shininess. You establish the color
        * of the specular highlights by setting the specular color to the desired RGBA color  the most common colors are
        * white or light gray. The values you set in the specular exponent control how sharp the specular effects are.
        * The default of the specular color is (0.9,0.9,0.9).
        * \returns  The specular color of this material */
      NVSG_API void setSpecularColor( const nvmath::Vec3f &sc );

      //! Get the specular exponent of this material.
      /** Specular reflection creates highlights on objects, making them appear shiny. The material contains two members
        * that describe the specular highlight color as well as the material's overall shininess. You establish the color
        * of the specular highlights by setting the specular color to the desired RGBA color  the most common colors are
        * white or light gray. The values you set in the specular exponent control how sharp the specular effects are.
        * The default of the specular exponent is 10.0.
        * \returns  The specular exponent of this material */
      NVSG_API void setSpecularExponent( const float se );

      //! Ask if this Material is transparent
      /** \returns  \c true, if it is transparent, otherwise \c false */
      NVSG_API virtual bool containsTransparentMaterial( void ) const;

    protected:
      //! Constructor
      NVSG_API Material( void );

      //! Partial Constructor
      NVSG_API Material( const StateAttribute &rhs );

      //! Copy Constructor
      NVSG_API Material( const Material &rhs );

      //! Protected destructor to prevent explicit creation on stack.
      NVSG_API virtual ~Material();

    private:
      nvutil::RCPtr<MaterialData> m_material; // shareable material data
  };

  inline const nvmath::Vec3f& Material::getAmbientColor( void ) const
  {
    __TRACE();
    return( m_material->m_ambientColor );
  }

  inline const nvmath::Vec3f& Material::getDiffuseColor( void ) const
  {
    __TRACE();
    return( m_material->m_diffuseColor );
  }

  inline const nvmath::Vec3f& Material::getEmissiveColor( void ) const
  {
    __TRACE();
    return( m_material->m_emissiveColor );
  }

  inline float Material::getOpacity( void ) const
  {
    __TRACE();
    return( m_material->m_opacity );
  }

  inline const nvmath::Vec3f& Material::getSpecularColor( void ) const
  {
    __TRACE();
    return( m_material->m_specularColor );
  }

  inline float Material::getSpecularExponent( void ) const
  {
    __TRACE();
    return( m_material->m_specularExponent );
  }

  inline void Material::setAmbientColor( const nvmath::Vec3f &ac )
  {
    __TRACE();
    nvutil::writeAccess( m_material );   //  nvutil::writeAccess performs a deep copy if necessary
    m_material->m_ambientColor = ac;
  }

  inline void Material::setDiffuseColor( const nvmath::Vec3f &dc )
  {
    __TRACE();
    nvutil::writeAccess( m_material );   //  nvutil::writeAccess performs a deep copy if necessary
    m_material->m_diffuseColor = dc;
  }

  inline void Material::setEmissiveColor( const nvmath::Vec3f &ec )
  {
    __TRACE();
    nvutil::writeAccess( m_material );   //  nvutil::writeAccess performs a deep copy if necessary
    m_material->m_emissiveColor = ec;
  }

  inline void Material::setOpacity( const float o )
  {
    __TRACE();
    __ASSERT( ( 0.0f <= o ) && ( o <= 1.0f ) );
    nvutil::writeAccess( m_material );   //  nvutil::writeAccess performs a deep copy if necessary
    if (    ( ( m_material->m_opacity == 1.0f ) && ( o < 1.0f ) )
        ||  ( ( m_material->m_opacity < 1.0f ) && ( o == 1.0f ) ) )
    {
      invalidateTransparentMaterialContainment();
    }
    m_material->m_opacity = o;
  }

  inline void Material::setSpecularColor( const nvmath::Vec3f &sc )
  {
    __TRACE();
    nvutil::writeAccess( m_material );   //  nvutil::writeAccess performs a deep copy if necessary
    m_material->m_specularColor = sc;
  }

  inline void Material::setSpecularExponent( const float se )
  {
    __TRACE();
    nvutil::writeAccess( m_material );   //  nvutil::writeAccess performs a deep copy if neccessary
    m_material->m_specularExponent = se;
  }

} //  namespace nvsg

