// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** @file */

#include "nvsgcommon.h"

// NBF is independent of packing. However, this forces
// the compile-time asserts below to fire on inconsistencies
#pragma pack(push, 1)

// convenient aliases for used built-in types
typedef char                    byte_t;   //!< Specifies an 8-bit signed type.
typedef unsigned char           ubyte_t;  //!< Specifies an 8-bit unsigned type.
typedef int                     int_t;    //!< Specifies a 32-bit signed integer type.
typedef unsigned int            uint_t;   //!< Specifies a 32-bit unsigned integer type.
typedef unsigned long long      uint64_t; //!< Specifies a 64-bit unsigned integer type.
typedef float                   float2_t[2]; //!< Specifies a 2-component float vector.
typedef float                   float3_t[3]; //!< Specifies a 3-component float vector.
typedef float                   float4_t[4]; //!< Specifies a 4-component float vector.

// convenient macro to add padding bits
#define PADDING(n) PADDING_i(n,__LINE__) 
#define PADDING_i(n,l) PADDING_ii(n,l)
#define PADDING_ii(n,l) ubyte_t padding##l[n] 

// NBF version
const ubyte_t NBF_VER_MAJOR  =  0x04; //!< NBF major version number
const ubyte_t NBF_VER_MINOR  =  0x01; //!< NBF version compatibility level
const ubyte_t NBF_VER_BUGFIX =  0x00; //!< NBF version bugfix level
  
// constants specifying a certain byte order
const ubyte_t NBF_LITTLE_ENDIAN = 0x00; //!< Specifies little endian byte order
const ubyte_t NBF_BIG_ENDIAN =  0x01;   //!< Specifies big endian byte order

// convenient helper types

//! The str_t structure specifies how a string is stored in a .NBF file.
struct str_t 
{ 
  uint_t      numChars;           //!< Specifies the number of characters in the actual string,
                                  //!< not including the terminating null character.
  uint_t      chars;              //!< Specifies the file offset to the string characters.
};
__CTASSERT(sizeof(str_t)%4==0);

//! The texCoordSet_t structure specifies how a texture coordinate set is stored in a .NBF file.
/** Texture coordinate sets, in this context, need to be considered in conjunction with NBFGeoSet objects. */
struct texCoordSet_t 
{
  uint_t      numTexCoords;       //!< Specifies the number of texture coordinates contained in the actual set.
  uint_t      coordDim;           //!< Specifies the dimension, in terms of float, of the contained texture coordinates.
                                  //!< Texture coordinates can be either one, two, three, or four dimensional.
  uint_t      texCoords;          //!< Specifies the file offset to the contained texture coordinates. 
};
__CTASSERT(sizeof(texCoordSet_t)%4==0);

//! The geometrySet_t structure specifies how a geometry set is stored in a .NBF file.
/** Geometry sets, in this context, need to be considered in conjunction with NBFGeoNode objects. */
struct geometrySet_t
{
  uint_t      drawable;           //!< Specifies the file offset to the actual NBFDrawable object. 
  uint_t      stateSet;           //!< Specifies the file offset to the corresponding NBFStateSet object.
};
__CTASSERT(sizeof(geometrySet_t)%4==0);

//! The indexSet_t structure specifies how an index set is stored in a .NBF file.
/** Index sets, in this context, need to be considered in conjunction with NBFStrips objects. */
struct indexSet_t
{
  uint_t      numIndices;         //!< Specifies the number of indices in the actual index set.
  uint_t      indices;            //!< Specifies the file offset to the indices. As specified by the NBF format,
                                  //!< a index is a 32-bit unsigned integer value.
};
__CTASSERT(sizeof(indexSet_t)%4==0);

//! The texBinding_t structure specifies how a texture binding is stored in a .NBF file.
/** Texture bindings, in this context, need to be considered in conjunction with 
  * NBFTectureAttribute and NBFTextureAttributeItem objects. */
struct texBinding_t
{
  uint_t      texUnit;            //!< Specifies the texture unit where the actual NBFTextureAttributeItem object is bound to.
  uint_t      texAttribItem;      //!< Specifies the file offset to the corresponding NBFTextureAttributeItem object.
};
__CTASSERT(sizeof(texBinding_t)%4==0);

//! The trafo_t structure specifies how a transfrmation is stored in a .NBF file.
struct trafo_t
{
  float4_t    orientation;        //!< Specifies the orientational part of the transformation.
  float3_t    scaling;            //!< Specifies the scaling part of the transformation.
  float3_t    translation;        //!< Specifies the translational part of the transformation.
  float3_t    center;             //!< Specifies the center of rotation of the transformation.
};
__CTASSERT(sizeof(trafo_t)%4==0);

//! The tweak_t structure specifies how a tweakable parameter is stored in a .NBF file.
/** Tweakables need to be considered in conjunction with NBFCgFx objects. */
struct tweak_t
{
  str_t       name;               //!< Specifies the tweakable's name as it appears in the .fx source file.
  uint_t      type;               //!< Specifies the tweakable's value type enum as specified in the CgFX format.
  union
  {
    uint_t      value;            //!< Specifies the file offset to a non-string value. This offset is not a valid offset
                                  //!< for string type enums. The value's size results from its type enum.
    str_t       strval;           //!< Specifies the string value in cases of string type enums. This is invalid for non-string
                                  //!< type enums.
  }; //!< Shared memory for non-string type enums and string type enums.
};
__CTASSERT(sizeof(tweak_t)%4==0);

//! The bone_t structure specifies how a bone is stored in a .NBF file. 
/** Bones need to be considered in conjunction with the skin_t structure. */
struct bone_t
{
  float       weight;             //!< Specifies a weight factor to be applied to the initial vertices and normals of the
                                  //!< actual skin.
  uint_t      animation;          //!< Specifies the file offset to the animation to be applied to the initial vertices 
                                  //!< and normals of the actual skin.
};
__CTASSERT(sizeof(bone_t)%4==0);

//! The skin_t structure specifies how a skin is stored in a .NBF file.
/** Skins need to be considered in conjunction with NBFSkinnedTriangles objects. */
struct skin_t
{
  uint_t      numBones;           //!< Specifies the number of bones describing the actual skin.
  uint_t      bones;              //!< Specifies the file offset to the bones describing the actual skin.
                                  //!< The way bones are stored in a .NBF file is specified by the bone_t structure. 
  float3_t    vertex;             //!< Specifies the initial vertex for the actual skin.
  float3_t    normal;             //!< Specifies the initial normal for the actual skin.
};
__CTASSERT(sizeof(skin_t)%4==0);

//! Unique NBF Object Codes
/** Each concrete NBFObject type is assigned to a unique NBF object code. 
  * This code is a 32-bit unsigned integer value, stored at offset 0, of each concrete NBFObject.
  * The purpose of the unique 'per-object' code is to provide a Load-Time Type Information (LTTI) 
  * to resolve concrete NBFObjects while loading NBF files. */   
enum
{
  // scene object 
  NBF_UNKNOWN                     = 0x00000000  //!< Unknown, and hence, invalid object code.
  // animation objects (0x100 - 0x1FF)
, NBF_TRAFO_ANIMATION           = 0x00000100
, NBF_VNVECTOR_ANIMATION
  // framed animations
, NBF_FRAMED_ANIMATION          = 0x00000120
, NBF_FRAMED_TRAFO_ANIMATION                  //!< Identifies an NBFFramedTrafoAnimation object.
, NBF_FRAMED_VNVECTOR_ANIMATION               //!< Identifies an NBFFramedVNVectorAnimation object.
  // interpolated animations 
, NBF_INTERPOLATED_ANIMATION    = 0x00000140
, NBF_LINEAR_INTERPOLATED_TRAFO_ANIMATION     //!< Identifies an NBFInterpolatedTrafoAnimation object.
, NBF_LINEAR_INTERPOLATED_VNVECTOR_ANIMATION  //!< Identifies an NBFInterpolatedVNVectorAnimation object.
  // camera objects (0x200 - 0x2FF)
, NBF_CAMERA                    = 0x00000200
, NBF_MONO_CAMERA
, NBF_JITTER_CAMERA
, NBF_SIMPLE_CAMERA
, NBF_PARALLEL_CAMERA                         //!< Identifies an NBFParallelCamera object.
, NBF_PERSPECTIVE_CAMERA                      //!< Identifies an NBFPerspectiveCamera object.
, NBF_STEREO_CAMERA  
  // drawable objects (0x300 - 0x3FF) 
, NBF_DRAWABLE                  = 0x00000300
, NBF_GEO_SET
, NBF_TRIANGLES                               //!< Identifies an NBFTriangles object.
, NBF_ANIMATED_TRIANGLES                      //!< Identifies an NBFAnimatedTriangles object.
, NBF_SKINNED_TRIANGLES                       //!< Identifies an NBFSkinnedTriangles object.
, NBF_TRISTRIPS                               //!< Identifies an NBFStrips object.
, NBF_QUADS                                   //!< Identifies an NBFQuads object (support since v0.3).
, NBF_ANIMATED_QUADS                          //!< Identifies an NBFAnimatedQuads object (support since v0.4)
, NBF_QUADSTRIPS                              //!< Identifies an NBFStrips object (support since v0.4)
  // node objects (0x400 - 0x4FF)
, NBF_NODE                      = 0x00000400
, NBF_GEO_NODE                                //!< Identifies a NBFGeoNode object.
, NBF_GROUP                                   //!< Identifies a NBFGroup object.
, NBF_LOD                                     //!< Identifies a NBFLOD object.
, NBF_SWITCH                                  //!< Identifies a NBFSwitch object.
, NBF_TRANSFORM                               //!< Identifies a NBFTransform object.
, NBF_ANIMATED_TRANSFORM                      //!< Identifies a NBFAnimatedTransform object.
, NBF_LIGHT_SOURCE
, NBF_DIRECTED_LIGHT                          //!< Identifies an NBFDirectedLight object.
, NBF_POINT_LIGHT                             //!< Identifies an NBFPointLight object.
, NBF_SPOT_LIGHT                              //!< Identifies an NBFSpotLight object.
  // state set objects (0x500 - 0x5FF)
, NBF_STATE_SET                 = 0x00000500  //!< Identifies an NBFStateSet object.
  // state attribute objects (0x600 - 0x6FF)
, NBF_STATE_ATTRIBUTE           = 0x00000600
, NBF_CGFX                                    //!< Identifies an NBFCgFx object.
, NBF_MATERIAL                                //!< Identifies an NBFMaterial object.
, NBF_FACE_ATTRIBUTE                          //!< Identifies an NBFFaceAttribute object.
, NBF_TEXTURE_ATTRIBUTE                       //!< Identifies an NBFTextureAttribute object.
, NBF_TEXTURE_ATTRIBUTE_ITEM                  //!< Identifies an NBFTextureAttributeItem object.
  // custom objects (>=0x700)
, NBF_CUSTOM_OBJECT                           //!< Custom objects must not have codes lower than this.
};

//! The NBFHeader structure represents the NBF header format.
/** The NBFHeader structure is the primary location where NBF specifics are stored.\n
  * For a valid NBF file, the NBFHeader structure is stored at file offset 0. Note that,
  * except for the NBFHeader object, a file offset of 0 indicates an invalid file offset!\n
  * This structure mainly serves as validation and compatibility checks for verification 
  * purposes. It also maintains the file offset to the contained NBFScene object, which  
  * represents a scene in the context of computer graphics. */ 
struct NBFHeader
{
  // signature
  byte_t      signature[4];       //!< A 4-byte signature identifying the file as a valid NBF file. The bytes are "#NBF".
  // NBF version
  ubyte_t     nbfMajorVersion;    //!< Specifies the major part of the NBF version used to save the file.
  ubyte_t     nbfMinorVersion;    //!< Specifies the minor part (compatibility level) of the NBF version used to save the file.
  ubyte_t     nbfBugfixLevel;     //!< Specifies the bugfix level of the NBF version used to save the file. This is optional
                                  //!< information, as a bugfix level does not influence compatibility issues, and hence 
                                  //!< must not be taken into account for compatibility checks.
  // NVSG version
  ubyte_t     nvsgMajorVersion;   //!< Specifies the major part of the NVSG version the content of this file is compatible to.
  ubyte_t     nvsgMinorVersion;   //!< Specifies the minor part of the NVSG version the content of this file is compatible to.
  ubyte_t     nvsgBugfixLevel;    //!< Specifies the bugfix level of the NVSG version. This is optional information, as a 
                                  //!< bugfix level does not influence compatibility issues, and hence must not be taken 
                                  //!< into account for compatibility checks.
  // Reserved bytes 
  ubyte_t     reserved[16];       //!< Reserved bytes for future extensions.
  // Date
  ubyte_t     dayLastModified;    //!< Specifies the day (1-31) of last modification.
  ubyte_t     monthLastModified;  //!< Specifies the month (1-12) of last modification.
  ubyte_t     yearLastModified[2]; //!< Specifies the year of last modification.
  // Time stamp
  ubyte_t     secondLastModified; //!< Specifies the second (0-59) of last modification. 
  ubyte_t     minuteLastModified; //!< Specifies the minute (0-59) of last modification.
  ubyte_t     hourLastModified;   //!< Specifies the hour (0-23) of last modification.
  // endianess
  ubyte_t     byteOrder;          //!< Specifies the byte order used to save the contained data.
                                  //!< A value of 0 implies little-endian byte order, a value of 1 implies big-endian byte order.
                                  //!< It is more convenient to use the symbolic constants NBF_LITTLE_ENDIAN and NBF_BIG_ENDIAN here.
  PADDING(2);                     //!< Padding bits to ensure offset of scene is on a 4-byte boundary, regardless of packing.
  // scene object
  uint_t      scene;              //!< Specifies the file offset to the contained NBFScene object.
  // optional view state
  uint_t      viewState;          //!< Specifies the file offset to an optional NBFViewState object. 
                                  //!< An offset of 0 indicates that no NBFViewState object is available in this file. 
};
__CTASSERT(sizeof(NBFHeader)%4==0);

//! The NBFScene structure represents a scene in the context of computer graphics.
/** A valid NBF file always contains one - and only one - NBFScene object. 
  * The file offset to this NBFScene object is specified within the NBFHeader structure. */
struct NBFScene
{
  float3_t    backColor;                //!< Specifies the scene's RGB background color used with rendering.
  uint_t      numCameras;               //!< Specifies the number of the scene's NBFCamera objects.
  uint_t      cameras;                  //!< Specifies the file offset to the offsets of the scene's NBFCamera objects.
  uint_t      numCameraAnimations;      //!< Specifies the number of the scene's NBFTrafoAnimation objects (camera animations). 
  uint_t      cameraAnimations;         //!< Specifies the offset to the offsets of the scene's NBFTrafoAnimation objects (camera animations).
  uint_t      defaultStateSet;          //!< Specifies the offset to the scene's optional default states. Default states are of type NBFStateSet.
  uint_t      overrideStateSet;         //!< Specifies the offset to the scene's optional override states. Override states are of type NBFStateSet.
  uint_t      numberOfAnimationFrames;  //!< For animated scenes, this specifies the number of animation frames.
  uint_t      root;                     //!< Specifies the file offset to the scene's root node, which always is of a NBFNode derived type.
};
__CTASSERT(sizeof(NBFScene)%4==0);

//! The NBFViewState represents an optional view state used to render the scene.
/** The file offset to an optional NBFViewState object is specified within the NBFHeader structure. */
struct NBFViewState
{
  uint_t      objectCode;         //!< Specifies the object code of the actual object. The object code is unique per object type! 
  uint_t      camera;             //!< Specifies the offset to the NBFCamera object to be used for viewing.
};
__CTASSERT(sizeof(NBFViewState)%4==0);

//! The NBFObject structure represents general object data. 
struct NBFObject
{
  uint_t      objectCode;         //!< Specifies the object code of the actual object. The object code is unique per object type! 
  ubyte_t     isShared;           //!< Indicates whether the data of the actual object is shared among different objects.
                                  //!< A value of 1 indicates that this object's data is shared, whereas a value of 0 indicates
                                  //!< that this object's data is not shared.
  PADDING(3);                     //!< Padding bits to ensure offset of objectDataID is on a 8-byte boundary, regardless of packing.
  uint64_t    objectDataID;       //!< A unique 64-bit value to identify shared object data while loading.
  uint_t      sourceObject;       //!< Specifies the file offset to the source object in case of data sharing. 
                                  //!< A file offset of 0 always indicates that no source object is available for the actual object.
  str_t       objectName;         //!< Specifies an optional name for the actual object.
  // NOTE: Because of the uint64_t member objectDataID above, which is 8-byte aligned, 
  // we need to ensure the size of NBFObject is fixed - that is, independent of whatever
  // the compilers actual packing value might be! We achieve this by making the size of 
  // NBFObject a multiple of 8 bytes (see compile time assert below).  
  PADDING(4);                     //!< Padding bits to ensure the size of NBFObject is a multiple of 8, regardless of packing.    
};
__CTASSERT(sizeof(NBFObject)%8==0);

//! The NBFTrafoAnimation represents an animation that will be applied to trafo_t objects.
/** A NBFTrafoAnimation serves as base class only and needs to be considered in conjunction
  * with either NBFFramedAnimation or NBFInterpolatedAnimation.\n
  * Valid object codes for a NBFTrafoAnimation are NBF_FRAMED_TRAFO_ANIMATION, 
  * and NBF_LINEAR_INTERPOLATED_TRAFO_ANIMATION. Further concrete object codes valid for
  * NBFTrafoAnimation objects are subject to future extensions of the NBF format. */
struct NBFTrafoAnimation : public NBFObject
{
  uint_t      trafos;             //!< Specifies the file offset to the trafo_t objects the animation will be applied to.
                                  //!< The actual number of trafo_t objects is specified by the corresponding animation
                                  //!< type, which is either NBFFramedAnimation or NBFInterpolatedAnimation.
  PADDING(4);        //!< Padding bits to ensure the size of NBFTrafoAnimation is a multiple of 8, regardless of packing.    
};
__CTASSERT(sizeof(NBFTrafoAnimation)%8==0);

//! The NBFVNVectorAnimation represents an animation that will be applied to VNVector objects.
/** A NBFVNVectorAnimation serves as base class only and needs to be considered in conjunction 
  * with either NBFFramedAnimation or NBFInterpolatedAnimation.\n
  * Concrete object codes valid for a NBFVNVectorAnimation are NBF_FRAMED_VNVECTOR_ANIMATION, 
  * and NBF_LINEAR_INTERPOLATED_VNVECTOR_ANIMATION. Further concrete object codes valid for
  * NBFVNVectorAnimation objects are subject to future extensions of the NBF format. */
struct NBFVNVectorAnimation : public NBFObject
{
  uint_t      numVertices;        //!< Specifies the number of vertices per vertex set. Each vertex set of 
                                  //!< the actual animation is specified to have the same count of vertices.
  uint_t      vertexSets;         //!< Specifies the file offset to the pre-transformed vertex sets for this animation. 
                                  //!< The vertex sets are stored adjoined to each other in contiguous memory.\n
                                  //!< The actual number of vertex sets, as well as normal sets, is specified by
                                  //!< the corresponding animation type, which is either NBFFramedAnimation or
                                  //!< NBFInterpolatedAnimation.
  uint_t      numNormals;         //!< Specifies the number of normals per normal set. Each normal set of 
                                  //!< the actual animation is specified to have the same count of normals.
  uint_t      normalSets;         //!< Specifies the file offset to the pre-transformed normal sets for this animation.
                                  //!< The normal sets are stored adjoined to each other in contiguous memory.\n
                                  //!< The actual number of normal sets, as well as vertex sets, is specified by
                                  //!< the corresponding animation type, which is either NBFFramedAnimation or
                                  //!< NBFInterpolatedAnimation.
                                  
};
__CTASSERT(sizeof(NBFVNVectorAnimation)%8==0);

//! The NBFFramedAnimation structure represents a framed animation.
/** A NBFFramedAnimation serves as base class only and needs to be considered in conjunction
  * with either NBFTrafoAnimation or NBFVNVectorAnimation.\n
  * Concrete object codes valid for a NBFFramedAnimation are NBF_FRAMED_TRAFO_ANIMATION and
  * NBF_FRAMED_VNVECTOR_ANIMATION. Further concrete object codes valid for NBFFramedAnimation 
  * objects are subject to future extensions of the NBF format. */
struct NBFFramedAnimation
{
  uint_t      numFrames;          //!< Specifies the number of animation frames.\n 
                                  //!< In conjunction with a NBFTrafoAnimation, this corresponds to the number of
                                  //!< trafo_t objects stored with a NBFTrafoAnimation.\n
                                  //!< In conjunction with a NBFVNVectorAnimation, this corresponds to the number of
                                  //!< vertex/normal sets stored with a NBFVNVectorAnimation.
};
__CTASSERT(sizeof(NBFFramedAnimation)%4==0);

//! The NBFFramedTrafoAnimation structure represents a framed animation that will be applied to trafo_t objects.
/** A NBFFramedTrafoAnimation is a concrete animation type. It publicly inherits from NBFTrafoAnimation and
  * NBFFramedAnimation. The object code for a NBFFramedTrafoAnimation is NBF_FRAMED_TRAFO_ANIMATION. */ 
struct NBFFramedTrafoAnimation : public NBFTrafoAnimation
                               , public NBFFramedAnimation
{
  PADDING(4);         //!< Padding bits to ensure the size of NBFFramedTrafoAnimation is a multiple of 8, regardless of packing.    
};
__CTASSERT(sizeof(NBFFramedTrafoAnimation)%8==0);

//! The NBFFramedVNVectorAnimation structure represents a framed animation that will be applied to VNVector objects.
/** A NBFFramedVNVectorAnimation is a concrete animation type. It publicly inherits from NBFVNVectorAnimation and
  * NBFFramedAnimation. The object code for a NBFFramedVNVectorAnimation is NBF_FRAMED_VNVECTOR_ANIMATION. */ 
struct NBFFramedVNVectorAnimation : public NBFVNVectorAnimation
                                  , public NBFFramedAnimation
{
  PADDING(4);         //!< Padding bits to ensure the size of NBFFramedVNVectorAnimation is a multiple of 8, regardless of packing.    
};
__CTASSERT(sizeof(NBFFramedVNVectorAnimation)%8==0);

//! The NBFInterpolatedAnimation structure represents a interpolated animation.
/** A NBFInterpolatedAnimation serves as base class only and needs to be considered in conjunction
  * with either NBFTrafoAnimation or NBFVNVectorAnimation.\n
  * Concrete object codes valid for a NBFInterpolatedAnimation are NBF_LINEAR_INTERPOLATED_TRAFO_ANIMATION and
  * NBF_LINEAR_INTERPOLATED_VNVECTOR_ANIMATION. Further concrete object codes valid for NBFInterpolatedAnimation 
  * objects are subject to future extensions of the NBF format. */
struct NBFInterpolatedAnimation
{
  uint_t      numKeys;            //!< Specifies the number of key frames
  uint_t      keys;               //!< Specifies the file offset to the key frames. As specified in the NBF format,
                                  //!< a key is a 32-bit unsigned integer value.
};
__CTASSERT(sizeof(NBFInterpolatedAnimation)%4==0);

//! The NBFInterpolatedTrafoAnimation structure represents a interpolated animation that will be applied to trafo_t objects.
/** A NBFInterpolatedTrafoAnimation is a concrete animation type. It publicly inherits from NBFTrafoAnimation and
  * NBFInterpolatedAnimation.\n 
  * The object code valid for a NBFInterpolatedTrafoAnimation is NBF_LINEAR_INTERPOLATED_TRAFO_ANIMATION. Further object codes
  * valid for NBFInterpolatedTrafoAnimation objects are subject to future extensions of the NBF format. */
struct NBFInterpolatedTrafoAnimation : public NBFTrafoAnimation
                                     , public NBFInterpolatedAnimation
{
};
__CTASSERT(sizeof(NBFInterpolatedTrafoAnimation)%8==0);

//! The NBFInterpolatedVNVectorAnimation structure represents a interpolated animation that will be applied to VNVector objects.
/** A NBFInterpolatedVNVectorAnimation is a concrete animation type. It publicly inherits from NBFVNVectorAnimation and
  * NBFInterpolatedAnimation.\n 
  * The object code valid for a NBFInterpolatedVNVectorAnimation is NBF_LINEAR_INTERPOLATED_VNVECTOR_ANIMATION. Further object codes
  * valid for NBFInterpolatedVNVectorAnimation objects are subject to future extensions of the NBF format. */
struct NBFInterpolatedVNVectorAnimation : public NBFVNVectorAnimation
                                        , public NBFInterpolatedAnimation
{
};
__CTASSERT(sizeof(NBFInterpolatedVNVectorAnimation)%8==0);

//! The NBFCamera structure represents a camera.
/** A NBFCamera serves as base class only. */
struct NBFCamera : public NBFObject
{
};
__CTASSERT(sizeof(NBFCamera)%8==0);

//! The NBFMonoCamera structure represents a monoscopic camera.
/** A NBFMonoCamera serves as base class only. */
struct NBFMonoCamera : public NBFCamera
{
};
__CTASSERT(sizeof(NBFMonoCamera)%8==0);

//! The NBFJitterCamera structure represents a monoscopic jitter camera.
/** \note NBFJitterCamera is subject to future extensions of the NBF format. */
struct NBFJitterCamera : public NBFMonoCamera
{
};
__CTASSERT(sizeof(NBFJitterCamera)%8==0);

//! The NBFSimpleCamera represents a simple monoscopic camera.
/** A NBFSimpleCamera serves as base class only.\n
  * Concrete object codes valid for a NBFSimpleCamera are NBF_PARALLEL_CAMERA
  * and NBF_PERSPECTIVE_CAMERA. Further object codes valid for a NBFSimpleCamera
  * are subject to future extensions for the NBF format. */
struct NBFSimpleCamera : public NBFMonoCamera
{
  uint_t      numHeadLights;    //!< Specifies the number of headlights attached.
  uint_t      headLights;       //!< Specifies the file offset to the offsets to the attached headlight objects.
                                //!< Headlights are of type NBFLightSource. 
  float3_t    upVector;         //!< Specifies the camera's normalized up vector.
  float3_t    position;         //!< Specifies the actual position of camera in world space.
  float3_t    direction;        //!< Specifies the normalized direction for the camera to look along.
  float       farDist;          //!< Specifies the distance from the actual camera position to the far clipping plane.
  float       nearDist;         //!< Specifies the distance from the actual camera position to the near clipping plane.
  float       targDist;         //!< Specifies the distance to the projection plane.
  float       projectionDx;     //!< Specifies the translation of projection (used with stereo for example).
  ubyte_t     isAutoClipPlanes; //!< Indicates if automatic generation of clipping planes is enabled.
  PADDING(3);      //!< Padding bits to ensure offset of windowSize is on a 4-byte boundary, regardless of packing
  float2_t    windowSize;       //!< Specifies the world-relative size of the viewing window. Whereas the x-component of 
                                //!< of the vector specifies the width, and the y-component of the vector specifies the height.
  float2_t    windowOffset;     //!< Specifies the world-relative offset from the viewing reference point to the center 
                                //!< of the viewing window.
};
__CTASSERT(sizeof(NBFSimpleCamera)%8==0);

//! The NBFParallelCamera represents a parallel camera.
/** A NBFParallelCamera is a concrete camera type. 
  * The object code for a NBFParallelCamera is NBF_PARALLEL_CAMERA. */
struct NBFParallelCamera : public NBFSimpleCamera
{
};
__CTASSERT(sizeof(NBFParallelCamera)%8==0);

//! The NBFPerspectiveCamera represents a perspective camera.
/** A NBFPerspectiveCamera is a concrete camera type. 
  * The object code for a NBFPerspectiveCamera is NBF_PERSPECTIVE_CAMERA. */
struct NBFPerspectiveCamera : public NBFSimpleCamera
{
};
__CTASSERT(sizeof(NBFPerspectiveCamera)%8==0);

//! The NBFStereoCamera represents a stereoscopic camera.
/** \note NBFStereoCamera is subject to future extensions of the NBF format. */
struct NBFStereoCamera : public NBFCamera
{
};
__CTASSERT(sizeof(NBFStereoCamera)%8==0);

//! The NBFDrawable represents any kind of a drawable object.
/** A NBFDrawable serves a base class only.\n
  * Concrete object codes valid for a NBFDrawable are NBF_TRIANGLES, NBF_ANIMATED_TRIANGLES, 
  * NBF_SKINNED_TRIANGLES, and NBF_TRISTRIPS. Further object codes valid for a NBFDrawable
  * are subject to future extensions of the NBF format. */
struct NBFDrawable : public NBFObject
{
};
__CTASSERT(sizeof(NBFDrawable)%8==0);

//! The NBFGeoSet structure represents a set of vertex attributes.
/** A NBFGeoSet serves as base class only. It maintains a full set of geometry
  * data (vertex attributes) that is used by all derived topology classes. */
struct NBFGeoSet : public NBFDrawable
{
  uint_t      numVertices;        //!< Specifies the number of contained vertices. 
  uint_t      vertices;           //!< Specifies the file offset to the vertices. Vertices are stored as float3_t.
  uint_t      numNormals;         //!< Specifies the number of contained normals.
  uint_t      normals;            //!< Specifies the file offset to the normals. Normals are stored as float3_t.
  uint_t      numTexCoordsSets;   //!< Specifies the number of contained texture coordinate sets.
  uint_t      texCoordsSets;      //!< Specifies the file offset to the texture coordinate sets. 
                                  //!< Texture coordinate sets are stored using the texCoordSet_t type.
  uint_t      numColors;          //!< Specifies the number of contained primary colors.
  uint_t      colorDim;           //!< Specifies the dimension, in terms of float, used for the primary colors.
                                  //!< Colors can be either three or four dimensional. 
  uint_t      colors;             //!< Specifies the file offset to the primary colors. In conformity to the color's dimension,
                                  //!< colors are stored as float3_t or float4_t respectively.
  uint_t      numSecondaryColors; //!< Specifies the number of contained secondary colors.
  uint_t      secondaryColorDim;  //!< Specifies the dimension, in terms of float, used for the secondary colors.
                                  //!< Colors can be either three or four dimensional. 
  uint_t      secondaryColors;    //!< Specifies the file offset to the secondary colors. In conformity to the color's dimension,
                                  //!< colors are stored as float3_t or float4_t respectively.
  uint_t      numFogCoords;       //!< Specifies the number of contained fog coordinates.
  uint_t      fogCoords;          //!< Specifies the file offset to the fog coordinates. Fog coordinates always are one 
                                  //!< dimensional, and are stored as 32-bit floating point values.
};
__CTASSERT(sizeof(NBFGeoSet)%8==0);

//! The NBFTriangles structure represents a set of independent triangles.
/** A NBFTriangles is a concrete topology class derived from NBFGeoSet. 
  * The object code for a NBFTriangles is NBF_TRIANGLES. */
struct NBFTriangles : public NBFGeoSet
{
  uint_t      numFaces;           //!< Specifies the number of contained triangles faces
  uint_t      faces;              //!< Specifies the file offset to the Face3 objects. As specified in the NBF format,
                                  //!< a face is a triple of 32-bit unsigned integer values.
};
__CTASSERT(sizeof(NBFTriangles)%8==0);

//! The NBFAnimatedTriangles structure represents animated triangles.
/** A NBFAnimatedTriangles is a concrete topology class derived from NBFTriangles. 
  * The object code for a NBFAnimatedTriangles is NBF_ANIMATED_TRIANGLES. */
struct NBFAnimatedTriangles : public NBFTriangles
{
  uint_t      animation;          //!< Specifies the file offset to the NBFVNVectorAnimation object 
                                  //!< to be applied to the triangle faces.
  PADDING(4);        //!< Padding bits to ensure the size of NBFAnimatedTriangles is a multiple of 8, regardless of packing.    
};
__CTASSERT(sizeof(NBFAnimatedTriangles)%8==0);

//! The NBFSkinnedTriangles structure represents skin animated triangles.
/** A NBFSkinnedTriangles is a concrete topology class derived from NBFTriangles. 
  * The object code for a NBFSkinnedTriangles is NBF_SKINNED_TRIANGLES. */
struct NBFSkinnedTriangles : public NBFTriangles
{
  uint_t      numSkins;           //!< Specifies the number of contained skins
  uint_t      skins;              //!< Specifies the file offset to the corresponding skin_t objects.
};
__CTASSERT(sizeof(NBFSkinnedTriangles)%8==0);

//! The NBFStrips structure represents a strip set.
/** A NBFStrips is a concrete topology class derived from NBFGeoSet. 
  * Valid object codes for a NBFStrips are NBF_TRISTRIPS and NBF_QUADSTRIPS. */
struct NBFStrips : public NBFGeoSet
{
  uint_t      numStrips;          //!< Specifies the number of strips
  uint_t      strips;             //!< Specifies the file offset to the strips. 
                                  //!< Strips are stored as indexSet_t objects.
};
__CTASSERT(sizeof(NBFStrips)%8==0);

//! The NBFQuads structure represents a set of independent quads.
/** A NBFQuads is a concrete topology class derived from NBFGeoSet. 
  * The object code for a NBFQuads is NBF_QUADS. */
struct NBFQuads : public NBFGeoSet
{
  uint_t      numFaces;           //!< Specifies the number of contained quad faces.
  uint_t      faces;              //!< Specifies the file offset to the quad faces. 
                                  //!< Quad faces are stored as Face4 objects.
};
__CTASSERT(sizeof(NBFQuads)%8==0);

//! The NBFAnimatedQuads structure represents animated quads.
/** A NBFAnimatedQuads is a concrete topology class derived from NBFQuads. 
  * The object code for a NBFAnimatedQuads is NBF_ANIMATED_QUADS. */
struct NBFAnimatedQuads : public NBFQuads
{
  uint_t      animation;          //!< Specifies the file offset to the NBFVNVectorAnimation object 
                                  //!< to be applied to the quad faces.
  PADDING(4);        //!< Padding bits to ensure the size of NBFAnimatedQuads is a multiple of 8, regardless of packing.    
};
__CTASSERT(sizeof(NBFAnimatedQuads)%8==0);

//! The NBFNode structure represents a general node.
/** A NBFNode serves as base class only. Concrete object codes valid for a NBFNode are
  * NBF_GEO_NODE, NBF_LOD, NBF_SWITCH, NBF_TRANSFORM, NBF_ANIMATED_TRANSFORM,
  * NBF_DIRECTED_LIGHT, NBF_POINT_LIGHT, and NBF_SPOT_LIGHT. Further concrete object
  * codes valid for a NBFNode are subject to future extensions of the NBF format. */
struct NBFNode : public NBFObject
{
  str_t       annotation;         //<! Specifies an optional annotation string.   
};
__CTASSERT(sizeof(NBFNode)%8==0);

//! The NBFGeoNode structure represents a geometry node.
/** The object code for a NBFGeoNode is NBF_GEO_NODE. */
struct NBFGeoNode : public NBFNode
{
  uint_t      numGeometrySets;    //<! Specifies the number of contained geometry sets.
  uint_t      geometrySets;       //<! Specifies the file offset to the geometry sets.
                                  //<! Geometry sets are stored as geometrySet_t objects.
};
__CTASSERT(sizeof(NBFGeoNode)%8==0);

//! The NBFGroup structure represents a group node.
/** A NBFGroup serves as base class only. Concrete object codes valid for a NBFGroup are
  * NBF_LOD, NBF_SWITCH, NBF_TRANSFORM, and NBF_ANIMATED_TRANSFORM. Further concrete object
  * codes valid for a NBFGroup are subject to future extensions of the NBF format. */
struct NBFGroup : public NBFNode
{
  uint_t      numChildren;        //!< Specifies the number of maintained children.
  uint_t      children;           //!< Specifies the file offset to the offsets to the maintained children.
                                  //!< NBFGroup's children always are of NBFNode-derived types.
};
__CTASSERT(sizeof(NBFGroup)%8==0);

//! The NBFLOD structure represents a 'Level Of Detail' group node.
/** The object code for a NBFLOD is NBF_LOD. */
struct NBFLOD : public NBFGroup
{
  float3_t    center;             //!< Specifies the center point used for distance calculations.
  uint_t      numRanges;          //!< Specifies the number of contained ranges.
  uint_t      ranges;             //!< Specifies the file offset to the ranges. 
                                  //!< Ranges are stored as 32-bit floating point numbers.
  PADDING(4);        //!< Padding bits to ensure the size of NBFLOD is a multiple of 8, regardless of packing.    
};
__CTASSERT(sizeof(NBFLOD)%8==0);

//! The NBFSwitch structure represents a switch group node.
/** The object code for a NBFSwitch is NBF_SWITCH. */
struct NBFSwitch : public NBFGroup
{
  uint_t      numActiveChildren;  //!< Specifies the number of children that are currently switched on.
  uint_t      activeChildren;     //!< Specifies the offset to indices to the active children.
                                  //!< Indices are stored as 32-bit unsigned integers.
};
__CTASSERT(sizeof(NBFSwitch)%8==0);

//! The NBFTransform structure represents a transform group node.
/** The object code for a NBFTransform is NBF_TRANSFORM. */
struct NBFTransform : public NBFGroup
{
  trafo_t             trafo;
  PADDING(4);        //!< Padding bits to ensure the size of NBFTransform is a multiple of 8, regardless of packing.    
};
__CTASSERT(sizeof(NBFTransform)%8==0);

//! The NBFAnimatedTransform structure represents an animated transform group node.
/** The object code for a NBFAnimatedTransform is NBF_ANIMATED_TRANSFORM. */
struct NBFAnimatedTransform : public NBFTransform
{
  uint_t      animation;          //!< Specifies the file offset to the NBFTrafoAnimation object
                                  //!< to be applied to the transform group node.
  PADDING(4);        //!< Padding bits to ensure the size of NBFAnimatedTransform is a multiple of 8, regardless of packing.    
};
__CTASSERT(sizeof(NBFAnimatedTransform)%8==0);

//! The NBFLightSource structure represents a light source node.
/** A NBFLightSource serves as base class only. Concrete object codes valid for 
  * a NBFLightSource are NBF_DIRECTED_LIGHT, NBF_POINT_LIGHT, and NBF_SPOT_LIGHT. */
struct NBFLightSource : public NBFNode
{
  float       intensity;          //!< Specifies the light intensity.
  float3_t    ambientColor;       //!< Specifies the ambient color term of the light.
  float3_t    diffuseColor;       //!< Specifies the diffuse color term of the light.
  float3_t    specularColor;      //!< Specifies the specular color term of the light.
  trafo_t     trafo;              //!< Specifies how to position the light in world space.
  uint_t      animation;          //!< Specifies the file offset to an optional NBFTrafoAnimation object
                                  //!< to be applied to the light transform (orientation and translation).
};
__CTASSERT(sizeof(NBFLightSource)%8==0);

// The NBFDirectedLight structure represents a directed light source.
/** The object code for a NBFDirectedLight is NBF_DIRECTED_LIGHT. */
struct NBFDirectedLight : public NBFLightSource
{
};
__CTASSERT(sizeof(NBFDirectedLight)%8==0);

// The NBFPointLight structure represents a point light source.
/** The object code for a NBFPointLight is NBF_POINT_LIGHT. */
struct NBFPointLight : public NBFLightSource
{
  float3_t    attenuation;        //!< Specifies the attenuation factors of the point light.
                                  //!< The x-component of the vector specifies the constant term of the attenuation,
                                  //!< the y-component of the vector specifies the linear term of the attenuation, and
                                  //!< the z-component of the vector specifies the quadratic term of the attenuation.
  PADDING(4);        //!< Padding bits to ensure the size of NBFPointLight is a multiple of 8, regardless of packing.    
};                        
__CTASSERT(sizeof(NBFPointLight)%8==0);

// The NBFSpotLight structure represents a spot light.
/** The object code for a NBFSpotLight is NBF_SPOT_LIGHT. */
struct NBFSpotLight : public NBFLightSource
{
  float3_t    attenuation;        //!< Specifies the attenuation factors of the spot light.
                                  //!< The x-component of the vector specifies the constant term of the attenuation,
                                  //!< the y-component of the vector specifies the linear term of the attenuation, and
                                  //!< the z-component of the vector specifies the quadratic term of the attenuation.
  float       cutoffAngle;        //!< Specifies the angle between the axis of the cone, the light is emitted to, and
                                  //!< a ray along the edge of the cone.
  float       falloffExponent;    //!< Controls the intensity distribution inside the cone, the light is mitted to.
  PADDING(4);                     //!< Padding bits to ensure the size of NBFSpotLight is a multiple of 8, regardless of packing.    
};
__CTASSERT(sizeof(NBFSpotLight)%8==0);

//! The NBFStateSet structure represents a set of heterogeneous NBFStateAttributes.
/** The object code for a NBFStateSet is NBF_STATE_SET. */
struct NBFStateSet : public NBFObject
{
  uint_t      numStateAttribs;    //!< Specifies the number of contained state attributes.
  uint_t      stateAttribs;       //!< Specifies the file offset to the offsets to the NBFStateAttriburte objects
};
__CTASSERT(sizeof(NBFStateSet)%8==0);

//! The NBFStateAttribute structure represents a state attribute.
/** A NBFStateAttribute serves a base class only. Concrete object codes valid for a NBFStateAttribute
  * are NBF_CGFX, NBF_MATERIAL, NBF_FACE_ATTRIBUTE, and NBF_TEXTURE_ATTRIBUTE. */
struct NBFStateAttribute : public NBFObject
{
};
__CTASSERT(sizeof(NBFStateAttribute)%8==0);

//! The NBFCgFx structure represents an effect attribute.
/** The object code for a NBFCgFx is NBF_CGFX. */
struct NBFCgFx : public NBFStateAttribute
{
  str_t       file;               //!< Specifies the name of the corresponding CgFX source file.
  uint_t      numTweaks;          //!< Specifies the number of tweakable parameters available for the effect.
  uint_t      tweaks;             //!< Specifies the file offset to the tweakables. 
                                  //!< Tweakables in the NBF format are stored as tweak_t objects.
  uint_t      currTech;           //!< Specifies the index of the current technique use with rendering.
  PADDING(4);                     //!< Padding bits to ensure the size of NBFCgFx is a multiple of 8, regardless of packing.    
};
__CTASSERT(sizeof(NBFCgFx)%8==0);

//! The NBFFaceAttribute structure represents a render mode.
/** The object code for a NBFFaceAttribute is NBF_FACE_ATTRIBUTE. */
struct NBFFaceAttribute : public NBFStateAttribute
{
  float3_t    ambientColor;         //!< Specifies the gloabal ambient color to be used for rendering.
  ubyte_t     backFaceCulling;      //!< Specifies if culling of back faces should be enabled for rendering.
  ubyte_t     polygonOffsetEnabled; //!< Specifies if polygon offset should be enabled for rendering.
  PADDING(2);                       //!< Padding bits to ensure offset of polygonOffsetFactor is on a 4-byte boundary, regardless of packing
  float       polygonOffsetFactor;  //!< Specifies the scale factor to be used to create a variable depth offset
                                    //!< for a polygon.
  float       polygonOffsetUnits;   //!< Specifies a unit that is multiplied by a render API implementation-specific
                                    //!< value to create a constant depth offset.
  ubyte_t     twoSidedLighting;     //!< Specifies if two-sided lighting should be enabled for rendering.
  PADDING(7);                       //!< Padding bits to ensure the size of NBFStateAttribute is a multiple of 4, regardless of packing.
};
__CTASSERT(sizeof(NBFFaceAttribute)%8==0);

//! The NBFMaterial structure represents a material.
/** The object code for a NBFMaterial is NBF_MATERIAL. */
struct NBFMaterial : public NBFStateAttribute
{
  float3_t    ambientColor;       //!< Specifies the ambient part of the material color. 
  float3_t    diffuseColor;       //!< Specifies the diffuse part of the material color.
  float3_t    emissiveColor;      //!< Specifies the emissive part of the material color.
  float       opacity;            //!< Specifies the opacity of the material.
  float3_t    specularColor;      //!< Specifies the specular part of the material color.
  float       specularExponent;   //!< Specifies the specular exponent of the material color.
};
__CTASSERT(sizeof(NBFMaterial)%8==0);

//! The NBFTextureAttribute structure represents a texture attribute.
/** The object code for a NBFTextureAttribute is NBF_TEXTURE_ATTRIBUTE. */
struct NBFTextureAttribute : public NBFStateAttribute
{
  uint_t      numBindings;        //!< Specifies the number of contained texture bindings.
  uint_t      bindings;           //!< Specifies the file offset to the texture bindings. Texture bindings in
                                  //!< the NBF format are stored as texBinding_t objects.
};
__CTASSERT(sizeof(NBFTextureAttribute)%8==0);

//! The NBFTextureAttributeItem structure represents a single texture object.
/** The object code for a NBFTextureAttributeItem is NBF_TEXTURE_ATTRIBUTE_ITEM. 
  * A NBFTextureAttributeItem needs to be consider in conjunction with a NBFTextureAttribute, 
  * which specifies the binding of texture objects. */
struct NBFTextureAttributeItem : public NBFObject
{
  str_t         file;               //!< Specifies the name of the texture image file.
  uint_t        texType;            //!< Specifies the texture type. Valid types are TT_AMBIENT, TT_BUMP, TT_DIFFUSE,
                                    //!< TT_DISPLACEMENT, TT_FILTER, TT_OPACITY, TT_REFLECTION, TT_REFRACTION, 
                                    //!< TT_SELF_ILLUM, TT_SHININESS, TT_SHINING_STRENGTH, and TT_SPECULAR. 
  uint_t        texEnvMode;         //!< Specifies the texture environment mode for the actual texture object. 
                                    //!< Valid modes are TEM_REPLACE, TEM_MODULATE, TEM_DECAL, TEM_BLEND, and TEM_ADD.
  uint_t        texWrapS;           //!< Specifies the wrap parameter for texture coordinate s. 
                                    //!< Valid values are TWM_REPEAT, TWM_CLAMP, and TWM_MIRROR
  uint_t        texWrapT;           //!< Specifies the wrap parameter for texture coordinate t.
                                    //!< Valid values are TWM_REPEAT, TWM_CLAMP, and TWM_MIRROR
  uint_t        texWrapR;           //!< Specifies the wrap parameter for texture coordinate r.
                                    //!< Valid values are TWM_REPEAT, TWM_CLAMP, and TWM_MIRROR
  uint_t        minFilter;          //!< Specifies the filter used with minimizing. 
                                    //!< Valid values are TFM_MIN_NEAREST, TFM_MIN_BILINEAR, and TFM_MIN_TRILINEAR.
  uint_t        magFilter;          //!< Specifies the filter used with magnifying.
                                    //!< Valid values are TFM_MAX_NEAREST, and TFM_MAZ_BILINEAR.
  float4_t      texBorderColor;     //!< Specifies the texture border RGBA color.
  trafo_t       trafo;              //!< Specifies the texture transformation
  uint_t        texGenMode[4];      //!< Specifies the texture coordinate generation modes
  float4_t      texGenPlane[4];     //!< Specifies the texture coordinate generation planes
};
__CTASSERT(sizeof(NBFTextureAttributeItem)%8==0);

#pragma pack(pop)
