// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include "nvsgcommon.h"

#include  "nvsg/Node.h"

namespace nvsg 
{
  //! Path representation in the tree.
  /** A Path represents a scene graph or subgraph. It contains
    * pointers to a chain of nodes, each being a child of the
    * previous. */
  class Path : public nvutil::RCObject
  {
    public:
      //! Default constructor.
      NVSG_API Path( void );

      //! Copy constructor
      NVSG_API Path( const Path& rhs );

      //! Get the the number of nodes in the path chain.
      /** \return Th number of nodes.*/
      NVSG_API unsigned int getLength( void ) const;

      //! Test on an empty path.
      /** \note This is possibly faster than testing the length.
        * \return \c true - path is empty, \c false - path is not empty */
      NVSG_API bool isEmpty( void ) const;  
      
      //! Get the head node of the path.
      /** \return The first node in the path chain.*/
      NVSG_API const Node * getHead( void ) const;

      //! Get the Tail Node
      /** \return The last node in the path chain. */
      NVSG_API const Node * getTail( void ) const;

      //! Get the ith node from the head.
      /** \return Pointer to ith node in the 
        * path chain. */
      NVSG_API const Node * getNodeFromHead( size_t i   //!< Index of the desired node (0 = head node)
                                           ) const;

      //! Get the ith node from the tail.
      /** \return Pointer to the ith node in the path 
        * chain (from the tail) */
      NVSG_API const Node * getNodeFromTail( size_t i   //!< Index of the desired node (0 = tail node)
                                           ) const;
      
      //! Pop the last node off.
      NVSG_API void pop( void );

      //! Push a node at the end of the chain.
      /** \note  We apply an addRef() so we do not lose the node. */
      NVSG_API void push( const Node * pNode  //!<  Node to push
                        );
           
      //! Remove all nodes from indexed node on.
      /** \note We apply a removeRef(). */
      NVSG_API void truncate( size_t start    //!< Start index. (0 removes all nodes from the path)
                            );

    protected:
      //! Prohibit explicit creation on stack by making the destructor protected.
      NVSG_API virtual ~Path(void);
      
    private:
      std::vector<const Node *> m_path;   //!< Vector of nodes representing a path chain.
  };

  // - - - - - - - - - - - - - - - - - - -
  // inlines
  // - - - - - - - - - - - - - - - - - - -

  /** getHead() returns NULL when the list is empty. */
  inline const Node * Path::getHead( void ) const
  {
    __TRACE();
    return( m_path.empty() ? NULL : m_path.front() );
  }

  /** getTail() returns NULL when the list is empty. */
  inline const Node * Path::getTail( void ) const
  {
    __TRACE();
    return( m_path.empty() ? NULL : m_path.back() );
  }

  /** getNodeFromHead() undefined for an invalid index. */
  inline const Node * Path::getNodeFromHead( size_t i ) const
  {
    __TRACE();
    __ASSERT(0 <= i && i < m_path.size());
    return( m_path[i] );
  }

  /** getNodeFromTail() undefined for an invalid index. */
  inline const Node * Path::getNodeFromTail( size_t i ) const
  {
    __TRACE();
    __ASSERT(0 <= i && i < m_path.size());
    return( m_path[m_path.size() - i-1] );
  }

  inline unsigned int Path::getLength( void ) const
  {
    __TRACE();
    return (unsigned int) m_path.size();
  }

  inline void Path::pop( void )
  {
    __TRACE();
    __ASSERT(getLength() > 0);
    m_path.back()->removeRef();
    m_path.pop_back();
  }

  inline bool Path::isEmpty( void ) const
  {
    __TRACE();
    return m_path.empty();
  }
} //  namspace nvsg
