// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** @file */

#include "nvsgcommon.h"

#include  "nvutil/PlugIn.h" // general PlugIn class definition

namespace nvsg
{
  class Scene;
  class Texture;
  class ViewState;

  //! Pure virtual base class for NVSG scene loader plug-ins
  /** User defined NVSG scene loader plug-ins must provide the \c SceneLoader
    * interface.
    * \par
    * Client code that needs to utilize a certain \c SceneLoader interface should
    * first query for a unique plug interface type ID through a call to \c nvutil::queryInterfaceType.
    * The unique plug interface type ID for a SceneLoader will be constructed from the \c UPITID_SCENE_SAVER
    * define and the actual NVSG version, which is coded in the UPITID_VERSION define, as shown in the code 
    * snippet below.
    * A call to \c nvutil::queryInterfaceType returns a list of all unique interface IDs (UPIIDSs)
    * found at given search paths.\n
    * After that, the client code can take the different UPIIDs to grab a scene loader interface 
    * through a call to \c nvutil::getInterface.\n
    * After usage, the client code should release the interface through a call to \c nvutil::releaseInterface:
    *
    * \code
    *   // Example:
    *   // --------
    *   // Get a scene loader interface capable to load 'nvb' files
    *   
    *   // 
    *   vector<string> searchPathes;
    *   // Add appropriate search paths here. Not relevant for the example here.
    *   // ...
    *
    *   // define the unique plug interface type ID for SceneLoaders
    *   const nvutil::UPITID PITID_SCENE_LOADER(UPITID_SCENE_LOADER, UPITID_VERSION);
    *
    *   nvsg::Scene * theScene = NULL;
    *   nvsg::ViewState * viewState = NULL;
    *   nvutil::UPIID nvbLoaderInterfaceID;
    *   bool foundAppropriate = false;
    *   vector<nvutil::UPIID> piids;
    *   if ( nvutil::queryInterfaceType(searchPathes, PITID_SCENE_LOADER, piids) )
    *   {
    *     vector<nvutil::UPIID>::iterator it = piids.begin();
    *     for ( ; it != piids.end(); ++it )
    *     {
    *       if ( !stricmp((*it).getPlugSpecificIDString(), ".nvb") )
    *       {
    *         // found, copy the ID
    *         nvbLoaderInterfaceID = *it;
    *         foundAppropriate = true;
    *         break; // look no further
    *       }
    *     }
    *     if ( foundAppropriate )
    *     {
    *       nvutil::PlugIn * plug; 
    *       if ( nvutil::getInterface(searchPathes, nvbLoaderInterfaceID, plug) )
    *       {
    *         nvsg::SceneLoader * loader = reinterpret_cast<SceneLoader*>(plug);
    *         theScene = loader->load("c:\\nvb\\sample.nvb", searchPathes, viewState);
    *         if ( theScene ) 
    *         { 
    *           theScene->addRef();
    *           if ( viewState )
    *           {
    *             viewState->addRef();
    *           }
    *         }
    *         nvutil::releaseInterface(nvbLoaderInterfaceID);
    *       }
    *     }
    *   }
    * \endcode
    */
  // SceneLoader interface
  class SceneLoader : public nvutil::PlugIn
  {
    public:
      //! Loading a scene
      /** Loads an NVSG scene from a file specified by \a filename. 
        * The function tries to look up this file as follows:
        * -# Looks at the specified location.
        * -# Looks at the current directory.
        * -# Uses the search paths stored in \a searchPaths to look up the file.
        * \returns A pointer to the loaded scene if successful, a null pointer otherwise.
        */
      NVSG_API virtual const Scene * load( const std::string& filename            //!< A string that holds the name of the scene file to be loaded.
                                         , const std::vector<std::string> &searchPaths //!< A collection of search pathes used to look up the file.
                                         , const ViewState *& viewState              /*!< If the function succeeded, this points to the optional
                                                                                          ViewState stored with the scene. */
                                         ) = 0;

    protected:
      //! Protected virtual destructor
      /** Prohibits ordinary client code from  
        * - creating a \c SceneLoader derived object on stack and
        * - calling \c delete on a pointer to \c SceneLoader. 
        * \note This will not prohibit the client code from calling \c delete
        * on a pointer to a \c SceneLoader derived object. Normally this 
        * is not that critical, because one almost always uses
        * a pointer to the base class when working with \c SceneLoader plug-ins.
        * \par
        * Nevertheless, it would be far less error prone if classes that
        * inherit from \c SceneLoader also make their destructor protected.
        */
      NVSG_API virtual ~SceneLoader();
  };

  //! Pure virtual base class for NVSG scene saver plug-ins
  /** User defined NVSG scene saver plug-ins must provide the \c SceneSaver
    * interface. */
  class SceneSaver : public nvutil::PlugIn
  {
    public:
      //! Saving a scene
      /** Saves an NVSG \a scene and a \a viewState to a file specified by \a filename. 
        * \returns \c true if successful, \c false otherwise.
        */
      NVSG_API virtual bool save( const Scene &scene, const ViewState *viewState, const std::string& filename ) = 0;

    protected:
      //! Protected virtual destructor
      /** Prohibits ordinary client code from  
        * - creating a \c SceneSaver derived object on stack and
        * - calling \c delete on a pointer to \c SceneSaver.
        * \note This will not prohibit the client code from calling \c delete
        * on a pointer to a \c SceneSaver derived object. Normally this 
        * is not that critical, because one almost always uses
        * a pointer to the base class when working with \c SceneSaver plug-ins.
        * \par
        * Nevertheless, it would be far less error prone if classes that
        * inherit from \c SceneSaver, also make their destructor protected.
        */
      NVSG_API virtual ~SceneSaver();
  };

#if 0 // not yet
# if ! defined( DOXYGEN_IGNORE )
  //! Pure virtual base class for texture loader plugins
  /** \note TextureLoader are not yet implemented as plugins. */
  class TextureLoader : public nvutil::PlugIn
  {
    public:
      //! Pure virtual interface function to load a Texture.
      /** Loads a texture file given by \a filename.
        * The function tries to look up the texture file as follows:
        * -# Looks at the specified location.
        * -# Looks at the current directory.
        * -# Uses the search pathes stored in \a searchPaths to look up the file.
        * \returns  A pointer to the loaded Texture if successful, a null pointer otherwise.
        */
      NVSG_API virtual Texture * load( const std::string& filename            //!<  texture file to load
                                     , const std::vector<std::string> &searchPaths //!<  paths to search through
                                     ) = 0;

    protected:
      //! Protected virtual destructor
      /** Prohibits ordinary client code from  
        * - creating a \c TextureLoader derived object on stack and
        * - calling \c delete on a pointer to \c TextureLoader.
        * \note This will not prohibit client code from calling \c delete
        * on a pointer to a \c TextureLoader derived object. Normally this 
        * is not that critical, because one almost always uses
        * a pointer to the base class when working with \c TextureLoader plug-ins.
        * \par
        * Nevertheless, it would be far less error prone if classes that
        * inherit from \c TextureLoader also make their destructor protected.
        */
      NVSG_API virtual ~TextureLoader();
  };
# endif
#endif

} //  namespace nvsg
