// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include "nvsgcommon.h"

#include "nvsg/nvsg.h"
#include "nvsg/GeoSet.h"

#ifdef _DEBUG
# ifndef new
#  define new new(__FILE__, __LINE__)
#  define _DEFINED_DBGNEW // restrict the 'new' macro to this header file only
# endif
#endif

namespace nvsg
{  
  //! GeoSet defining quad strips.
  /** A QuadStrips is a GeoSet that represents the vertex attributes of the GeoSet as quad strips. The strips are
    * defined by a number of IndexSet elements. Each IndexSet is a std::vector<unsigned int> and represents one quad strip.
    * If you have, for example, an array of eight vertices in the GeoSet, one possibility to make them a cube is, to define
    * three IndexSet elements like that:
    * \code
    *   set[0] = { 0, 1, 2, 3, 4, 5, 6, 7, 0, 1 };
    *   set[1] = { 1, 7, 3, 5 };
    *   set[2] = { 6, 0, 4, 2 };
    * \endcode
    * The ordering ensures that the triangles are all drawn with the same orientation so that the triangle strip correctly
    * form part of a surface. Preserving the orientation is important for some operations, such as culling. The number of
    * vertices per IndexSet must be at least 4 for anything to be drawn. */
  class QuadStrips : public GeoSet
  {
    public:
      //! Create a QuadStrips.
      /** \returns a constant pointer to a QuadStrips. */
      NVSG_API static const QuadStrips * create( void );

      //! Create a QuadStrips by copying from a GeoSet.
      /** \returns a constant pointer to a QuadStrips. */
      NVSG_API static const QuadStrips * createFromBase( const GeoSet &rhs );

      //! Get a clone of this QuadStrips object
      /** \returns a constant pointer to a QuadStrips. */
      NVSG_API virtual const QuadStrips * clone( void ) const;

      //! Determine whether the data of this object is shared.
      /** \returns  \c true if the object's data is shared, otherwise \c false.  */
      NVSG_API virtual bool isDataShared( void ) const;

      //! Get the DataID of the data of this object.
      /** \returns  DataID of the object's data.  */
      NVSG_API virtual DataID getDataID( void ) const;

      //! Ask if this QuadStrips has strips.
      /** \returns  \c true, if this QuadStrips has strips, otherwise \c false */
      NVSG_API bool hasStrips( void ) const;

      //! Get the number of strips.
      /** \returns  The number of strips */
      NVSG_API size_t getNumberOfStrips( void ) const;

      //! Get constant pointer to strips.
      /** \return A constant pointer to strips  */
      NVSG_API const IndexSet * getStrips() const;

      //! Set the strips for this QuadStrip.
      /** Copies \a numStrips strips from \a pStrips into this QuadStrips. */
      NVSG_API void setStrips( const IndexSet* pStrips    //!<  strips to set
                             , size_t numStrips           //!<  number of strips
                             );

      //! Set the strips for this QuadStrips.
      /** Copies \a numStrips strips from \a pStrips into this QuadStrips, starting at position \a pos
        * inside the range of pre-existing strips.
        *
        * Pre-existing strips in the range [\a pos, \a pos + \a numStrips) will be replaced.
        * Pre-existing strips outside this range remain untouched.
        *
        * If you specify -1 for \a pos or if \a pos specifies the number of strips currently stored,
        * the strips pointed to by \a pStrips will be appended to the pre-existing faces. 
        *
        * If you not specify -1 for \a pos or \a pos neither specifies the number of faces currently
        * stored nor does it specify a valid position inside the range of pre-existing strips, 
        * the behaviour is undefined! */
      NVSG_API void setStrips( size_t pos //!< Start position inside the range of pre-existing strips
                             , const IndexSet* pStrips  //!<  strips to set
                             , size_t numStrips            //!<  number of strips
                             );

      //! Calculate tangents and binormals.
      /** The two dimensional texture coordinates out of texture unit \a tu are used as the parameterization for the
        * calculation. The tangents are written to texture unit \a tg and the binormals to texture unit \a bn.
        */
      NVSG_API virtual void calculateTangentSpace( size_t tu = 0    //!<  texture unit to get parameterizing coords
                                                 , size_t tg = 1    //!<  texture unit to write tangents to
                                                 , size_t bn = 2    //!<  texture unit to write binormals to
                                                 );

    protected:
      //! Constructor
      NVSG_API QuadStrips( void );

      //! Partial Constructor from a GeoSet
      NVSG_API QuadStrips( const GeoSet &rhs );

      //! Copy Constructor
      NVSG_API QuadStrips( const QuadStrips &rhs );

      //! Protected destructor to prevent explicit creation on stack.
      NVSG_API virtual ~QuadStrips( void );

    private:
      void  calculateTangentBinormal( size_t tu, size_t i, size_t i1, size_t i2, nvmath::Vec3f &tangent, nvmath::Vec3f &binormal );

      nvutil::RCPtr<nvutil::RCVector<IndexSet> >  m_quadStrips; // quadstrip indices
  };

  // - - - - - - - - - - - - - - - - - - - - - - - - - 
  // inlines
  // - - - - - - - - - - - - - - - - - - - - - - - - - 

  inline bool QuadStrips::hasStrips( void ) const 
  {
    __TRACE();
    return( !m_quadStrips->empty() ); 
  }

  inline size_t QuadStrips::getNumberOfStrips( void ) const
  { 
    __TRACE();
    return( m_quadStrips->size() ); 
  }

  inline const IndexSet * QuadStrips::getStrips( void ) const 
  { 
    __TRACE();
    return( &(*m_quadStrips)[0] ); 
  }
}

#ifdef _DEFINED_DBGNEW
# undef new
#endif
