// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include "nvsgcommon.h"

#include "nvsg/Animation.h"
#include "nvsg/Camera.h"
#include "nvsg/Node.h"

namespace nvmath
{
  class Trafo;
}
namespace nvsg
{
  class StateSet;
}

namespace nvsg
{
  //! Shrink wrap class to subsume all scene related information.
  class Scene : public nvutil::RCObject
              , public nvutil::ISWMRSync
  {
    public:
      //! Create a Scene.
      /** \returns a constant pointer to a Scene. */
      NVSG_API static const Scene * create( void );
      
      //! Set the background color.
      NVSG_API void setBackColor( const nvmath::Vec3f &color    //!<  color to use as background color
                                );

      //! Get the background color.
      NVSG_API const nvmath::Vec3f & getBackColor( void )  const;

      //! Add a camera to this scene.
      /** A scene holds a list of cameras from which one can be selected for viewing. The reference count for the
        * camera is incremented. */
      NVSG_API void addCamera( const Camera * pCamera   //!<  camera to add
                             );

      //! Remove a camera from the scene.
      NVSG_API void removeCamera( size_t index //!< Index of the camera to remocve 
                                );

      //! Get a camera of the scene.
      /** \returns  A constant pointer to the camera at position \a index, or NULL if this doesn't exist */
      NVSG_API const Camera * getCamera( size_t index = 0   //!<  index of camera to get
                                 ) const;

      //! Get the number of cameras in the scene.
      /** A scene holds a list of cameras from which one can be selected for viewing.
        * \returns  The number of cameras in the scene */
      NVSG_API size_t getNumberOfCameras( void ) const;

      //! Get the number of camera animations in the scene.
      /** A scene holds a list of camera animations from which one can be selected (independently from the camera) 
        * for animated viewing.
        * \returns  The number of camera animations in the scene */
      NVSG_API size_t getNumberOfCameraAnimations( void ) const;

      //! Add a camera animation to this scene.
      /** A scene holds a list of camera animations from which one can be selected (independently from the camera) 
        * for animated viewing. The reference count for the animation is incremented. */
      NVSG_API void addCameraAnimation( const Animation<nvmath::Trafo> * pAnimation   //!<  animation to add
                                      );

      //! Get a camera animation of the scene.
      /** \returns  A constant pointer the animation at position \a index, or NULL if this doesn't exist */
      NVSG_API const Animation<nvmath::Trafo> * getCameraAnimation( size_t index = 0    //!<  index of animation to get
                                                                  ) const;

      //! Remove a camera animation from the scene.
      NVSG_API void removeCameraAnimation( size_t index //!< Index of the camera animation to remove
                                         );

      //! Get the number of animation frames.
      /** All animations in a scene must have the same number of steps. This number is queried here.
        * \returns  The number of animation frames (of every animation in the scene) */
      NVSG_API size_t getNumberOfFrames( void ) const;

      //! Set the root node.
      /** If there already is a root node, it is dereferenced; the reference count of the new root node is incremented. */
      NVSG_API void setRootNode( const Node * root    //!<  new root node
                               );

      //! Get the root node of the scene tree.
      /** \returns  A constant pointer to the root node of the scene tree, or NULL if there isn't any */
      NVSG_API const Node * getRootNode( void ) const;

      //! Set the default state set.
      /** If a geometry in this scene doesn't have a StateAttribute of a special kind, but there is one in this default
        * StateSet, the default one is used. If there is already a defaul StateSet, it is dereferenced;
        * the reference count of this StateSet is incremented. */
      NVSG_API void setDefaultStateSet( const StateSet * pStateSet   //!<  StateSet to set as the default one
                                      );

      //! Get the default state set of the scene.
      /** If a geometry in this scene doesn't have a StateAttribute of a special kind, but there is one in this default
        * StateSet, the default one (if there is one) is used. */
      NVSG_API const StateSet * getDefaultStateSet( void ) const;

      //! Set the override state set.
      /** If there is an override state set in the scene, all geometries are rendered using that state set, ignoring
        * any other state set in the scene. */
      NVSG_API void setOverrideStateSet( const StateSet * pStateSet   //!<  StateSet to set as the override one
                                       );

      //! Get the override state set of the scene.
      /** If there is an override state set in the scene, all geometries are rendered using that state set, ignoring
        * any other state set in the scene. */
      NVSG_API const StateSet * getOverrideStateSet( void ) const;

      //! Ask if this Scene contains any CgFx attribute.
      /** \returns  \c true, if the Scene contains a CgFx attribute in its scene tree, otherwise \c false.  */
      NVSG_API bool containsCgFx( void ) const;

      //! Ask if this Scene contains any LightSource.
      /** \returns  \c true, if the Scene contains a LightSource in its scene tree, otherwise \c false.  */
      NVSG_API bool containsLight( void ) const;

      //! Ask if this Scene contains any transparent Material attribute.
      /** \returns  \c true, if the Scene contains a transparent Material attribute in its scene tree, otherwise \c false. */
      NVSG_API bool containsTransparentMaterial( void ) const;

      //! Ask if this Scene contains any transparent Texture.
      /** \returns  \c true, if the Scene contains a transparent Texture in its scene tree, otherwise \c false. */
      NVSG_API bool containsTransparentTexture( void ) const;

      //! Acquire read access in a multithreaded environment.
      /** For safety in a multithreaded environment, an object maintains a read/write locking mechanism that allows 
        * contemporaneous access for multiple readers but write access for only a single writer. 
        * \par
        * This function returns either when it gets read access or when the time-out interval elapses. By default,
        * this function waits indefinitely.
        * \returns \c true, if read access could be acquired, \c false otherwise. */
      NVSG_API bool waitToRead( size_t milliseconds = 0xFFFFFFFF //<! Time-out interval, in miliseconds. 
                              ) const;

      //! Signal finished reading the object.
      /** A read access to this object, that has been acquired by a previous call to waitToRead, will be ended
        * by a call to this function. */
      NVSG_API void doneReading( void ) const;
      
      //! Acquire write access in a multithreaded enviroment.
      /** For safety in a multithreaded environment, an object maintains a read/write locking mechanism that allows 
        * contemporaneous access for multiple readers but write access for only a single writer. 
        * \par
        * This function returns either when it gets write access or when the time-out interval elapses. By default,
        * this function waits indefinitely.
        * \returns \c true, if write access could be acquired, \c false otherwise. */
      NVSG_API bool waitToWrite( size_t milliseconds = 0xFFFFFFFF //!< Time-out interval, in millisceconds.
                               ) const;
      
      //! Signal when finished writing the object.
      /** A write access to this object that has been acquired by a previous call to waitToWrite will be ended
        * by a call to this function. */
      NVSG_API void doneWriting( void ) const;

    protected:
      //! Constructor
      NVSG_API Scene( void );

      //! Protected destructor to prevent explicit creation on stack.
      NVSG_API virtual ~Scene( void );

    private:
      nvmath::Vec3f                                   m_backColor;
      std::vector<const Camera *>                     m_cameras;              //!< vector of cameras in scene
      std::vector<const Animation<nvmath::Trafo>*>    m_cameraAnimations;     //!< vector of camera animations
      const StateSet                                * m_defaultStateSet;      //!< default scene state set. Could be NULL!
      const StateSet                                * m_overrideStateSet;     //!< override scene state set. Could be NULL!
      const Node                                    * m_root;                 //!< Scene root node. Could be NULL!

      nvutil::SWMRSync            m_lock;
  };

  inline void Scene::addCamera( const Camera *pCamera )
  {
    __TRACE();
    m_cameras.push_back( pCamera );
    pCamera->addRef();
  }

  inline void Scene::removeCamera( size_t index )
  {
    __TRACE();
    __ASSERT(index < m_cameras.size());

    m_cameras[index]->removeRef();
    m_cameras.erase(m_cameras.begin()+index);
  }

  inline void Scene::addCameraAnimation( const Animation<nvmath::Trafo> *pAnimation )
  {
    __TRACE();
    m_cameraAnimations.push_back( pAnimation );
    pAnimation->addRef();
  }

  inline void Scene::removeCameraAnimation( size_t index )
  {
    __TRACE();
    __ASSERT(index < m_cameraAnimations.size());

    m_cameraAnimations[index]->removeRef();
    m_cameraAnimations.erase(m_cameraAnimations.begin()+index);
  }

  inline const Camera * Scene::getCamera( size_t index ) const
  {
    __TRACE();
    return( ( index < m_cameras.size() ) ? m_cameras[index] : NULL );
  }

  inline const Animation<nvmath::Trafo> * Scene::getCameraAnimation( size_t index ) const
  {
    __TRACE();
    return( ( index < m_cameraAnimations.size() ) ? m_cameraAnimations[index] : NULL );
  }

  inline const StateSet  * Scene::getDefaultStateSet( void ) const
  {
    __TRACE();
    return( m_defaultStateSet );
  }

  inline const StateSet * Scene::getOverrideStateSet( void ) const
  {
    __TRACE();
    return( m_overrideStateSet );
  }

  inline size_t Scene::getNumberOfCameras( void ) const
  {
    __TRACE();
    return( m_cameras.size() );
  }

  inline size_t Scene::getNumberOfCameraAnimations( void ) const
  {
    __TRACE();
    return( m_cameraAnimations.size() );
  }

  inline const Node * Scene::getRootNode( void ) const
  {
    __TRACE();
    return( m_root );
  }

  inline size_t Scene::getNumberOfFrames( void ) const
  {
    __TRACE();
    __ASSERT( m_root );
    size_t nof = m_root->getNumberOfFrames();
    for ( size_t i=0 ; !nof && i<m_cameraAnimations.size() ; i++ )
    {
      nof = m_cameraAnimations[i]->getNumberOfFrames();
    }

#if defined( _DEBUG )
    for ( size_t i=0 ; i<m_cameraAnimations.size() ; i++ )
    {
      size_t nofca = m_cameraAnimations[i]->getNumberOfFrames();
      __ASSERT( ( nofca == 0 ) || ( nofca == nof ) );
    }
#endif

    return( nof );
  }

  inline bool Scene::containsCgFx( void ) const
  {
    __TRACE();
    return( m_root->containsCgFx() );
  }

  inline bool Scene::containsLight( void ) const
  {
    __TRACE();
    return( m_root->containsLight() );
  }

  inline bool Scene::containsTransparentMaterial( void ) const
  {
    __TRACE();
    return( m_root->containsTransparentMaterial() );
  }

  inline bool Scene::containsTransparentTexture( void ) const
  {
    __TRACE();
    return( m_root->containsTransparentTexture() );
  }

  inline const nvmath::Vec3f& Scene::getBackColor( void ) const
  {
    __TRACE();
    return( m_backColor );
  }

  inline void Scene::setBackColor( const nvmath::Vec3f &color )
  {
    __TRACE();
    m_backColor = color;
  }

  inline bool Scene::waitToRead( size_t milliseconds ) const
  {
    __TRACE();
    size_t status;
    return m_lock.waitToRead(milliseconds, status);
  }

  inline void Scene::doneReading( void ) const
  {
    __TRACE();
    m_lock.doneReading();
  }

  inline bool Scene::waitToWrite( size_t milliseconds ) const
  {
    __TRACE();
    size_t status;
    return m_lock.waitToWrite(milliseconds, status);
  }

  inline void Scene::doneWriting( void ) const
  {
    __TRACE();
    m_lock.doneWriting();
  }
} //  namspace nvsg
