// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include "nvsgcommon.h"

#include "nvmath/Vec2f.h"
#include "nvmath/Vec3f.h"
#include "nvmath/Mat44f.h"
#include "nvmath/Quatf.h"
#include "nvsg/MonoCamera.h"

namespace nvsg
{
  //! A simple camera
  /** This camera type is used to group the simple camera types (as Parallel/Perspective). */
  class SimpleCamera : public MonoCamera
  {
    public:
      //! Interface for cloning a SimpleCamera.
      NVSG_API virtual const SimpleCamera * clone( void ) const = 0;

      //! Determine whether the data of this object is shared.
      /** \returns  \c true if the object's data is shared, otherwise \c false.  */
      NVSG_API virtual bool isDataShared() const;

      //! Get the DataID of the data of this object.
      /** \returns  DataID of the object's data.  */
      NVSG_API virtual DataID getDataID() const;

      //
      // stereo stuff
      //

      //! Get the projection translation (used with stereo only).
      /** \returns  The projection translation in x */
      NVSG_API virtual float getProjectionDx() const;

      //! Set the project translation (used with stereo only).
      NVSG_API virtual void setProjectionDx( float dx   //!<  shift of projection in x-direction
                                           );

      //
      // head light interface
      //

      //! Get number of headlights attached to the camera.
      /** \returns  The number of headlights attached to camera */
      NVSG_API virtual size_t  getNumberOfHeadLights() const;

      //! Get the indexed headlight of this camera
      /** \returns  A constant pointer to the \a indexed headlight*/
      NVSG_API virtual const LightSource * getHeadLight( size_t index   //!<  index of head light to get
                                                       ) const;

      //! Set the headlight position and orientation relative to the camera
      /** \returns \c true if a headlight with this index exists, otherwise \c false */
      NVSG_API virtual bool setHeadLightTrafo( const nvmath::Vec3f &trans   //!<  position of headlight
                                             , const nvmath::Quatf &quat    //!<  orientation of headlight
                                             , size_t index = 0             //!<  index of headlight
                                             );

      //! Get the position/orientation of the indexed headlight of this camera
      /** It is assumed that the camera holds at least \a index headlights. */
      NVSG_API virtual void getHeadLightTrafo( 
        nvmath::Vec3f &trans       //!<  translation relative to the camera
      , nvmath::Quatf &quat        //!<  rotation relative to the camera
      , size_t index = 0           //!<  index of the headlight to get information from
      ) const;

      //! Add a headlight on this camera.
      /** A camera can have a number of headlights that are positioned relative to the camera, i.e. they move together
        * with the camera. These lights are not part of the scene, but part of the camera. The default position and
        * orientation of a new headlight is the camera position and orientation. */
      NVSG_API virtual void addHeadLight( 
        const LightSource * pLight                                               //!<  The light source to add
      , const nvmath::Vec3f &position = nvmath::Vec3f( 0.0f, 0.0f, 0.0f )        //!<  The position relative to the camera
      , const nvmath::Quatf &quat     = nvmath::Quatf( 0.0f, 0.0f, 0.0f, 1.0f )  //!<  The rotation relative to the camera
      );

      //! Remove a headlight.
      /** \returns \c true if headlight was removed, otherwise \c false */
      NVSG_API virtual bool removeHeadLight( size_t index = 0   //!<  index of headlight to remove
                                           );

      //
      // basic camera interface
      //

      //! Get the distance to the target point.
      /** \returns  The distance to the target point */
      NVSG_API virtual float getTargetDistance() const;

      //! Set the distance to the target point.
      NVSG_API virtual void setTargetDistance( float td //!< distance to target point (or viewing plane)
                                             );

      //! Get the camera position.
      /** \returns  The position of camera */
      NVSG_API virtual const nvmath::Vec3f & getPosition() const;

      //! Set the camera position.
      NVSG_API virtual void setPosition( const nvmath::Vec3f& pos   //!<  position
                                       );

      //! Get the up direction.
      /** \returns  The up direction of camera */
      NVSG_API virtual nvmath::Vec3f getUpVector() const;

      //! Set the up direction.
      NVSG_API virtual void setUpVector( const nvmath::Vec3f &up //!< up direction (might change viewing direction)
                                       );

      //! Get viewing direction
      /** The default viewing direction of a SimpleCamera is along the negative z-axis.
        * \returns  The viewing direction */
      NVSG_API virtual nvmath::Vec3f getDirection() const;

      //! Set the viewing direction.
      NVSG_API virtual void setDirection( const nvmath::Vec3f& dir  //!<  new viewing direction (might change upVector)
                                        );
      
      //! Set the orientation of the camera.
      NVSG_API virtual void setOrientation( const nvmath::Quatf &quat   //!<  quaternion that determines direction and up vector
                                          );

      //! Set the orientation of the camera.
      NVSG_API virtual void setOrientation( const nvmath::Vec3f &dir    //!<  direction to look at
                                          , const nvmath::Vec3f &up     //!<  up vector
                                          );

      //! Interface for setting the orientation of the camera.
      NVSG_API virtual void setOrientation( const nvmath::Vec3f &axis   //!<  axis to rotate about
                                          , float angle                 //!<  angle to rotate
                                          );

      //! Get the current orientation.
      /** \returns  quaternion that describes the orientation */
      NVSG_API virtual nvmath::Quatf getOrientation( void ) const;

      //! Get the world-to-view transformation.
      /** \returns  world-to-camera transformation */
      NVSG_API virtual nvmath::Mat44f getWorldToViewMatrix() const;

      //! Get the inverse of the camera specific world-to-view transformation
      /** \returns  inverse of the world-to-view transformation (i.e. view to world) */
      NVSG_API virtual nvmath::Mat44f getViewToWorldMatrix()  const;

      //! Set the window offset.
      NVSG_API virtual void setWindowOffset( const nvmath::Vec2f &offset    //!<  window ofset to viewing reference point
                                           );

      //! Get the window offset.
      /** \returns  The window offset */
      NVSG_API virtual const nvmath::Vec2f & getWindowOffset() const;

      //!  Set the window size.
      NVSG_API virtual void setWindowSize( const nvmath::Vec2f &size    //!<  window size (world coordinates)
                                         );

      //! Get the window size.
      /** \returns  The window size */
      NVSG_API virtual const nvmath::Vec2f & getWindowSize() const;

      //! Set the aspect ratio.
      NVSG_API virtual void setAspectRatio( float ar                //!<  new aspect ratio
                                          , bool keepWidth = false  //!<  keep width or height constant
                                          );
     
      //! Get the current aspect ratio
      /** \returns  The width versus height of the camera window */
      NVSG_API virtual float getAspectRatio() const;
      
      //! Set the relative window region to view
      NVSG_API virtual void setWindowRegion( const nvmath::Vec2f &lowerLeft  //!< relative lower left position
                                           , const nvmath::Vec2f &upperRight //!< relative upper right position
                                           );

      //! Get the relative window region to view
      NVSG_API virtual void getWindowRegion( nvmath::Vec2f &lowerLeft     //!< relative lower left position
                                           , nvmath::Vec2f &upperRight    //!< relative upper right position
                                           ) const;

            //! Enable/Disable automatic clip plane determination.
      NVSG_API virtual void setAutoClipPlanes( bool on    //!<  true to enable, false to disable
                                             );

      //! Get auto clipping state.
      /** If auto clip is on, the camera determines the optimal near and front clipping distances out of the scene. If
        * the user has additional information about the scene that might help clipping, it should be turned off.
        * \returns  \c true, if auto clip is on, otherwise \c false */
      NVSG_API virtual bool isAutoClipPlanes() const;

      //! Calculate the distances to the near and far clipping planes.
      /** Near and far distances can be used internally to the camera to improve depth resolution. This function
        * should determine these values such that the sphere given by \a center and \a radius just fits between them. */
      NVSG_API virtual void calcNearFarDistances( const nvmath::Sphere3f &sphere  //!<  bounding sphere of the world to look at
                                                );
 
      //! Set the near clipping distance.
      NVSG_API virtual void setNearDistance( float nd   //!<  near clipping distance
                                           );
      
      //! Get distance to near clipping plane
      /** Near and far distances can be used internally to the camera to improve depth resolution.
        * \returns  The distance to the near clipping plane */
      NVSG_API virtual float getNearDistance() const;

      //! Get distance to far clipping plane
      /** Near and far distances can be used internally to the camera to improve depth resolution.
        * \returns  The distance to the far clipping plane */
      NVSG_API virtual float getFarDistance() const;

      //! Set the far clipping distance.
      NVSG_API virtual void setFarDistance( float fd    //!<  far clipping distance
                                          );

      //
      // operations to perform on a camera
      //

      //! Move the camera.
      NVSG_API virtual void move( const nvmath::Vec3f& delta  //!<  delta to move the camera (camera coordinates)
                                );

      //! Orbit the camera.
      NVSG_API virtual void orbit( const nvmath::Vec3f& axis  //!<  axis to orbit around (camera coordinates)
                                 , float rad                  //!<  angle to orbit (radians)
                                 );

      //! Interface for rotating the camera.
      NVSG_API virtual void rotate( const nvmath::Vec3f& axis     //!<  axis to rotate around 
                                  , float rad                     //!<  angle to rotate (radians)
                                  , bool  cameraRelative = true   //!<  flag to indicate if the axis is camera or world relative
                                  );

      //! Zoom by a factor.
      NVSG_API virtual void zoom( float factor    //!< zoom in/out by this factor; changes windowSize
                                );

      //! Zoom to fit view to a sphere.
      NVSG_API virtual void zoom( const nvmath::Sphere3f &sphere  //!<  sphere to zoom on
                                , float fovy = -1.0f              //!<  field of view; if negative, keep the current
                                );

    protected:
      //! Constructor
      NVSG_API SimpleCamera();

      //! Copy constructor
      NVSG_API SimpleCamera( const SimpleCamera& rhs //!< source camera object
                           );

      //! Protected destructor to prevent explicit creation on stack.
      NVSG_API virtual ~SimpleCamera();

    private:
      std::vector<const LightSource *>  m_headLights;       //!< vector of the headlights

      nvmath::Quatf m_orientation;  //!<  orienation of the camera
      nvmath::Vec3f  m_position;      //!< position of camera
      float          m_targDist;      //!< distance to projection plane; for non-perspective cameras this somewhat artificial, for ease of use

      bool  m_isAutoClipPlanes;       //!< true, to use automatic generated clipping planes
      float m_farDist;                //!< distance from position to far clipping plane
      float m_nearDist;               //!< distance from position to near clipping plane

      float m_projectionDx;           //!< translation of projection (used with stereo for example)

      nvmath::Vec2f m_windowSize;     //!< width (x) and height (y) of viewing window
      nvmath::Vec2f m_windowOffset;   //!< offset from the viewing reference point to the center of the viewing window
      nvmath::Vec2f m_lowerLeft;      //!< lower left position of window region to view
      nvmath::Vec2f m_upperRight;     //!< upper right position of window region to view
  };

  inline const LightSource * SimpleCamera::getHeadLight( size_t index ) const
  {
    __TRACE();
    __ASSERT( index < m_headLights.size() );
    return m_headLights[index];
  }

} // namespace
