// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include "nvsgcommon.h"

#include "nvmath/Trafo.h"
#include "nvsg/Animation.h"
#include "nvsg/Triangles.h"

#ifdef _DEBUG
# ifndef new
#  define new new(__FILE__, __LINE__)
#  define _DEFINED_DBGNEW // restrict the 'new' macro to this header file only
# endif
#endif


namespace nvsg
{
  //! Internal helper class for SkinnedTriangles.
  /** Instead of the vertices and normals from Triangles, SkinnedTriangles use a special construction that has a vector of
    * "Bones" per vertex, from which the vertices and normals are calculated. This calculation depends on the animation
    * step. A Bone essentially consists of an Animation, an offset, and a weight. Out of these values a vertex is
    * calculated using a vector of Bones. */
  class Bone
  {
    public:
      //! Constructor
      Bone( void );

      //! Copy constructor
      Bone( const Bone &rhs );

      //! Destructor
      ~Bone( void );

    public:
      //! Calculate a vertex.
      /** For animation step \a frame and reference point \a v, a modified vertex is calculated. All Bones of a Skin equally
        * contribute to the animated vertex this way.
        * \returns  An animated vertex relative to \a v */
      nvmath::Vec4f operator()( size_t frame, const nvmath::Vec4f &v )  const;

      //! Get the Animation.
      /** \returns  A pointer to the Animation */
      const Animation<nvmath::Trafo> * getAnimation( void ) const;

      //! Get the transformation of the animation step.
      /** \returns  The animated transform on step \a frame */
      const nvmath::Mat44f & getMatrix( size_t frame   //!<  animation step to get transformation at
                                      ) const;

      //! Get the weight.
      /** \returns  The weight of this Bone */
      float getWeight( void ) const;

      //! Set the Animation.
      /** Setting the Animation also sets the base inverse as the inverse of the first animation step.  */
      NVSG_API void setAnimation( const Animation<nvmath::Trafo> *pAnimation   //!<  new animation for this Bone
                                );

      //! Set the base inverse.
      /** All animation steps are done relative to this base inverse. Usually it is calculated out of the first
        * animation step. */
      void setInverse( const nvmath::Mat44f &inverse    //!<  new base inverse to use
                     );

      //! Set the trafo.
      /** Just sets the trafo to use and invalidates the animation step.  */
      void setMatrix( const nvmath::Mat44f& trafo    //!<  trafo to use
                    );

      //! Set the weight.
      /** The weight is used to weight this Bones contribution to the vertex calculation of a Skin. */
      void setWeight( float weight    //!<  weight for this bone
                    );

    private:
      nvmath::Mat44f              m_baseInverse;
      const Animation<nvmath::Trafo>  * m_animation;
      float               m_weight;

      mutable size_t  m_currFrame;
      mutable nvmath::Mat44f  m_currTrafo;
  };

  inline Bone::Bone( void )
  : m_animation(NULL)
  , m_currFrame(0xFFFFFFFF)
  {
    __TRACE();
  }

  inline Bone::Bone( const Bone &rhs )
  : m_baseInverse( rhs.m_baseInverse )
  , m_weight( rhs.m_weight )
  , m_currFrame( rhs.m_currFrame )
  , m_currTrafo( rhs.m_currTrafo )
  {
    __TRACE();
    m_animation = rhs.m_animation;
    if ( m_animation )
    {
      m_animation->addRef();
    }
  }

  inline Bone::~Bone( void )
  {
    __TRACE();
    if ( m_animation )
    {
      m_animation->removeRef();
    }
  }

  inline nvmath::Vec4f Bone::operator()( size_t frame, const nvmath::Vec4f &vIn ) const
  {
    __TRACE();
    nvmath::Vec4f vOut;
    if ( m_animation )
    {
      vOut = (*m_animation)[frame].getMatrix() * m_baseInverse * vIn * m_weight;
    }
    else
    {
      vOut = vIn * m_weight;
    }
    return( vOut );
  }

  inline const Animation<nvmath::Trafo> * Bone::getAnimation( void ) const
  {
    __TRACE();
    return( m_animation );
  }

  inline const nvmath::Mat44f & Bone::getMatrix( size_t frame ) const
  {
    __TRACE();
    if ( m_animation )
    {
      if ( m_currFrame != frame )
      {
        m_currTrafo = (*m_animation)[frame].getMatrix();
        m_currFrame = frame;
      }
    }
    return( m_currTrafo );
  }

  inline float Bone::getWeight( void ) const
  {
    __TRACE();
    return( m_weight );
  }

  inline void Bone::setAnimation( const Animation<nvmath::Trafo> *pAnimation )
  {
    __TRACE();
    if ( m_animation != pAnimation )
    {
      if ( m_animation )
      {
        m_animation->removeRef();
      }
      m_animation = pAnimation;
      if ( m_animation )
      {
        m_animation->addRef();

        const nvmath::Trafo trafo = (*m_animation)[0];
        nvmath::Quatf rot = trafo.getOrientation();
        nvmath::Vec3f pos = trafo.getTranslation();
        m_baseInverse = nvmath::Mat44f( nvmath::Vec3f( 0.0f, 0.0f, 0.0f ), nvmath::Quatf( rot[0], rot[1], rot[2], -rot[3] ) )

                      * nvmath::Mat44f( nvmath::Vec4f( 1.0f, 0.0f, 0.0f, -pos[0] ),
                                nvmath::Vec4f( 0.0f, 1.0f, 0.0f, -pos[1] ),
                                nvmath::Vec4f( 0.0f, 0.0f, 1.0f, -pos[2] ),
                                nvmath::Vec4f( 0.0f, 0.0f, 0.0f, 1.0f ) );
        m_currFrame   = 0;
        m_currTrafo   = nvmath::Mat44f( pos, rot );
      }
      else
      {
        m_currFrame = 0xFFFFFFFF;
      }
    }
  }

  inline void Bone::setInverse( const nvmath::Mat44f &inverse )
  {
    __TRACE();
    m_baseInverse = inverse;
  }

  inline void Bone::setMatrix( const nvmath::Mat44f &trafo )
  {
    __TRACE();
    m_currTrafo = trafo;
    m_currFrame = 0xFFFFFFFF;
  }

  inline void Bone::setWeight( float weight )
  {
    __TRACE();
    m_weight = weight;
  }

  //! Internal helper class for SkinnedTriangles.
  /** Instead of the vertices and normals from Triangles, SkinnedTriangles use a special construction that has a vector of
    * "Bones" per vertex, from which the vertices and normals are calculated. This calculation depends on the animation
    * step.
    * \par
    * The vertex and the normal of a Skin are starting values, that are calculated out of the first animation step of
    * each bone, together with their offset and weight. These values are used as starting values on the following
    * animation steps. */
  typedef struct
  {
    std::vector<Bone>  m_bones;    //!<  vector of "Bones" from wich a vertex and a normal is calculated
    nvmath::Vec3f         m_normal;   //!<  current normal, calculated from the vector of "Bones"
    nvmath::Vec3f         m_vertex;   //!<  current vertex, calculated from the vector of "Bones"
  } Skin;

  //! GeoSet for animated Triangles.
  /** This class has additional information to calculate the vertices and normals of a Triangles depending on the
    * animation step. */
  class SkinnedTriangles : public Triangles
  {
    public:
      //! Create a SkinnedTriangles.
      /** \returns a constant pointer to a SkinnedTriangles. */
      NVSG_API static const SkinnedTriangles * create( void );

      //! Create a SkinnedTriangles by copying from a Triangles.
      /** \returns a constant pointer to a SkinnedTriangles. */
      NVSG_API static const SkinnedTriangles * createFromBase( const Triangles &rhs );

      //! Get a clone of this SkinnedTriangles object
      /** \returns a constant pointer to a SkinnedTriangles. */
      NVSG_API virtual const SkinnedTriangles * clone( void ) const;

      //! Determine whether the data of this object is shared.
      /** \returns  \c true if the object's data is shared, otherwise \c false.  */
      NVSG_API virtual bool isDataShared( void ) const;

      //! Get the DataID of the data of this object.
      /** \returns  DataID of the object's data.  */
      NVSG_API virtual DataID getDataID( void ) const;

      //! Get the number of animation frames that are part of this SkinnedTriangles.
      /** \returns The number of animation frames if there is an animation attached, otherwise zero.  */
      NVSG_API virtual size_t getNumberOfFrames( void ) const;

      //! Get number of skins.
      /** In addition to vertices, normals, and so on, a SkinnedTriangles contains so-called "skins". A skin is the
        * information where one vertex and its normal can be calculated from, depending on the animation frame. */
      NVSG_API size_t getNumberOfSkins( void ) const;

      //! Get the skins.
      /** \returns  A constant pointer to the skins */
      NVSG_API const Skin * getSkins( void ) const;

      //! Calculate vertices and normals depending on the animation frame.
      /** SkinnedTriangles contain animation information on each vertex. When the animation frame is set, the vertices
        * and normals are recalculated. If bCalculateTangentSpace is true, tangents and binormals are also recalculated. */
      NVSG_API void setAnimationFrame( size_t frame             //!<  animation frame to set
                                     , bool bCalculateTangentSpace    //!<  if true, recalculate tangents and binormals
                                     );

      //! Overwrite of Triangles::setNormals.
      /** Sets the normals in the skins and then calls to Triangles::setNormals. */
      NVSG_API void setNormals( const nvmath::Vec3f * pNor    //!<  normals to set
                              , size_t numNor         //!<  number of normals
                              );

      //! Overwrite of Triangles::setNormals.
      /** Sets the normals in the skins and then calls to Triangles::setNormals. */
      NVSG_API void setNormals( size_t pos // Start position
                              , const nvmath::Vec3f * pNor    //!<  normals to set
                              , size_t numNor         //!<  number of normals
                              );
      //! Set the skins.
      /** A Skin contains the information how a vertex and a normal is calculated for each animation step.  */
      NVSG_API void setSkins( const Skin *pSkins    //!<  skins to set
                            , size_t numSkins       //!<  number of skins
                            );

      //! Set skins
      /** Copies /a numSkins skins pointed to by /a pSkins, starting at position /a pos 
        * inside the range of pre-existing skins.
        *
        * Pre-existing skins in the range [\a pos, \a pos + \a numSkins) will be replaced.
        * Pre-existing skins outside this range remain untouched.
        *
        * If you specify -1 for \a pos or if \a pos specifies the number of skins currently stored,
        * the skins pointed to by \a pSkins will be appended to the pre-existing faces. 
        *
        * If you not specify -1 for \a pos or \a pos neither specifies the number of skins currently
        * stored nor does it specify a valid position inside the range of pre-existing skins, 
        * the behaviour is undefined! */
      NVSG_API void setSkins( size_t pos
                            , const Skin *pSkins
                            , size_t numSkins );

    protected:
      //! Constructor.
      NVSG_API SkinnedTriangles( void );

      //! Partial Constructor
      NVSG_API SkinnedTriangles( const Triangles &rhs );

      //! Copy Constructor
      NVSG_API SkinnedTriangles( const SkinnedTriangles &rhs );

      //! Protected destructor to prevent explicit creation on stack.
      NVSG_API virtual ~SkinnedTriangles( void );

    private:
      nvutil::RCPtr<nvutil::RCVector<Skin> >  m_skins; // sharable skins
  };

  // - - - - - - - - - - - - - - - - - - - - - - - - - 
  // inlines
  // - - - - - - - - - - - - - - - - - - - - - - - - - 

  inline size_t SkinnedTriangles::getNumberOfSkins( void )  const
  { 
    __TRACE();
    return( m_skins->size() ); 
  }

  inline const Skin * SkinnedTriangles::getSkins( void )  const 
  { 
    __TRACE();
    return( &(*m_skins)[0] ); 
  }

} //  namespace nvsg

#ifdef _DEFINED_DBGNEW
# undef new
#endif
