// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include "nvsgcommon.h"

#include "nvsg/Object.h"
#include "nvutil/Incarnation.h"

namespace nvsg
{  
  class StateSet;

  //! Pure virtual base class of all state attribute objects.
  /** Defines the interface that every state attribute (e.g. FaceAttribute) has to resolve.  */
  class StateAttribute : public Object
  {
    public:
      //! Get a clone of this attribute
      NVSG_API virtual const StateAttribute * clone( void ) const = 0;

      //! Determine whether the data of this object is shared.
      /** \returns  \c true if the object's data is shared, otherwise \c false.  */
      NVSG_API virtual bool isDataShared( void ) const;

      //! Get the DataID of the data of this object.
      /** \returns  DataID of the object's data.  */
      NVSG_API virtual DataID getDataID( void ) const;

      //! Get attribute's incarnation for comparison purposes.
      /** \returns The current incarnation. */
      NVSG_API const nvutil::Incarnation& getIncarnation( void ) const;

      //! Get the number of parents of this StateAttribute.
      /** \returns  The number of parents if this StateAttribute */
      NVSG_API size_t getNumberOfParents( void ) const;

      //! Get a constant pointer to the parent StateSet at position \a index.
      /** \returns  The constant pointer to the parent StateSet at position \a index */
      NVSG_API const StateSet * getParent( size_t index    //!<  The position to get the parent at
                                         ) const;

      //! Ask if this StateAttribute is a transparent Material
      /** \returns  \c false */
      NVSG_API virtual bool containsTransparentMaterial( void ) const;

      //! Invalidate the transparent Material containment cache.
      /** Invalidating the cache is carried to all parents of this Node.  */
      NVSG_API virtual void invalidateTransparentMaterialContainment( void );

      //! Ask if this StateAttribute is a transparent Texture
      /** \returns  \c false */
      NVSG_API virtual bool containsTransparentTexture( void ) const;

      //! Invalidate the transparent Texture containment cache.
      /** Invalidating the cache is carried to all parents of this Node.  */
      NVSG_API virtual void invalidateTransparentTextureContainment( void );

    protected:
      //! Constructor
      NVSG_API StateAttribute( void );
      //! Copy constructor
      NVSG_API StateAttribute(const StateAttribute& rhs);

      //! Protected destructor to prevent explicit creation on stack.
      NVSG_API virtual ~StateAttribute( void );

    protected:
      nvutil::Incarnation m_incarnation;    //!< current incarnation

    private:
      // A StateAttribute can be the child of more than one parent. 
      // The child-to-parent connection is used, for example, for 
      // invalidating the parent's transparent Material containment flag.
      friend class StateSet;
      NVSG_API void addParent( const StateSet * stateSet );
      NVSG_API void removeParent( const StateSet * stateSet );

    private:
      std::vector<const StateSet*> m_parents;     //!<  vector of parent StateSet objects
  };

  inline size_t StateAttribute::getNumberOfParents( void ) const
  {
    __TRACE();
    return( m_parents.size() );
  }

  inline const StateSet * StateAttribute::getParent( size_t index ) const
  {
    __TRACE();
    __ASSERT( index < m_parents.size() );   
    return( m_parents[index] );
  }

  inline const nvutil::Incarnation& StateAttribute::getIncarnation( void ) const
  {
    __TRACE();
    return m_incarnation;
  }

} //  namespace nvsg

