// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include "nvsgcommon.h"

#include "nvsg/Object.h"
#include <vector>

namespace nvsg
{
  class GeoNode;
  class StateAttribute;

  //! Class to hold a list of StateAttributes.
  /** A StateSet is a complete description of the visual appearance of some geometry. This is achieved by a list of
    * StateAttributes, each describing parts of the appearance (e.g. FaceAttribute, Material, TextureAttribute).
    */
  class StateSet : public Object
  {
    public:
      //! Create a StateSet.
      /** \returns a constant pointer to a StateSet. */
      NVSG_API static const StateSet * create( void );

      //! Create a StateSet by copying from an Object.
      /** \returns a constant pointer to a StateSet. */
      NVSG_API static const StateSet * createFromBase( const Object &rhs );

      //! Create a StateSet by cloning this.
      /** \returns a constant pointer to a StateSet. */
      NVSG_API virtual const StateSet * clone( void ) const;

      //! Determine whether the data of this object is shared.
      /** \returns \c true if the object's data is shared, otherwise \c false.  */
      NVSG_API virtual bool isDataShared( void ) const;

      //! Get the DataID of the data of this object.
      /** \returns  DataID of the object's data.  */
      NVSG_API virtual DataID getDataID( void ) const;

      //! Add a StateAttribute to the attribute list.
      /** There is always only one attribute of the same type in the list. 
        * So for example if you add a second StateAttribute of the type CgFX the other 
        * CgFX attribute will be replaced.
        * \note In RenderTraverser::RT_EFFECTED mode the CgFX attribute is used with higher 
        * priority than every other attribute. E.g an additional material attribute will be 
        * neglected until you use RenderTraverser::RT_SHADED.*/
      NVSG_API void addAttribute( const StateAttribute *attrib    //!< attribute to add
                                );

      //! Get the number of StateAttributes in this StateSet.
      /** \returns  number of StateAttributes in this StateSet */
      NVSG_API size_t getNumberOfAttributes( void ) const;

      //! Returns a pointer to the constant StateAttribute at position \a index.
      /** \returns A pointer to the constant StateAttribute at position \a index. 
        * \note The behavior is undefined if called with an invalid index. */
      NVSG_API const StateAttribute * getAttribute( size_t index    //!<  position to get StateAttribute at
                                                  ) const;

      //! Get a constant pointer to the StateAttribute with the object code \a oc.
      /** \returns constant pointer to the StateAttribute with object code \a oc or NULL, if there is none. 
        * \note You will find the non custom object codes here: #OC_OBJECT */
      NVSG_API const StateAttribute * getAttributeByObjectCode( unsigned int oc   //!< object code of StateAttribute to get
                                                              ) const;

      //! Remove a specified StateAttribute.
      /** \returns true, if the attribute was successfully removed, false otherwise. */
      NVSG_API bool removeAttribute( const StateAttribute * attrib    //!<  StateAttribute to remove
                                   );

      //! Remove the StateAttribute at position \a index.
      /** \returns true, if the attribute was successfully removed, false otherwise. */
      NVSG_API bool removeAttribute( size_t index   //!<  index of StateAttribute to remove
                                   );

      //! Remove the StateAttribute with the object code \a oc.
      /** \returns true, if the attribute was successfully removed, false otherwise. 
        * \note You will find the non custom object codes here: #OC_OBJECT */
      NVSG_API bool removeAttributeByObjectCode( unsigned int oc    //!<  StateAttribute class to remove
                                               );
      
      //! Get the number of owners of this StateSet.
      /** \returns  The number of owners if this StateSet */
      NVSG_API size_t getNumberOfOwners( void ) const;

      //! Get a constant pointer to the owner node at position \a index.
      /** \returns  The constant pointer to the owner node at position \a index */
      NVSG_API const GeoNode * getOwner( size_t index    //!<  The position to get the owner at
                                       ) const;

      //! Ask if this StateSet contains any CgFx attributes.
      /** \returns  \c true if the StateSet contains a CgFx attribute, otherwise \c false.  */
      NVSG_API virtual bool containsCgFx( void ) const;

      //! Invalidate the CgFx containment cache.
      /** The cache is invalidated and all owners of this StateSet are invalidated on CgFx containment.  */
      NVSG_API virtual void invalidateCgFxContainment( void );

      //! Ask if this StateSet contains any transparent Material attributes.
      /** \returns  \c true if the StateSet contains a transparent Material attribute, otherwise \c false.  */
      NVSG_API virtual bool containsTransparentMaterial( void ) const;

      //! Invalidate the transparent Material containment cache.
      /** The cache is invalidated and all owners of this StateSet are invalidated on CgFx containment.  */
      NVSG_API virtual void invalidateTransparentMaterialContainment( void );

      //! Ask if this StateSet contains any transparent Texture.
      /** \returns  \c true if the StateSet contains a transparent Texture, otherwise \c false.  */
      NVSG_API virtual bool containsTransparentTexture( void ) const;

      //! Invalidate the transparent Texture containment cache.
      /** The cache is invalidated and all owners of this StateSet are invalidated on CgFx containment.  */
      NVSG_API virtual void invalidateTransparentTextureContainment( void );
    protected:
      //! Constructor
      NVSG_API StateSet( void );

      //! Partial Constructor
      NVSG_API StateSet( const Object &rhs );

      //! Copy constructor
      NVSG_API StateSet( const StateSet& rhs );

      //! Protected destructor to prevent explicit creation on stack.
      NVSG_API virtual ~StateSet( void );

      //! Invalidate all caches of this StateSet, maybe depending on the added or removed StateAttribute.
      /** If the CgFx containment changes by adding/removing \a attrib, the CgFx cache is invalidated. */
      NVSG_API void invalidateCaches( const StateAttribute *attrib, bool insert );

    private:
      // A Node can be owned by more than one GeoNode. The owner connection is 
      // used for example to invalidate the owner's containment cache.
      friend class GeoNode;
      void addOwner( const GeoNode * geoNode );
      void removeOwner( const GeoNode * geoNode );

    private:
      std::vector<const StateAttribute*>  m_stateAttribs;
      std::vector<const GeoNode*>         m_owners;       // owners of this state set are collected here

      //  State caches
      mutable bool  m_containsCgFx;
      mutable bool  m_containsCgFxValid;
      mutable bool  m_containsTransparentMaterial;
      mutable bool  m_containsTransparentMaterialValid;
      mutable bool  m_containsTransparentTexture;
      mutable bool  m_containsTransparentTextureValid;

  };

  inline const StateAttribute * StateSet::getAttribute( size_t index ) const
  {
    __TRACE();
    __ASSERT(index<m_stateAttribs.size()); // undefined behavior for invalid index
    return m_stateAttribs[index];
  }

  inline size_t StateSet::getNumberOfAttributes( void ) const
  {
    __TRACE();
    return( m_stateAttribs.size() );
  }

  inline size_t StateSet::getNumberOfOwners( void ) const
  {
    __TRACE();
    return( m_owners.size() );
  }

  inline const GeoNode * StateSet::getOwner( size_t index ) const
  {
    __TRACE();
    __ASSERT( index < m_owners.size() );
    return m_owners[index];
  }

//>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

  /*! \relates StateSet
   *  Compare two StateSets.
   *  Two StateSets are considered to be equal, if they have the same number of StateAttributes and each of these
   *  StateAttributes are pairwise equal.
   *  \returns \c true, if the StateSets are equal, \c false otherwise. */
  NVSG_API bool operator==( const StateSet &lhs   //!<  first StateSet to compare
                          , const StateSet &rhs   //!<  second StateSet to compare
                          );

} //  namespace nvsg
