// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include "nvsgcommon.h"

#include  "nvsg/nvsg.h"
#include  "nvsg/Camera.h"

namespace nvsg
{
  class CullData;

  //! Class to encapsulate the view specific states.
  class ViewState : public nvutil::RCObject
                  , public nvutil::ISWMRSync
  {
    public:
      //! Instantiates a default-constructed constant ViewState.
      /** \returns A pointer to the new created ViewState. */ 
      NVSG_API static const ViewState * create();

      //! Instantiates a constant copy of this ViewState object.
      /** \note The new copy does not inherit this ViewState's cull 
        * information but will initialy have its cull tree cleared!
        * \returns A pointer to the copy-constructed ViewState. */
      NVSG_API const ViewState * clone() const;

      //! Get the current animation frame.
      /** \returns  The current animation frame */
      NVSG_API size_t getAnimationFrame( void ) const;

      //! Get the current camera.
      /** \returns  A pointer to the camera in use with this view */
      NVSG_API const Camera * getCamera( void ) const;

      //! Ask if the view is animated.
      /** \returns \c true if this view is currently animated, otherwise \c false */
      NVSG_API bool isAnimating( void ) const;

      //! Ask if the view is in stereo mode.
      /** \returns \c true if this view is currently in stereo mode, otherwise \c false */
      NVSG_API bool isStereo( void )  const;

      //! Set the animation frame.
      NVSG_API void setAnimationFrame( size_t frame   //!<  frame to use for next evaluation
                                     );

      //! Enable/Disable animation.
      NVSG_API void setAnimation( bool state    //!<  true, to enable animation in this view, otherwise false
                                );

      //! Set the current camera.
      NVSG_API void setCamera( const Camera * pCamera   //!<  pointer to camera to use in this view
                             );

      //! Enable/Disable stereo.
      NVSG_API void setStereo( bool state   //!<  true, to enable stereo display, otherwise false
                             );

      //! Clear the current cull tree.
      NVSG_API void clearCullTree( void ) const ;

      //! Get the root of the CullData-tree.
      NVSG_API const CullData * getCullTree( void ) const;

      //! Set a given cull tree as the current.
      NVSG_API void setCullTree( const CullData *cullTree );

      //! Acquire read access in a multithreaded environment.
      /** For safety in a multithreaded environment, an object maintains a read/write locking mechanism that allows 
        * contemporaneous access for multiple readers but write access for only a single writer. 
        * \par
        * This function returns either when it gets read access or when the time-out interval elapses. By default,
        * this function waits indefinitely.
        * \returns \c true, if read access could be acquired, \c false otherwise. */
      NVSG_API bool waitToRead( size_t milliseconds = 0xFFFFFFFF //<! Time-out interval, in milliseconds. 
                              ) const;

      //! Signal when finished reading the object.
      /** A read access to this object, that has been acquired by a previous call to waitToRead, will be ended
        * by a call to this function. */
      NVSG_API void doneReading( void ) const;
      
      //! Acquire write access in a multithreaded enviroment.
      /** or safety in a multithreaded environment, an object maintains a read/write locking mechanism that allows 
        * contemporaneous access for multiple readers but write access for only a single writer. 
        * \par
        * This function returns either when it gets write access or when the time-out interval elapses. By default,
        * this function waits indefinitely.
        * \returns \c true, if write access could be acquired, \c false otherwise. */
      NVSG_API bool waitToWrite( size_t milliseconds = 0xFFFFFFFF //!< Time-out interval, in milliseconds.
                               ) const;
      
      //! Signal finished writing the object.
      /** A write access to this object, that has been acquired by a previous call to waitToWrite, will be ended
        * by a call to this function. */
      NVSG_API void doneWriting( void ) const;

    protected:
      //! Default-initializes a ViewState object.
      /** \note A default-constructwed ViewState has no camera attached! */
      NVSG_API ViewState( void );

      //! Initializes a ViewState as a copy of another ViewState object.
      /** \note The new constructed ViewState will not inherit the cull tree from the source ViewState 
        * - that is the cull tree for the new constructed ViewState initially is cleared! */ 
      NVSG_API ViewState( const ViewState& rhs );

      //! Protected destructor to prevent explicit creation on stack.
      NVSG_API virtual ~ViewState(void);

    private:
      // assigning a ViewState is not permitted!
      ViewState& operator=(const ViewState&);   

    private:
      size_t          m_animationFrame;
      const Camera  * m_camera;
      bool            m_stereoState;
      // the animation state might be modified by concurrently executing threads, 
      // declare it as volatile for this reason
      volatile bool   m_animationState;

      mutable const CullData  * m_cullTree;

      nvutil::SWMRSync  m_lock;
  };

  
  inline size_t ViewState::getAnimationFrame( void ) const
  {
    return( m_animationFrame );
  }

  inline bool ViewState::isAnimating( void ) const
  {
    return( m_animationState );
  }

  inline const Camera * ViewState::getCamera( void ) const
  {
    return( m_camera );
  }

  inline bool ViewState::isStereo( void ) const
  {
    return( m_stereoState );
  }

  inline void ViewState::setAnimationFrame( size_t frame )
  {
    if ( m_animationFrame != frame )
    {
      m_animationFrame = frame;
    }
  }

  inline void ViewState::setAnimation( bool state )
  {
    m_animationState = state;
  }

  inline void ViewState::setCamera( const Camera * pCamera )
  {
    if ( m_camera != pCamera )
    {
      if ( m_camera )
      {
        m_camera->removeRef();
      }
      m_camera = pCamera;
      if ( m_camera )
      {
        m_camera->addRef();
      }
    }
  }

  inline void ViewState::setStereo( bool state )
  {
    m_stereoState = state;
  }

  inline const CullData * ViewState::getCullTree( void ) const
  {
    return( m_cullTree );
  }

  inline void ViewState::setCullTree( const CullData *cullTree )
  {
    clearCullTree();
    m_cullTree = cullTree;
  }

  inline bool ViewState::waitToRead( size_t milliseconds ) const
  {
    size_t status;
    return m_lock.waitToRead(milliseconds, status);
  }

  inline void ViewState::doneReading( void ) const
  {
    m_lock.doneReading();
  }

  inline bool ViewState::waitToWrite( size_t milliseconds ) const
  {
    size_t status;
    return m_lock.waitToWrite(milliseconds, status);
  }

  inline void ViewState::doneWriting( void ) const
  {
    m_lock.doneWriting();
  }
} // namespace
