// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#pragma push_macro("free")
#undef free
#include <valarray>
#pragma pop_macro("free")

#include <set>

#include "nvsgcommon.h"

#include "nvsg/nvsg.h"
#include "nvtraverser/Traverser.h"

namespace nvtraverser
{  

  //! Traverser that can condense some drawables.
  /** Currently, this CondenseTraverser supports Triangles only.  */
  class CondenseTraverser : public Traverser
  {
  public:
    //! Constructor
    NVSG_API CondenseTraverser(void);

    //! Set the epsilon to use to distinguish vertex values.
    /** This epsilon multiplied by the size of the scene gives the tolerance for detecting equal vertices.  */
    NVSG_API void  setEpsilon( 
      float epsilon                 //!<  Epsilon to use
      );

  protected:
    //! Proteced destructor to prevent instantiation of a CondenseTraverser on stack.
    NVSG_API virtual ~CondenseTraverser(void);

    //! Initiate traversal of the scene with a view state.
    /** Calculates the tolerance for the scene and calls the Traverser::doApply().  */
    NVSG_API virtual void  doApply( const nvsg::ViewState *pViewState    //!<  ViewState to handle
      , const nvsg::Scene *pScene            //!<  Scene to traverse and handle
      );

    //! Condense all Quads and Triangles of a GeoNode.
    NVSG_API virtual void handleGeoNode( const nvsg::GeoNode *p   //!<  GeoNode to condense
      );

  private:
    bool    areSimilar( const std::valarray<float> &v0, const std::valarray<float> &v1, float eps );
    void    condense( const nvsg::Quads *p );
    void    condense( const nvsg::Triangles *p );
    size_t  hash( const std::valarray<float> &vertex, const std::valarray<float> &base, float scale, size_t numVertices );
    void    postRemapping( const nvsg::GeoSet *p, std::vector<std::valarray<float> > &valuesOut );
    void    prepareRemapping( const nvsg::GeoSet *p, std::vector<std::valarray<float> > &valuesIn );
    bool    remap( const nvsg::IndexSet &iIn, const std::vector<std::valarray<float> > &vIn,
      nvsg::IndexSet &iOut, std::vector<std::valarray<float> > &vOut, float epsilon );

  private:
    float m_epsilon;
    float m_tolerance;
    std::set<const nvsg::Quads *>     m_quads;
    std::set<const nvsg::Triangles *> m_triangles;
  };

  inline  void  CondenseTraverser::setEpsilon( float epsilon )
  {
    m_epsilon = fabsf( epsilon );
  }
}
