// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include "nvsgcommon.h"
#include <list>
#include "nvgl/GLContext.h"
#include "nvtraverser/RenderTraverser.h"
#include "nvsg/GeoSet.h"

namespace nvgl
{
  class GLContext;
  class PBuffer;
}
namespace nvsg
{
  class Texture;
}

namespace nvtraverser
{

// common device context helper data types for Windows and Linux
#if defined(_WIN32)
  typedef HDC   DEVICE_CONTEXT;
  typedef HGLRC RENDER_CONTEXT;      
#elif defined(LINUX)
  typedef Drawable   DEVICE_CONTEXT;
  typedef GLXContext RENDER_CONTEXT;
#endif
  
//! Use this to make sure that the right OpenGL context is current.
#if defined(_WIN32) && defined(_DEBUG)
  #define __ASSERT_CONTEXT() assertContext(); 
#else
  #define __ASSERT_CONTEXT()
#endif

  //! RenderTraverser using OpenGL
  class GLTraverser : public RenderTraverser
  {
    public:
      //! Constructor
      NVSG_API GLTraverser(void);

      //--  Functions specific to GLTraverser --

      //--  Functions implemented from RenderTraverser  --
      NVSG_API virtual JitterTechnique getJitterTechnique( void ) const;
      NVSG_API virtual bool isJitterTechniqueSupported( JitterTechnique jt               //!< Jitter technique.
                                                      ) const;
      NVSG_API virtual bool setJitterTechnique( JitterTechnique jt                       //!< Jitter technique
                                              );
      NVSG_API virtual RenderTechnique getRenderTechnique( void ) const;
      NVSG_API virtual bool isRenderTechniqueSupported( RenderTechnique rt               //! Render technique.
                                                      ) const;
      NVSG_API virtual bool setRenderTechnique( RenderTechnique rt                       //! Render technique.
                                              );
      NVSG_API virtual TransparencyTechnique getTransparencyTechnique( void ) const;
      NVSG_API virtual bool isTransparencyTechniqueSupported( TransparencyTechnique tt   //! Transparency technique.
                                                            ) const;
      NVSG_API virtual bool setTransparencyTechnique( TransparencyTechnique tt           //! Transparency technique.
                                                    );
      NVSG_API virtual OITTechnique getOITTechnique( void ) const;
      NVSG_API virtual bool isOITTechniqueSupported( OITTechnique ot                     //! OIT technique 
                                                   ) const;
      NVSG_API virtual bool setOITTechnique( OITTechnique ot                             //! OIT technique 
                                           );
      NVSG_API virtual size_t getOITMaxPasses( void ) const;
      NVSG_API virtual bool setOITMaxPasses( size_t maxPasses                            //! Maximum passes to render, even when threshold is not reached.
                                           );
      NVSG_API virtual float getOITThreshold( void ) const;
      NVSG_API virtual bool setOITThreshold( float threshold  //!< Threshold [0,1]
                                           );
      NVSG_API virtual GeometryCacheTechnique getGeometryCacheTechnique( void ) const;
      NVSG_API virtual bool isGeometryCacheTechniqueSupported( GeometryCacheTechnique gct  //! GeometryCache technique.
                                                             ) const;
      NVSG_API virtual bool setGeometryCacheTechnique( GeometryCacheTechnique gct          //! GeometryCache technique.
                                                     );
      NVSG_API virtual void getTextureSizeLimits( size_t &width, size_t &height ) const;
      NVSG_API virtual void setTextureSizeLimits( size_t width, size_t height );

      // CachedGeometryCreator interface
      NVSG_API virtual void onReleaseDAL(nvsg::HDAL hDAL);

    protected:
      //! Destructor
      NVSG_API virtual ~GLTraverser(void);

      // doApply override
      NVSG_API virtual void doApply( const nvsg::ViewState *pViewState, const nvsg::Scene *pScene );
     
      // handler routines overrides

      //! Routine to handle a \link nvsg::Transform Transform \endlink node while traversing the scene graph.
      NVSG_API virtual void handleTransform(const nvsg::Transform * p);
      //! Routine to handle a \link nvsg::DirectedLight DirectedLight \endlink node while traversing the scene graph.
      NVSG_API virtual void handleDirectedLight(const nvsg::DirectedLight * p);
      //! Routine to handle a \link nvsg::PointLight PointLight \endlink node while traversing the scene graph.
      NVSG_API virtual void handlePointLight(const nvsg::PointLight * p);
      //! Routine to handle a \link nvsg::SpotLight SpotLight \endlink node while traversing the scene graph.
      NVSG_API virtual void handleSpotLight(const nvsg::SpotLight * p);
      //! Routine to handle a \link nvsg::Quads Quads \endlink object while traversing the scene graph.
      NVSG_API virtual void handleQuads(const nvsg::Quads * p);
      //! Routine to handle a \link nvsg::AnimatedQuads AnimatedQuads \endlink object while traversing the scene graph.
      NVSG_API virtual void handleAnimatedQuads(const nvsg::AnimatedQuads * p);
      //! Routine to handle a \link nvsg::QuadStrips QuadStrips \endlink object while traversing the scene graph.
      NVSG_API virtual void handleQuadStrips(const nvsg::QuadStrips * p);
      //! Routine to handle a \link nvsg::Triangles Triangles \endlink object while traversing the scene graph.
      NVSG_API virtual void handleTriangles(const nvsg::Triangles * p);
      //! Routine to handle a \link nvsg::AnimatedTriangles AnimatedTriangles \endlink object while traversing the scene graph.
      NVSG_API virtual void handleAnimatedTriangles(const nvsg::AnimatedTriangles * p);
      //! Routine to handle a \link nvsg::SkinnedTriangles SkinnedTriangles \endlink object while traversing the scene graph.
      NVSG_API virtual void handleSkinnedTriangles(const nvsg::SkinnedTriangles * p);
      //! Routine to handle a \link nvsg::TriStrips TriStrips \endlink object while traversing the scene graph.
      NVSG_API virtual void handleTriStrips(const nvsg::TriStrips * p);
      //! Routine to handle a \link nvsg::StateSet StateSet \endlink object while traversing the scene graph.
      NVSG_API virtual void handleStateSet(const nvsg::StateSet * p);
      //! Routine to handle a \link nvsg::CgFx CgFx \endlink attribute while traversing the scene graph.
      NVSG_API virtual void handleCgFx(const nvsg::CgFx * p);
      //! Routine to handle a \link nvsg::FaceAttribute FaceAttribute \endlink attribute while traversing the scene graph.
      NVSG_API virtual void handleFaceAttribute(const nvsg::FaceAttribute * p);
      //! Routine to handle a \link nvsg::Material Material \endlink attribute while traversing the scene graph.
      NVSG_API virtual void handleMaterial(const nvsg::Material * p);
      //! Routine to handle a \link nvsg::TextureAttribute TextureAttribute \endlink attribute while traversing the scene graph.
      NVSG_API virtual void handleTextureAttribute(const nvsg::TextureAttribute * p);
      //! Routine to handle a \link nvsg::TextureAttributeItem TextureAttributeItem \endlink object while traversing the scene graph.
      NVSG_API virtual void handleTextureAttributeItem(const nvsg::TextureAttributeItem * p, size_t tu);
      //! Routine to handle a \link nvsg::JitterCamera JitterCamera \endlink object while traversing the scene graph.
      NVSG_API virtual void handleJitterCamera(const nvsg::JitterCamera * p, const nvsg::Node * root);
      //! Routine to handle a \link nvsg::ParallelCamera ParallelCamera \endlink object while traversing the scene graph.
      NVSG_API virtual void handleParallelCamera(const nvsg::ParallelCamera * p, const nvsg::Node * root);
      //! Routine to handle a \link nvsg::PerspectiveCamera PerspectiveCamera \endlink object while traversing the scene graph.
      NVSG_API virtual void handlePerspectiveCamera(const nvsg::PerspectiveCamera * p, const nvsg::Node * root);
      //! Routine to handle a \link nvsg::StereoCamera StereoCamera \endlink object while traversing the scene graph.
      NVSG_API virtual void handleStereoCamera(const nvsg::StereoCamera * p, const nvsg::Node * root);
      //! Routine to handle a \link nvsg::HighlightBox HighlightBox \endlink object while traversing the scene graph.
      NVSG_API virtual void handleHighlightBox(const nvsg::HighlightBox * p, const nvsg::GeoSet * geoSet);
      //! Routine to handle a \link nvsg::HighlightEdges HighlightEdges \endlink object while traversing the scene graph.
      NVSG_API virtual void handleHighlightEdges(const nvsg::HighlightEdges * p, const nvsg::GeoSet * geoSet);

      //! Handles stuff to do between transform stack adjustment and traversal.
      /** The model view matrix stack is already adjusted, so we can load it to the GL-engine.
        * \return result of the base class (RenderTraverser) */
      NVSG_API virtual bool preTraverseTransform( const nvsg::Transform *p);

      //! Handles stuff to do just before the scene is traversed.
      /** This initializes the camera related part of the transform stack (after traversal of the headlights). */
      NVSG_API virtual bool preTraverseRoot( const nvsg::SimpleCamera *p, const nvsg::Node *root);

      //! Handles stuff to do just before a Drawable is traversed.
      /** This tests, if the Drawable is to be drawn regarding the current render states. */
      NVSG_API virtual bool preTraverseDrawable( const nvsg::Drawable *p );

      // This function is called from the RenderTraverser framework whenever a LightSource is to be disabled.
      NVSG_API virtual void disableLightSource(const nvsg::LightSource *p);

#if defined(_DEBUG) && defined(_WIN32)
      void assertContext();
#endif

    protected:
      const nvsg::CgFx      * m_effect;                 // current effect attrib
      bool                    m_resetStateAttribute;    // used with internal state caching

    private:
#if defined(_DEBUG) && defined(_WIN32)
      HGLRC m_hglrc; // only for asserting the right current context
#endif

      void  adjustRenderTechnique( bool on );

      void  initGLState( void );

      void  syncCgFxTweakables( const nvsg::CgFx * p);
      void  syncCgFxTransforms(const nvsg::CgFx * p);

      bool  loadTexture(const nvsg::Texture * pTex, GLuint texHdl);
      void  uploadTexture1D( const nvsg::Texture *p );
      void  uploadTexture2D( const nvsg::Texture *p, unsigned int imageIndex = 0, GLenum target = GL_TEXTURE_2D );
      void  uploadTexture3D( const nvsg::Texture *p );

      void  clearJitterTechnique( void );
      void  initJitterTechnique( void );
      void  initJitterTechniquePBuffer( void );
      void  jitterAccum( const nvsg::JitterCamera *p, const nvsg::Node *root );
      void  jitterPBuffer( const nvsg::JitterCamera *p, const nvsg::Node *root );
      void  jitterReadPixels( const nvsg::JitterCamera *p, const nvsg::Node *root );

      void  clearTransparencyTechnique( void );
      void  initTransparencyTechnique( void );
      void  clearOITTechnique( void );
      void  initOITTechnique( void );
      void  allocOITPBuffers( void );
      void  freeOITPBuffers( void );
      void  allocOITTextures( void );
      void  freeOITTextures( void );
      void  setupOITRendering( void );

      void  prepareHighlight( const nvmath::Vec3f &color, float lineWidth );

      void  freeDeviceDependentData(std::vector<nvsg::HDAL>& data, unsigned int dataID);

      //! Handle the general part of a light source.
      GLenum  glTraverse( const nvsg::LightSource *p        //!<  light source to handle
                        , size_t index                //!<  number of lights on stack
                        );

      // re-routing to base implementation used with glTraverse template functions below
      void handlePerspectiveCameraCallBaseImplementation(const nvsg::PerspectiveCamera * p, const nvsg::Node *root);
      void handleParallelCameraCallBaseImplementation(const nvsg::ParallelCamera * p, const nvsg::Node *root);

      template<typename camera, typename defaultHandler>
      void  glTraverse( const camera *p, const nvsg::Node *root, defaultHandler defHandler );

      template<typename camera, typename defaultHandler>
      void  glTraverseOIT( const camera *p, const nvsg::Node *root, defaultHandler defHandler );


      // select render path to go 
      void selectRenderPath();
      // function pointer via to invoke quad rendering routines below
      void (GLTraverser::*m_pfnProcessQuads)( const nvsg::Quads *p );
      // function pointer via to invoke QuadStrips rendering routines below
      void (GLTraverser::*m_pfnProcessQuadStrips)( const nvsg::QuadStrips *p );
      // function pointer via to invoke triangle rendering routines below
      void (GLTraverser::*m_pfnProcessTriangles)( const nvsg::Triangles *p );
      // function pointer via to invoke TriStrips rendering routines below
      void (GLTraverser::*m_pfnProcessTriStrips)( const nvsg::TriStrips *p );

      // prepare effected GeoSet
      void  GLTraverser::prepareEffect( void );

      // process quads using a default render path 
      void processQuads( const nvsg::Quads *p );
      // process QuadStrips using a default render path
      void processQuadStrips( const nvsg::QuadStrips *p );
      // process triangles using a default render path 
      void processTriangles( const nvsg::Triangles *p );
      // process TriStrips using a default render path
      void processTriStrips( const nvsg::TriStrips *p );
      // common processing on GeoSet
      void processGeoSet( const nvsg::GeoSet *p );

      // draw a highlight box (ignoring current tree states)
      void drawHighlightBox( const nvsg::HighlightBox *box );
      
      // check if the viewport size has changed, and if yes resize
      // our texture or pbuffer objects for OIT
      void resizeOITBuffers();

      struct VBOInfo;
      // process quads using vertex buffer objects (VBO) extension 
      void processQuadsVBO( const nvsg::Quads * p );
      // process QuadStrips using vertex buffer objects (VBO) extension 
      void processQuadStripsVBO( const nvsg::QuadStrips * p );
      // process triangles using vertex buffer objects (VBO) extension 
      void processTrianglesVBO( const nvsg::Triangles * p );
      // process TriStrips using vertex buffer objects (VBO) extension 
      void processTriStripsVBO( const nvsg::TriStrips * p );
      void processGeoSetVBO( const nvsg::GeoSet *p, const VBOInfo* vboInfo );
      VBOInfo * getCachedVBOInfo(const nvsg::GeoSet * p) const;

      struct DLInfo;
      // process quads using OpenGL display lists 
      void processQuadsDL( const nvsg::Quads * p );
      // process QuadStrips using OpenGL display lists 
      void processQuadStripsDL( const nvsg::QuadStrips * p );
      // process triangles using OpenGL display lists 
      void processTrianglesDL( const nvsg::Triangles * p );
      // process TriStrips using OpenGL display lists 
      void processTriStripsDL( const nvsg::TriStrips * p );
      DLInfo * processGeoSetDL( const nvsg::GeoSet *p );
      DLInfo * getCachedDLInfo(const nvsg::GeoSet * p) const;

      const nvsg::Scene     * m_scene;                  // a reference to the currently traversed scene
      bool                    m_isCgFxAware;            // assugns whether the current system is CgFX aware
//      const nvsg::CgFx      * m_effect;                 // current effect attrib

      const nvsg::StateSet  * m_stateSet;
//      bool                    m_resetStateAttribute;    // used with internal state caching
                                                        // need to check this inside the StateAttribute handler routines
      bool                    m_textured;               // the current geometry is textured

      std::list<const nvsg::StateAttribute *>   m_attrList[2];
      std::list<const nvsg::StateAttribute *> * m_currentList;
      std::list<const nvsg::StateAttribute *> * m_nextList;

      int                   m_viewportWidth;
      int                   m_viewportHeight;
      JitterTechnique       m_jitterTechnique;
      float                 m_fraction;                     // independent of special jitter technique
      GLuint                m_dataSize;                     // for read-pixels-based jittering
      GLubyte             * m_imageData;
      GLushort            * m_accumData;
      bool                  m_accumulationInitialized;      // for pbuffer-based jittering
      GLuint                m_fpAccumulate;
      GLuint                m_fpLoad;
      GLuint                m_fpReturn;
      nvgl::PBuffer       * m_fBuffer[2];
      nvgl::PBuffer       * m_jitterPBuffer;

      RenderTechnique       m_renderTechnique;

      // A little helper class for screen door transparency.
      class StipplePatterns
      {
	    public:
        StipplePatterns();
        unsigned char  * getBits(int index);

      private:
        unsigned char  patBytes[ 64 ][ 8 ];
		    unsigned long stippleBits[64][32];

        StipplePatterns( const StipplePatterns &stipplePatterns );
        StipplePatterns& operator=( const StipplePatterns &stipplePatterns );
      };

      TransparencyTechnique m_transparencyTechnique;
      StipplePatterns       m_stipple;                      // for screen door transparency
      bool                  m_transparentPass;              // for blended transparency
      bool                  m_isTransparentMaterial;
      bool                  m_isTransparentTexture;
      OITTechnique          m_oitTechnique;
      size_t                m_oitMaxPasses;
      bool                  m_oitInitialized;
      bool                  m_oitSetupPending;
      int                   m_oitWidth;
      int                   m_oitHeight;
      std::vector<GLuint>   m_oitRGBATextureQueries;
      std::vector<GLuint>   m_oitPBufferQueries;
      std::vector<GLuint> * m_oitOcclusionQueries;
      GLuint                m_fpPeel[2];
      float                 m_oitThreshold;
      size_t                m_oitThresholdPixels;
      GLuint                m_oitDepthTexture;
      GLint                 m_oitDepthFormat;
      std::vector<GLuint>        m_oitRGBATextures;
      std::vector<nvgl::PBuffer*>      m_oitPBuffers;
      GeometryCacheTechnique  m_geometryCacheTechnique;

      // identifier for a certain cached geometry 
      enum
      {
        GL_VBO = RT_GL | 0x01
      , GL_DL  = RT_GL | 0x02
      };
      // cached geometry storage 
      std::vector<nvsg::HDAL> m_cachedGeometry;
      
      enum {
        VERTEX_BUFFER = 0  // index of buffer holding vertex data
      , INDEX_BUFFER       // index of buffer holding indices
      };

      struct VBOInfo : public nvsg::DALData
      {
        VBOInfo(nvsg::DALDataCreator * creator, const nvsg::DALHost * host); 
        ~VBOInfo(); 

        const nvgl::GLContext * context;
        DEVICE_CONTEXT    hdc;            
        RENDER_CONTEXT    hglrc;

        GLuint vBufferID[2];   // buffer ids for vertex and index buffer
        // index buffer related
        GLenum indexType;      // type used with indices (either GL_UNSIGNED_INT or GL_UNSIGNED_SHORT)
        GLsizei numIndices;     // total number of indices
        // multi draw elements related
        GLvoid ** primitiveOffsets;    // per primitive offset into index buffer 
        GLsizei * primitiveIndexCount; // per primitive index count
        // vertex buffer related
        GLsizei numBytes;       // size in bytes of the vertex buffer
        GLsizei stride;         // byte offset between consecutive vertex attribute data of same type
        size_t attribOffsets[nvsg::GeoSet::NUMBER_OF_VERTEX_ATTRIBUTES]; // per vertex attribute offset into vertex buffer
      };

      // predicate that takes a pointer to a VBOInfo objects and decides 
      // whether that VBOInfo is valid for the current render context
      static bool isVBOInfoValid(const GLTraverser::VBOInfo * info);

      struct DLInfo : public nvsg::DALData
      {
        DLInfo(nvsg::DALDataCreator * creator, const nvsg::DALHost * host); 
        ~DLInfo();

        const nvgl::GLContext * context;
        GLuint listID;
      };

      // predicate that takes a pointer to a DLInfo objects and decides 
      // whether that DLInfo is valid for the current render context
      static bool isDLInfoValid(const GLTraverser::DLInfo* info);
      void processDL( DLInfo *dlInfo );

      // identifier for TexInfo data
      enum
      {
        GL_TEX = RT_GL | 0x04
      };
      // DAL data carrying device dependent texture IDs  
      struct TexInfo : public nvsg::DALData
      {
        TexInfo(nvsg::DALDataCreator * creator, const nvsg::DALHost * host, const nvsg::Texture * tex);
        ~TexInfo();
        const nvsg::Texture * texture;
        GLuint textureHandle;
      };

      GLint               m_maxTextureSize;
      size_t              m_textureLimitWidth;
      size_t              m_textureLimitHeight;
      std::vector<nvsg::HDAL> m_textures; // DAL handles to TexInfos will be collected here

      nvmath::Vec3f m_accScaling;   //  accumulated scaling factors for enabling/disabling normalization of normals
  };

  inline  unsigned char *  GLTraverser::StipplePatterns::getBits( int index )
  {
    return( (unsigned char *) &stippleBits[ index ][ 0 ] );
  }
}
