// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include "nvsgcommon.h"

#include "nvtraverser/Traverser.h"
#include "nvutil/TransformStack.h"


namespace nvtraverser
{

//! Traverser that can handle the model view transform stack.
/** Use this class as a base for traversers that need to maintain the
  * current model view transformation. */
class ModelViewTraverser : public Traverser
{
protected:
  //! Protected constructor to prevent instantiation of a ModelViewTraverser.
  /** \note A ModelViewTraverser doesn't change anything in the scene graph, but might be used by a modifying traverser.
    * therefore, it gets the readOnly flag as a parameter to pass to Traverser. */
  NVSG_API ModelViewTraverser( bool readOnly = true );

  NVSG_API virtual void handleLOD( const nvsg::LOD *p );

  //! Protected destructor to prevent instantiation of a ModelViewTraverser.
  NVSG_API virtual ~ModelViewTraverser();

  //! Provide special treatment of a ParallelCamera object.
  /** The function initializes the model view transformation stack and calls the base class handler function. */
  NVSG_API virtual void handleParallelCamera( 
    const nvsg::ParallelCamera *p     //!<  ParallelCamera to handle
  , const nvsg::Node *root            //!<  root node of scene to traverse
  );

  //! Provide special treatment of a PerspectiveCamera object.
  /** The function initializes the model view transformation stack and calls the base class handler function. */
  NVSG_API virtual void handlePerspectiveCamera( 
    const nvsg::PerspectiveCamera *p  //!<  PerspectiveCamera to handle
  , const nvsg::Node *root            //!<  root node of scene to traverse
  );

  //! Provide special treatment of a Transform node.
  /** On a Transform the modelview matrix is modified, the children are traversed, and the modelview matrix is
    * restored. */
  NVSG_API virtual void handleTransform(
    const nvsg::Transform *p          //!<  Transform to handle
  );

  //! Handles actions to take between transform stack adjustment and traversal.
  /** In this base class, this is a NOP.  */
  NVSG_API virtual bool preTraverseTransform( 
    const nvsg::Transform *p          //!< Transform node to traverse next 
  );

  //! Handles actions to take between traversal and transform stack adjustment.
  /** When this function returns true, the subtree beneath is traversed. Otherwise it isn't.
    * \return true */
  NVSG_API virtual void postTraverseTransform(
    const nvsg::Transform *p          //!< Transform node that was traversed immediately before this call. 
  );

  //! Handles actions to take just before the scene is traversed.
  /** This initializes the camera related part of the transform stack (after traversal of the headlights). */
  NVSG_API virtual bool preTraverseRoot( 
    const nvsg::SimpleCamera *p  //!<  camera that is used here
  , const nvsg::Node *root       //!<  root node to traverse from
  );

protected:
  nvutil::TransformStack  m_transformStack; //!<  stack of transformations that holds the actual transformations while traversing a tree
};
}
