// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include "nvsgcommon.h"

#include "nvtraverser/CullTraverser.h"
#include "nvtraverser/OcclusionCullGLTraverser.h"

namespace nvtraverser
{
  //! \c CullTraverser culling on occlusion information.
  /** This \c CullTraverser gathers information on all \c Node s in the tree to determine if it changes any fragment on the
    * screen. All \c Node s (and potentially subtrees) that don't touch any fragment are marked as outside and can be
    * skipped on subsequent traversals. */
  class OcclusionCullTraverser : public CullTraverser
  {
    public:
      //! Constructor
      NVSG_API OcclusionCullTraverser(void);

      // We need this for the contained OcclusionCullGLTraverser
      //
      //! Get the current geoemtry caching technique.
      //** \return The currently used geometry caching technique. */
      NVSG_API RenderTraverser::GeometryCacheTechnique getGeometryCacheTechnique( void ) const;
      //! Test on supported geometry caching techhniques.
      /** \return true - supported, false - not supported.*/
      NVSG_API bool isGeometryCacheTechniqueSupported( RenderTraverser::GeometryCacheTechnique gct  //! GeometryCache technique.
                                                     ) const;
      //! Set the geometry caching technique the render should use.
      /** \return true - everything went fine, false - not set */
      NVSG_API bool setGeometryCacheTechnique( RenderTraverser::GeometryCacheTechnique gct          //! GeometryCache technique.
                                             );
    protected:
      //! Destructor
      NVSG_API virtual ~OcclusionCullTraverser(void);
        
      // doApply override
      NVSG_API virtual void doApply( const nvsg::ViewState *pViewState, const nvsg::Scene *pScene );

      //  Reimplemented from CullTraverser
      NVSG_API virtual nvsg::CullCode cullCode( const nvsg::Drawable *p ) const;
      NVSG_API virtual nvsg::CullCode cullCode( const nvsg::Node *p ) const;

    private:
      OcclusionCullGLTraverser  * m_glTraverser;
      mutable size_t              m_index;
      std::vector<GLuint>       * m_queries;
#if defined( _DEBUG )
      std::vector<const nvsg::Drawable *> * m_drawables;
#endif
  };

  inline RenderTraverser::GeometryCacheTechnique OcclusionCullTraverser::getGeometryCacheTechnique() const
  {
    __ASSERT( m_glTraverser );
    return m_glTraverser->getGeometryCacheTechnique();
  }

  inline bool OcclusionCullTraverser::isGeometryCacheTechniqueSupported(RenderTraverser::GeometryCacheTechnique gct) const
  {
    __ASSERT( m_glTraverser );
    return m_glTraverser->isGeometryCacheTechniqueSupported(gct);
  }

  inline bool OcclusionCullTraverser::setGeometryCacheTechnique(RenderTraverser::GeometryCacheTechnique gct)
  {
    __ASSERT( m_glTraverser );
    return m_glTraverser->setGeometryCacheTechnique(gct);
  }
                                             
}

