// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include <set>

#include "nvsgcommon.h"
#include "nvtraverser/Traverser.h"

namespace nvsg
{
  class StateAttribute;
  class TextureAttributeItem;
}

namespace nvtraverser
{
  //! Traverser that unifies identical StateSets.
  /** Identical StateSets are recognized and linked to the same one.
    * Identical StateAttributes are recognized and linked to the same one.
    * This way, redundant objects are removed, thereby reducing storage space as well as run time. */
  class StateSetUnifyTraverser : public Traverser
  {
    public:
      //! Constructor
      NVSG_API StateSetUnifyTraverser( void );

    public:
      //! Get the 'ignore names' flag.
      /** If the 'ignore names' flag is set, on comparing StateSets and StateAttributes their names are ignored.
        * \return true if the names will be ignored, otherwise false */
      NVSG_API bool getIgnoreNames( void ) const;

      //! Set the 'ignore names' flags.
      /** If the 'ignore names' flag is set, on comparing StateSets and StateAttributes their names are ignored. */
      NVSG_API void setIgnoreNames( bool ignore   //!<  set true to ignore names on comparision
                                  );

    protected:
      //! Protected destructor to prevent instantiation of a StateSetUnifyTraverser on stack.
      NVSG_API virtual ~StateSetUnifyTraverser( void );

      //! Routine to handle a \link nvsg::CgFx CgFx \endlink object while traversing the scene graph.
      /** After traversing the CgFx \a p, this handler compares it to the other previously encountered CgFx objects.
        * If there is another identical CgFx object, \a p is marked to be replaced by that in the handleStateSet routine. */
      NVSG_API virtual void handleCgFx( const nvsg::CgFx *p             //!< Points to the currently visited CgFx object.
                                      );

      //! Routine to handle a \link nvsg::GeoNode GeoNode \endlink object while traversing the scene graph.
      /** After traversing the GeoNode \ p, this handler replaces all StateSet objects in it's geometries that have been
        * marked for replacement in the handleStateSet routine. */
      NVSG_API virtual void handleGeoNode( const nvsg::GeoNode *p   //!< Points to the currently visited GeoNode object.
                                         );

      //! Routine to handle a \link nvsg::FaceAttribute FaceAttribute \endlink object while traversing the scene graph.
      /** After traversing the FaceAttribute \a p, this handler compares it to the other previously encountered FaceAttribute
        * objects. If there is another identical FaceAttribute object, \a p is marked to be replaced by that in the handleStateSet
        * routine. */
      NVSG_API virtual void handleFaceAttribute( const nvsg::FaceAttribute *p //!< Points to the currently visited FaceAttribute object.
                                               );

      //! Routine to handle a \link nvsg::Material Material \endlink object while traversing the scene graph.
      /** After traversing the Material \a p, this handler compares it to the other previously encountered Material
        * objects. If there is another identical Material object, \a p is marked to be replaced by that in the handleStateSet
        * routine. */
      NVSG_API virtual void handleMaterial( const nvsg::Material *p     //!< Points to the currently visited Material object.
                                          );

      //! Routine to handle a \link nvsg::StateSet StateSet \endlink object while traversing the scene graph.
      /** After traversing the StateSet \a p, this handler replaces a CgFx, Material, FaceAttribute, and TextureAttribute,
        * that has been marked for replacement in the corresponding handler. Then \a p is compared to the other
        * previously encountered StateSet objects. If there is another identical StateSet object, \a p is marked to be
        * replaced by that in the handleGeoNode routine. */
      NVSG_API virtual void handleStateSet( const nvsg::StateSet * p    //!< Points to the currently visited StateSet object.
                                          );

      //! Routine to handle a \link nvsg::TextureAttribute TextureAttribute \endlink object while traversing the scene graph.
      /** After traversing the TextureAttribute \a p, this handler all it's TextureAttributeItems, that has been marked
        * for replacement in the handleTextureAttributeItem routine. Then \a p is compared to the other previously
        * encountered TextureAttribute objects. If there is another identical TextureAttribute object, \a p is marked to be
        * replaced by that in the handleSateSet routine. */
      NVSG_API virtual void handleTextureAttribute( const nvsg::TextureAttribute *p //!< Points to the currently visited TextureAttribute object.
                                                  );

      //! Routine to handle a \link nvsg::TextureAttributeItem TextureAttributeItem \endlink object while traversing the scene graph.
      /** After traversing the TextureAttributeItem \a p, this handler compares it to the other previously encountered
        * TextureAttributeItem objects. If there is another identical TextureAttributeItem object, \a p is marked to be replaced
        * by that in the handleTextureAttribute routine. */
      NVSG_API virtual void handleTextureAttributeItem( const nvsg::TextureAttributeItem *p //!< Points to the currently visited TextureAttributeItem object.
                                                      , size_t textureUnit            //!< Texture unit to use with this TextureAttributeItem
                                                      );

    private:
      const nvsg::CgFx                            * m_cgfx;
      std::set<const nvsg::CgFx*>                   m_cgfxs;
      const nvsg::FaceAttribute                   * m_faceAttribute;
      std::set<const nvsg::FaceAttribute*>          m_faceAttributes;
      const nvsg::Material                        * m_material;
      std::set<const nvsg::Material*>               m_materials;
      std::vector<std::pair<const nvsg::StateSet*,const nvsg::StateSet*> > m_stateSet;
      std::set<const nvsg::StateSet*>               m_stateSets;
      const nvsg::TextureAttribute                * m_textureAttribute;
      std::set<const nvsg::TextureAttribute*>       m_textureAttributes;
      std::vector<std::pair<size_t,const nvsg::TextureAttributeItem*> > m_textureAttributeItem;
      std::set<const nvsg::TextureAttributeItem*>   m_textureAttributeItems;

      bool  m_ignoreNames;
  };

  inline bool StateSetUnifyTraverser::getIgnoreNames( void ) const
  {
    return( m_ignoreNames );
  }

  inline void StateSetUnifyTraverser::setIgnoreNames( bool ignore )
  {
    m_ignoreNames = ignore;
  }

}
