// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#include <list>

#include "nvsgcommon.h"

#include "nvtraverser/Traverser.h"

namespace nvtraverser
{
  //! Traverser that unifies identical Transforms.
  /** Identical Transforms (only siblings!) are recognized and their children are gathered beneath a single Transform.
    * This way, redundant objects are removed, thereby reducing storage space as well as run time.
    * \note When there is a LightSource inbetween of some Transform nodes, unifying is not allowed (and therefore not
    * performed) across this LightSource. */
  class TransformUnifyTraverser : public Traverser
  {
    public:
      //! Constructor
      NVSG_API TransformUnifyTraverser( void );

    public:
      //! Get the 'ignore names' flag.
      /** If the 'ignore names' flag is set, the names of the Transforms are ignored.
        * \return true if the names will be ignored, otherwise false */
      NVSG_API bool getIgnoreNames( void ) const;

      //! Set the 'ignore names' flags.
      /** If the 'ignore names' flag is set, the names of the Transforms are ignored. */
      NVSG_API void setIgnoreNames( bool ignore   //!<  set true to ignore names on comparison
                                  );

    protected:
      //! Protected destructor to prevent instantiation of a StateSetUnifyTraverser on stack.
      NVSG_API virtual ~TransformUnifyTraverser( void );

      //! doApply override
      NVSG_API virtual void doApply( const nvsg::ViewState *viewState, const nvsg::Scene *scene );

      //! Routine to handle a \link nvsg::Transform Transform \endlink node while traversing the scene graph.
      /** This handler compares the Transform \a p with previously encountered Transforms. If it is identical to one of them
        * all it's children are moved to that other Transform and \a p is marked for deletion. */
      NVSG_API virtual void handleTransform( const nvsg::Transform * p //!< Points to the currently visited Transform object.
                                           );

      //! Routine to handle a \link nvsg::AnimatedTransform AnimatedTransform \endlink node while traversing the scene graph.
      /** This handler compares the AnimatedTransform \a p with previously encountered AnimatedTransforms. If it is
        * identical to one of them all it's children are moved to that AnimatedTransform and \a p is marked for deletion. */
      NVSG_API virtual void handleAnimatedTransform( const nvsg::AnimatedTransform * p //!< Points to the currently visited AnimatedTransform object.
                                                   );

      //! This function will be called from the framework immediately before a Group is to be traversed.
      /** \remarks Custom traversers can override this function for any tasks that need to be done immediately 
        * before the Group pointed to by \a p is traversed. */
      NVSG_API virtual void preTraverseGroup( 
        const nvsg::Group *p  //!< Points to the Group that will be traversed immediately after this call.
      );

      //! This function will be called from the framework immediately after a Group has been traversed.
      /** \remarks Custom traversers can override this function for any tasks that need to be done immediately 
        * after the Group pointed to by \a p has been traversed. */
      NVSG_API virtual void postTraverseGroup( 
        const nvsg::Group *p  //!< Points to the Group that was traversed immediately before this call.
      );

    private:
      bool                                                    m_ignoreLights;
      bool                                                    m_ignoreNames;
      std::stack<std::list<const nvsg::Transform*>*>          m_toDeleteTransform;
      std::stack<std::list<const nvsg::Transform*>*>          m_toKeepTransform;
      std::stack<std::list<const nvsg::AnimatedTransform*>*>  m_toDeleteAnimatedTransform;
      std::stack<std::list<const nvsg::AnimatedTransform*>*>  m_toKeepAnimatedTransform;
  };

  inline bool TransformUnifyTraverser::getIgnoreNames( void ) const
  {
    return( m_ignoreNames );
  }

  inline void TransformUnifyTraverser::setIgnoreNames( bool ignore )
  {
    m_ignoreNames = ignore;
  }

}
