// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#ifndef LINUX

#pragma once

#include "nvsgcommon.h"

#include "nvgl/GLContext.h"
#include "nvui/RenderArea.h"

/** \file */

namespace nvui
{
  //! OpenGL render area
  /** This class is responsible for MS Windows specific OpenGL functionality.
    * It sets up and destroys the OpenGL context, handles the creation, 
    * pixelformat, refresh, mouse input, etc. */    
  class GLWinRenderArea : public RenderArea
  {
    public:
      //! Default constructor.
      NVSG_API GLWinRenderArea(void);
      //! Default destructor.
      NVSG_API virtual ~GLWinRenderArea(void);

      //! Attach a window to the render area.
      /** You have to create an OpenGL capable window and attach it before calling init(),
          or you can derive from this class to integrate the functionality.  */          
      NVSG_API virtual bool attachWindow( HWND hWnd  //!< The window handle of an existing window
                                        );
      //! Create the render area.
      /** This method creates and initializes the render area.
        * \return true - everything went fine
        * \note Create and attach a valid Window before calling this function. */
      NVSG_API virtual bool init( RenderArea * shareArea = NULL  //!< Pointer to a render area with which to share display lists, textures, vertex programs, etc.
                                );
      //! Destroy the render area.
      /** This method destroys and cleans up the render area. */
      NVSG_API virtual void destroy( void );
      
      //! Make OpenGL context current.
      /** This is a more clever way to make the context current. 
        * This method minimizes wglMakeCurrent calls. 
        * \note Use only this method instead of wglMakeCurrent 
        * so we save time. */
      NVSG_API void  makeCurrent( void );

      //! Handle mouse button down.
      /** React on mouse button down events (for example, report it to the current camera manipulator). */
      NVSG_API virtual void mouseButtonDown( UINT flags //!< Concatenation of pressed buttons (Windows: MK_LBUTTON, MK_MBUTTON, MK_RBUTTON
                                           , int x      //!< Current x position of the mouse
                                           , int y      //!< Current y position of the mouse
                                           );
      //! Handle mouse button up.
      /** React on mouse button up events (for example, report it to the current camera manipulator. */
      NVSG_API virtual void mouseButtonUp( UINT flags   //!< Concatenation of released buttons (Windows: MK_LBUTTON, MK_MBUTTON, MK_RBUTTON
                                         , int x        //!< Current x position of the mouse
                                         , int y        //!< Current y position of the mouse
                                         );
      //! Handle mouse motion.
      /** React on mouse motion events (for example, report it to the current camera manipulator. */
      NVSG_API virtual void mouseMotion( int x  //!< Current x position of the mouse
                                       , int y  //!< Current y position of the mouse
                                       );

      //! Handle the viewport resizing
      /** Resize the viewport. React on parent window size changes. 
        * Report to OpenGL, to the render window, or to the camera manipulator. */
      NVSG_API virtual void setViewportSize( size_t x  //!< Width of the viewport
                                           , size_t y  //!< Height of the viewport
                                           );

      //! Share the OpenGL lists
      /** Share display lists, textures, vertex programs, etc. 
        * with another rendering context.
        * \return true - everything went fine */
      NVSG_API bool shareLists( GLWinRenderArea* ra ); 

      //! Render the scene.
      /** Do the rendering of the entire scene. (app, cull, draw the whole tree) 
        * If you want to trigger this method, use TriggerRedraw() to force a redraw 
        * of the scene. */
      NVSG_API virtual void renderScene( void );  
      
      //!< Trigger a redraw of the scene
      /** \note Use this method to trigger a redraw 
        * instead of calling the drawing method directly! */
      NVSG_API virtual void triggerRedraw( void );       
      
    protected:
      HGLRC     m_hglrc;  //!< Our OpenGL context
      HDC       m_hdc;    //!< Our device context

      //! Initialize additional OpenGL features.
      /** This is a handy method that is called from create. 
        * Deriving from this class simplifies the initialization of  
        * special features like HW swap locking, etc. 
        * \return true - everything went fine */
      NVSG_API virtual bool  initGL( void );

      //! Get the Pixelformat index.
      /** Determine the OpenGL pixelformat index with respect to the 
        * requested features, such as stereo, multisample, etc.. 
        * \return Pixel format index. Error: 0 */
      NVSG_API UINT  determinePixelFormat( void );
    
  private:
      HWND      m_hWindow;   //!< Our Window handle
};

}

#endif //LINUX