// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include "nvsgcommon.h"

#include "nvui/CameraManipulator.h"

namespace nvui
{  
 
  //!Walk Camera Manipulator Class.
  /** This manipulator is a special CameraManipulator that interprets/converts mouse movement
    * into walk-through camera movements.\n
    * \n
    * <b>Keys & Buttons </b>                        \n
    * Left mouse button           : Walk forward.   \n
    * Middle mouse button         : Walk backward.  \n
    * W-key                       : Increase speed. \n
    * S-key                       : Decrease speed. \n
    * Ctrl + Left mouse button    : Tilt.           \n
    * Ctrl + Middle mouse button  : Pan.            \n
    * \n
    * \note If you use this manipulator your application must use a mechanism (such as a small thread  
    * that triggers redraw, etc. ) that provides continuous camera updates when walking forward or backward. \n
    * \n
    * \note If you implement your own manipulator, refer to the button state 
    * configuration to see how to add additional button combinations in the same way.
    */
  class WalkCamManip : public CameraManipulator
  {
    public:
      //!Default Constructor
      NVSG_API WalkCamManip(void);
      //!Default Destructor
      NVSG_API ~WalkCamManip(void);

      //! Additional keys for walking.
      enum
      { _W_KEY        = 0x0400 //!< Speed up.
      , _S_KEY        = 0x0800 //!< Speed up.
      };

      //! Handle key repeats.
      /** Adapt key repeats. */
      NVSG_API virtual void keyRepeat();

      //! Apply the states to the camera.
      /** Recalculate the camera with respect to the manipulator states. */
      NVSG_API virtual void apply();
            
      //! Signals if the manipulator is in \e walk mode
      /** This method compares the current button or key state 
        * with the configured button states 
        * of this object. e.g. TRANSLATE1
        * \return true: manipulator is in walk mode
        */
      bool isWalk(void);

      //! Signals if the manipulator is in \e translation mode
      /** This method compares the current button or key state 
        * with the configured button states 
        * of this object. e.g. TRANSLATE1
        * \return true: manipulator is in translation mode
        */
      bool isTranslate(void);

      //! Signals if the manipulator is in \e tilting mode
      /** This method compares the current button or key state 
        * with the configured button states 
        * of this object. e.g. TRANSLATE1
        * \return true: manipulator is in tilting mode
        */
      bool isTilt(void);
      
      //! Set the screen size.
      /** Set the screen resolution. The resolution is needed for the 
        * sensitivity of the manipulator. Default is 1280x1024. */
      void setScreenSize( UINT x  //!< width of the screen (pixels) -  must not be 0 
                        , UINT y  //!< height of the screen (pixels) -  must not be 0
                        );

      //! Set the world up-vector.
      /** Define where the 'sky' is. */
      void setWorldUpVector( nvmath::Vec3f up //!< World up-vector
                           );
      //! Get the world up-vector.
      /** \return Vector of the world pointing to the 'sky' */
      const nvmath::Vec3f& getWorldUpVector();

      //! Define speed increase factor.
      /** The current speed will be multiplied or divided by this factor */
      void setSpeedIncreaseFactor(float factor   //!< Speed increase factor - default 2.f \note This factor must not be 0!
                                 );
      
      //! Get the speed increase factor.
      /** \return  Speed increase factor.*/
      float getSpeedIncreaseFactor();

      //! Define horizontal sensitivity.
      /** The horizontal sensitivity is a factor that influences how 
        * sensitive the manipulator reacts to horizontal rotation changes */
      void setHSensitivity(float sensitivity //!< Sensitivity factor - default 1.f
                          );
      
      //! Get the horizontal sensitivity.
      /** \return  Horizontal sensitivity factor.*/
      float getHSensitivity();

      //! Set the speed.
      /** This parameter controls the speed of the camera movement */
      void setSpeed(float speed //!< Speed value - default 1.f
                   );

      //! Get the movement speed.
      /** \return  Speed value.*/
      float getSpeed();

  protected:
      UINT  m_ScreenSizeX; //!< width of the screen (pixels)
      UINT  m_ScreenSizeY; //!< height of the screen (pixels)
      
      nvmath::Vec3f m_worldUpVector;  //!< Up-vector of the world.

      float m_speedIncreaseFactor;  //!< Factor that influences the stepping. Default: 2.0 NOTE: This factor must not be 0!
      float m_hSensitivity;         //!< Factor that influences the horizontal rotation. Default: 1.0
      
      float m_speed; //!< Walk speed. Controlled via key or button input. Default: 1.0

      static NVSG_API const UINT WALKFORWARD;   //!< Key/button configuration walking.
      static NVSG_API const UINT WALKBACKWARD;  //!< Key/button configuration walking.
      static NVSG_API const UINT TRANSLATE;     //!< Key/button configuration for translating up, down, left and right.
      static NVSG_API const UINT TILT;          //!< Key/button configuration looking left, right, up and down.
      static NVSG_API const UINT INCREASESPEED; //!< Key/button configuration speed up.
      static NVSG_API const UINT DECREASESPEED; //!< Key/button configuration slow down.

      static NVSG_API const UINT AREAOFPEACE;   //!< Size of the area where the mouse does not trigger the manipulator.

      //! Map your keys.
      /** \return Manipulator key code */
      NVSG_API virtual UINT convertKey(UINT key  //!< External key code.
                                      );

      //! Signals if the manipulator is in \e forward walking mode
      /** This method compares the current button or key state 
        * with the configured button states 
        * of this object. e.g. TRANSLATE1
        * \return true: manipulator is in forward walking mode
        */
      bool isWalkForward( void );
      
      //! Signals if the manipulator is in \e backward walking mode
      /** This method compares the current button or key state 
        * with the configured button states 
        * of this object. e.g. TRANSLATE1
        * \return true: manipulator is in backward walking mode
        */
      bool isWalkBackward( void );

      //! Do the camera walk.
      /** Walk forward or backward. You can simply 
        * replace it with your own way to handle this. */
      NVSG_API virtual void walk( void );
      
      //! Do the camera translate.
      /** Translate the camera perpendicular to the 
        * movement plane. (up, down, left and right)
        * You can simply replace it with your own way to handle this. */
      NVSG_API virtual void translate( void );
      
      //! Do the camera tilt.
      /** Tilt the camera. You rotate the camera in place so you 
        * can look at your feet, up into the sky, or rotate left and right.
        * You can simply replace it with your own way to handle this. */
      NVSG_API virtual void tilt( void );
  };

  inline bool WalkCamManip::isWalk( void )
  {
    return( isWalkForward() || isWalkBackward() );
  }

  inline bool WalkCamManip::isWalkForward( void )
  {
    return(WALKFORWARD == m_buttonState && (  m_keyState == 0
                                           || m_keyState == INCREASESPEED 
                                           || m_keyState == DECREASESPEED));
  }

  inline bool WalkCamManip::isWalkBackward( void )
  {
    return(WALKBACKWARD == m_buttonState && (  m_keyState == 0 
                                            || m_keyState == INCREASESPEED 
                                            || m_keyState == DECREASESPEED));
  }

  inline bool WalkCamManip::isTranslate( void )
  {
    return( TRANSLATE == ((m_buttonState | m_keyState)) );
  }

  inline bool WalkCamManip::isTilt( void )
  {
    return( TILT == ((m_buttonState | m_keyState)) );
  }

  inline void WalkCamManip::setScreenSize(UINT x, UINT y)
  {
    __ASSERT(x>0);
    __ASSERT(y>0);
    m_ScreenSizeX = x;
    m_ScreenSizeX = y;
  }

  inline void WalkCamManip::setWorldUpVector( nvmath::Vec3f up )
  {
    m_worldUpVector = up; 
  }

  inline const nvmath::Vec3f& WalkCamManip::getWorldUpVector()
  {
    return m_worldUpVector; 
  }

  inline void WalkCamManip::setSpeedIncreaseFactor(float factor)
  {
    __ASSERT(fabs(factor) > FLT_EPSILON);
    m_speedIncreaseFactor = factor; 
  }
  
  inline float WalkCamManip::getSpeedIncreaseFactor()
  {
    return m_speedIncreaseFactor; 
  }

  inline void WalkCamManip::setHSensitivity(float sensitivity)
  {
    m_hSensitivity = sensitivity; 
  }
  
  inline float WalkCamManip::getHSensitivity()
  {
    return m_hSensitivity; 
  }

  inline void WalkCamManip::setSpeed(float speed)
  {
    m_speed = speed; 
  }

  inline float WalkCamManip::getSpeed()
  {
    return m_speed;
  }
} //  namespace nvui
