// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

// rcobject.h
//
// Support for reference counted object
//
#pragma once
/** \file */

#include "nvsgcommon.h"

#include "nvutil/Allocator.h" // for IAllocator interface

// memory leak detection
#ifdef _DEBUG
# ifndef new
#  define new new(__FILE__, __LINE__)
#  define _DEFINED_DBGNEW // restrict the 'new' macro to this header file only
# endif
#endif

namespace nvutil
{
  //! A reference counted object interface
  /** Provides an interface for managing reference counted objects. \c RCObject inherits
    * from \c IAllocator to utilize an optimized heap manager for small object allocation.
    * \par 
    * The purpose of a reference counted object is to share its data with other reference counted 
    * objects as long as this is possible. If a certain reference counted bject is constructed from 
    * another reference counted bject, it will initially share its data with that source object, 
    * same with assignment. That sharing can be sustained as long as non of the objects will be modified.
    *
    * \note A \c RCObject can only be constructed on heap. The compiler will reject any
    * attempt to construct a \c RCObject on stack.
    *
    * \note It is prohibited to explicitely delete a \c RCObject by calling \c delete on a pointer
    * to a \c RCObject received from a previous call to \c new. The compiler will complain, if any 
    * client code attempts to do so. If client code creates a \c RCObject, it must increment the 
    * object's reference count (\c addRef) before using it and decrement the object's reference 
    * count (\c removeRef) after usage. The object will be automatically deleted, if it's reference 
    * count reaches zero.
    *
    * \note The reference count of a newly created \c RCObject initially is zero.
    * 
    */
  class RCObject : public IAllocator
  {
    public:
      //! Default constructor
      /** Initializes the reference count to zero. Newly created RCObjects are shareable by default. */
      RCObject() : m_refcnt(0), m_shareable(true) {};

      //! Increment reference count
      /** For client code to ensure that the data of a reference counted object is valid as long as it
        * uses the object, it should first increment the objects reference count. After usage, client
        * code should decrement the objects reference count to avoid resource leaks.
        *
        * \note The reference count of a newly created \c RCObject initially is zero.
        */
      NVSG_API void addRef()            const; 
      //! Decrement reference count
      /** \note It is prohibited to explicitely delete a \c RCObject. 
        * The object will be automatically deleted, if reference count yields zero.
        */
      NVSG_API void removeRef()         const;  
      //! Mark the object as not shareable
      /** If a \c RCObject is marked as not shareable, it will never share its data with other \c RCObjects.
        * \note An \c RCObject initially is shareable.
        */
      NVSG_API void markUnshareable();
      //! Query for the current shareable state
      /** \note An \c RCObject initially is shareable.
        * \returns \c true if shareable, \c false otherwise.
        */
      NVSG_API bool isShareable()       const; 
      //! Get to know if the object is currently shared
      /** If a \c RCObject has been constructed from a shareable \c RCObject, 
        * it initially shares its data with that object.
        * \returns \c true if currently shared, \c false otherwise.
        */
      NVSG_API bool isShared()          const; 

    protected:

      //! Protected copy constructor
      /** Reference counted objects need to live on heap, so we prohibit construction on stack by making constructors
        * and assignment protected. */
      RCObject(const RCObject& rhs) : m_refcnt(0), m_shareable(true) {}

      //! Protected assignment operator
      /** Reference counted objects need to live on heap, so we prohibit construction on stack by making constructors
        * and assignment protected. */
      RCObject& operator=(const RCObject& rhs) { return *this; }

      //! Protected destructor
      /** To prohibit explicit deletion of reference counted objects, the destructor is protected. Reference counted
        * objects cannot deleted other than by being unreferenced with removeRef(). */
      virtual ~RCObject() {}

    private:
      /** m_refcnt must be mutable, so we can also call addRef() / removeRef() from const pointers. */
      mutable int m_refcnt;
      bool m_shareable;
  };

  //! A reference counted vector of elements.
  /** A RCVector is a template class publicly derived from both std::vector<> and RCObject. It is used to hold a
    * reference counted vector of elements of any type. It is used for example in \c Quads for the faces or in \c
    * TriStrips for the strip indices.  */
  template<typename T>
  class RCVector : public std::vector<T>, public RCObject
  {
    public:
      //! Default Constructor
      RCVector()
      {
      }

      //! Constructor
      /** the vector of elements is sized to \a n. */
      RCVector( size_t n      //!<  size of vector
              )
        : std::vector<T>(n)
      {
      }
  };

  //! A smart pointer template for reference counted objects
  /** Offers automatic reference counting through utilizing the \c RCObject interface. 
    * Like a smart pointer this can be used as an ordinary built-in pointer.
    * 
    * \note \c T must be convertible to RCObject. The compiler will reject any attempt
    * to instantiate a \c RCPtr object for types other than RCObject convertible
    * types.
    */
  template <typename T>
  class RCPtr
  {
    //! Initiate write access
    /** This must be called prio to modify the underlying data of the specified \c RCPtr
      * object.
      *
      * \remarks This has been introduced to manually decide wether the 
      * underlying data really gets modified by a non-const access.
      * This helps not to clone every shared object, when a non-const access 
      * occurs but the underlying data remains untouched. 
      * With the introduction of \c writeAccess, full automatic seceding a
      * shared object with a non-const access has been removed. To nevertheless 
      * maintain correctness with shared data, one have to explicitely initiate 
      * every modifying code with a call to \c writeAccess on the corresponding 
      * \c RCPtr object. 
      *
      * \note As \c U is used as a template parameter for \c RCPtr, \c U must be
      * convertible to \c RCObject.
      */
    template <typename U>
    friend void writeAccess( RCPtr<U> &rcPtr //!< The \c RCPtr object, for which write access should be initiated.
                          );
   
    public: 
      explicit RCPtr(T* pRCObj=0);              //!< Construct a \c RCPtr object from a pointer to a \c RCObject.
      RCPtr(const RCPtr& rhs);                  //!< Copy constructor.
      ~RCPtr();                                 //!< Destructor.
      RCPtr<T>& operator=(const RCPtr<T>& rhs); //!< Assigning a \c RCPtr object.
      
      T* operator->();                //!<  Member selection operator to make this behave like a dump pointer.
      const T* operator->() const;    //!<  Constant member selection operator to make this behave like a dump pointer.
      T& operator*();                 //!<  Pointer dereference operator to make this behave like a dump pointer.
      const T& operator*() const;     //!<  Constant pointer dereference operator to make this behave like a dump pointer.

    private:
      void init();                              //!< initialize a RCPtr<T> object
      void clone();                             //!< deep copy the embedded object

      // the dump pointer
      T * m_ptr;
  };

  //! A smart pointer template for reference counted objects.
  /** For objects we have no access to their sources we need an additional level of indirection
    * because we can't make them publicly inherit from RCObject.
    * The 'I' stands for 'indirect'.
    */
  template <typename T>
  class RCIPtr
  {
    //! Initiate write access
    /** This must be called prio to modify the underlying data of the specified \c RCPtr
      * object.
      *
      * \remarks This has been introduced to manually decide wether the 
      * underlying data really gets modified by a non-const access.
      * This helps not to clone every shared object, when a non-const access 
      * occurs but the underlying data remains untouched. 
      * With the introduction of \c writeAccess, full automatic seceding a
      * shared object with a non-const access has been removed. To nevertheless 
      * maintain correctness with shared data, one have to explicitely initiate 
      * every modifying code with a call to \c writeAccess on the corresponding 
      * \c RCPtr object. 
      *
      * \note As \c U is used as a template parameter for \c RCPtr, \c U must be
      * convertible to \c RCObject.
      */
    template <typename U>
    friend void writeAccess( RCIPtr<U> &rcPtr //!< The \c RCIPtr object, for which write access should be initiated.
                          );

    public:
      explicit RCIPtr(T* pObj=0);                 //!< Construct a \c RCIPtr object from a pointer to an object.
      RCIPtr(const RCIPtr& rhs);                  //!< Copy constructor.
      ~RCIPtr();                                  //!< Destructor.
      RCIPtr<T>& operator=(const RCIPtr<T>& rhs); //!< Assigning a \c RCIPtr object.
      
      T* operator->();                //!<  Member selection operator to make this behave like a dump pointer.
      const T* operator->() const;    //!<  Constant member selection operator to make this behave like a dump pointer.
      T& operator*();                 //!<  Pointer dereference operator to make this behave like a dump pointer.
      const T& operator*() const;     //!<  Constant pointer dereference operator to make this behave like a dump pointer.

    private:
      // CountHolder acts as an additional indirection and holds
      // a dump pointer to the opject beeing reference counted
      struct CountHolder : public RCObject
      {
        // destructor only gets called when the objects 
        // reference count reaches zero
        ~CountHolder() { delete m_ptr; }
        T * m_ptr;
      };

      CountHolder * m_cntHolder;

      void init();
      void clone(); // copy on write
  };

  //! AutoPtr for reference counted objects
  /** Enables us to be exception safe on file load time.
    */
  template <typename T>
  class RCObject_AutoPtr
  {
    template<typename U> 
    friend class RCObject_AutoPtr;  

    public:
      explicit RCObject_AutoPtr(T * ptr=0);               //!< Construct a \c RCObject_AutoPtr object from a pointer.
      template <typename U>
      RCObject_AutoPtr(const RCObject_AutoPtr<U>& rhs);   //!< Copy constructor.
      ~RCObject_AutoPtr();                                //!< Destructor.
      template <typename U>
      RCObject_AutoPtr<T>& operator=(const RCObject_AutoPtr<U>& rhs); //!< Assigning a \c RCObject_AutoPtr object.

      T * get() const;            //!<  Equivalent to the selection operator
      T * operator->() const;     //!<  Member selection operator to make this behave like a dump pointer.
      T&  operator*() const;      //!<  Pointer dereference operator to make this behave like a dump pointer.
      T * release();              //!<  Release the pointer to the object.
      void reset(T * ptr);        //!<  Reset the pointer to the object.

    private:
      void deletePointer();
      T * m_ptr;
  };

  //----------------------------------------------------------------------------
  // inline implementations following
  //----------------------------------------------------------------------------

  inline void RCObject::addRef() const 
  { 
    ++m_refcnt; 
  } 

  inline void RCObject::removeRef() const
  { 
    __ASSERT(m_refcnt>0); 
    if (!--m_refcnt) 
    { 
      delete this; 
    } 
  }

  inline void RCObject::markUnshareable() 
  { 
    m_shareable=false; 
  }

  inline bool RCObject::isShareable() const 
  { 
    return m_shareable==true; 
  }

  inline bool RCObject::isShared() const 
  { 
    return m_refcnt>1; 
  }
  //>>
  template <typename T>
  inline RCPtr<T>::RCPtr(T* pRCObj) 
  : m_ptr(pRCObj) 
  { 
    init(); 
  }

  template <typename T>
  inline RCPtr<T>::RCPtr(const RCPtr& rhs) 
  : m_ptr(rhs.m_ptr) 
  { 
    init(); 
  }

  template <typename T>
  inline RCPtr<T>::~RCPtr() 
  { 
    if ( m_ptr ) 
    { 
      m_ptr->removeRef(); 
    } 
  }

  template <typename T>
  inline RCPtr<T>& RCPtr<T>::operator=(const RCPtr<T>& rhs) 
  {
    if ( m_ptr != rhs.m_ptr )
    {
      // remove reference to former value first
      m_ptr->removeRef();
      m_ptr = rhs.m_ptr;
      init();
    }
    return *this;
  }
    
  template <typename T>
  inline T* RCPtr<T>::operator->() 
  { 
    return m_ptr; 
  }

  template <typename T>
  inline const T* RCPtr<T>::operator->() const 
  { 
    return m_ptr; 
  }

  template <typename T>
  inline T& RCPtr<T>::operator*() 
  { 
    return *m_ptr; 
  }

  template <typename T>
  inline const T& RCPtr<T>::operator*() const 
  { 
    return *m_ptr; 
  }

  template <typename T>
  inline void RCPtr<T>::init()
  {
    if ( !m_ptr ) return;
    
    if ( !m_ptr->isShareable() )
    {
      // deep copy required!
      m_ptr = new T(*m_ptr);
    }
    m_ptr->addRef();
  }

  template <typename T>
  inline void RCPtr<T>::clone()
  {
    if ( m_ptr->isShared() )
    { 
      // need to wedge off an own copy if original data is shared 
      T * pOldVal = m_ptr; // remember the old object pointer for deep copying below
      m_ptr->removeRef(); // unreference the object before beeing new assigned
      m_ptr = new T(*pOldVal); // deep copy -> T(const T&) required!
      m_ptr->addRef();
    }
  }
  //>>
  template <typename T>
  inline RCIPtr<T>::RCIPtr(T * pObj)
  : m_cntHolder(new CountHolder)
  {
    m_cntHolder->m_ptr = pObj;
    init();
  }

  template <typename T>
  inline RCIPtr<T>::RCIPtr(const RCIPtr<T>& rhs)
  : m_cntHolder(rhs.m_cntHolder)
  {
    init();
  }

  template <typename T>
  inline RCIPtr<T>::~RCIPtr()
  {
    m_cntHolder->removeRef();
  }

  template <typename T>
  inline RCIPtr<T>& RCIPtr<T>::operator=(const RCIPtr<T>& rhs)
  {
    if ( m_cntHolder != rhs.m_cntHolder )
    {
      m_cntHolder->removeRef(); // unreference the object before beeing new assigned
      m_cntHolder = rhs.m_cntHolder;
      init();
    }
    return *this;
  }

  template <typename T>
  inline const T* RCIPtr<T>::operator->() const
  {
    // const access, no need for copy on write here
    return m_cntHolder->m_ptr;
  }

  template <typename T>
  inline T* RCIPtr<T>::operator->() 
  {
    // perform copy on write with every non-const access to the pointed-to object
    // clone();
    return m_cntHolder->m_ptr;
  }

  template <typename T>
  inline const T& RCIPtr<T>::operator*() const
  {
    // const access, no need for copy on write here
    return *(m_cntHolder->m_ptr);
  }

  template <typename T>
  inline T& RCIPtr<T>::operator*() 
  {
    // perform copy on write with every non-const access to the pointed-to object
    // clone();
    return *(m_cntHolder->m_ptr);
  }

  template <typename T>
  inline void RCIPtr<T>::init()
  {
    if ( !m_cntHolder->isShareable() )
    {
      // non-shareables must not be shared (refcnt==1)
      _ASSERTE(!m_cntHolder->isShared());
      // must not unreference unshared objects because this 
      // would delete the pointed-to object immediately! 
      T * pOldVal = m_cntHolder->m_ptr;
      m_cntHolder = new CountHolder;
      m_cntHolder->m_ptr = new T(*pOldVal);
    }
    m_cntHolder->addRef();
  }

  template <typename T>
  inline void RCIPtr<T>::clone()
  {
    if ( m_cntHolder->isShared() )
    {
      T * pOldVal = m_cntHolder->m_ptr; // remember the old object pointer for deep copying below
      m_cntHolder->removeRef(); // unreference the object before beeing new assigned
      m_cntHolder = new CountHolder;
      m_cntHolder->m_ptr = new T(*pOldVal); // deep copy
      m_cntHolder->addRef();
    }
  }
  //>>

  template <typename T>
  inline RCObject_AutoPtr<T>::RCObject_AutoPtr(T * ptr)
  : m_ptr(ptr)
  {
  }

  template <typename T>
  template <typename U>
  inline RCObject_AutoPtr<T>::RCObject_AutoPtr(const RCObject_AutoPtr<U>& rhs)
  : m_ptr(rhs.release())
  {
    // !! don't add reference here !!
  }

  template <typename T>
  inline RCObject_AutoPtr<T>::~RCObject_AutoPtr()
  {
    deletePointer();
  }
    
  template <typename T>
  template <typename U>
  inline RCObject_AutoPtr<T>& RCObject_AutoPtr<T>::operator=(const RCObject_AutoPtr<U>& rhs)
  {
    if ( this != &rhs )
    {
      reset(rhs.release());
    }
    return *this;
  }

  template <typename T>
  inline T * RCObject_AutoPtr<T>::get() const
  {
    return m_ptr;
  }

  template <typename T>
  inline T * RCObject_AutoPtr<T>::operator->() const
  {
    return m_ptr;
  }

  template <typename T>
  inline T&  RCObject_AutoPtr<T>::operator*() const
  {
    return *m_ptr;
  }
    
  template <typename T>
  inline T * RCObject_AutoPtr<T>::release()
  {
    T * ptr = m_ptr;
    m_ptr = 0;
    return ptr;
  }

  template <typename T>
  void RCObject_AutoPtr<T>::reset(T * ptr)
  {
    if ( m_ptr != ptr )
    {
      deletePointer();
      m_ptr = ptr;
      // !! don't ad reference here !!
    }
  }

  template <typename T>
  void RCObject_AutoPtr<T>::deletePointer()
  {
    if ( m_ptr )
    {
      m_ptr->addRef();
      m_ptr->removeRef();
    }
  }

  //>>
  template <typename T>
  inline void writeAccess( RCPtr<T> &rcPtr )
  {
    // perform a deep copy of the embedded object if shared
    rcPtr.clone();
  }

  //>>
  template <typename T>
  inline void writeAccess( RCIPtr<T> &rcPtr )
  {
    // perform a deep copy of the embedded object if shared
    rcPtr.clone();
  }

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
  //! Convenient helper functor that can be used in STL algorithms to add a reference to reference counted objects.
  /** With this helper functor it's easy and efficient to add a reference to each object contained in a STL container.
    * To copy a container of reference counted objects and increase their reference count you could do something like
    * that:
    *   \code
    *   for ( iter=src.begin() ; iter!=src.end() ; ++iter )
    *   {
    *     dst.push_back( *iter );
    *     (*iter)->addRef();
    *   }
    *   \endcode
    * or, equivalently but more efficiently, using this functor:
    *   \code
    *   transform( src.begin(), src.end(), inserter( dst, dst.begin()), addReference() );
    *   \endcode
    */
  struct addReference
  {
    //! Function call operator for one argument, being the pointer to an object.
    /** The member function \a addRef() of an RCObject is called on the argument.
      * \return the argument for usage in the algorithm */
    template <typename RCOBJPTR>
    RCOBJPTR operator()( RCOBJPTR pRCObject   //!<  pointer to reference counted object
                      );

    //! Function call operator for a pair of arguments.
    /** The member function \a addRef() of an RCObject is called on the second element of the pair of arguments.
      * \return the argument for usage in the algorithm */
    template <typename AnyType, typename RCOBJPTR>
      const std::pair<AnyType, RCOBJPTR>& operator()( std::pair<AnyType,RCOBJPTR>& p  //!< pointer to a pair of any argument and a reference counted object
                                            );
  };

  template <typename RCOBJPTR>
  RCOBJPTR addReference::operator()( RCOBJPTR pRCObject )
  {
    pRCObject->addRef();
    return pRCObject;
  }

  template <typename AnyType, typename RCOBJPTR>
  const std::pair<AnyType, RCOBJPTR>& addReference::operator()( std::pair<AnyType,RCOBJPTR>& p )
  {
    p.second->addRef();
    return p;
  }
//>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

  //! Convenient helper functor that can be used in STL algorithms
  struct removeReference
  {
    //! Function call operator for one argument, being the pointer to an object.
    template <typename RCOBJPTR>
    RCOBJPTR operator()(RCOBJPTR pRCObject) 
      { pRCObject->removeRef(); return pRCObject; }

    //! Function call operator for a pair of arguments.
    template <typename AnyType, typename RCOBJPTR>
    const std::pair<AnyType, RCOBJPTR>& operator() (std::pair<AnyType,RCOBJPTR>& p) 
      { p.second->removeRef(); return p; }
  };

} // namespace nvutil

#ifdef _DEFINED_DBGNEW
# undef new
#endif
