#include "nvsg/nvsgapi.h" // NVSG storage-class defines
#include "GLXRenderArea.h"
#include "SimpleScene.h"

#include "nvsg/PointLight.h"
#include "nvsg/PerspectiveCamera.h"
#include "nvui/TrackballCamManip.h"

#include "nvsg/PlugInterface.h"
#include "nvsg/PlugInterfaceID.h"
#include "nvutil/PlugIn.h"
#include "nvutil/Tools.h"

#include <X11/keysym.h>

#include "nvutil/DbgNew.h" // this must be the last include

using namespace nvsg;
using namespace nvui;
using nvmath::Vec3f;
using nvmath::Quatf;
using std::string;
using std::vector;
using nvutil::GetDirFromPath;

// define a unique plug-interface ID for SceneLoader 
const nvutil::UPITID PITID_SCENE_LOADER(UPITID_SCENE_LOADER, UPITID_VERSION);
const nvutil::UPITID PITID_SCENE_SAVER(UPITID_SCENE_SAVER, UPITID_VERSION);

GLXRenderArea * renderArea=NULL;
const Scene * scene=NULL;
const ViewState * viewState=NULL;
TrackballCamManip * camManip=NULL;

int attrList[] = { 
    GLX_RGBA, 
    GLX_DOUBLEBUFFER, 
    GLX_RED_SIZE, 4, 
    GLX_GREEN_SIZE, 4, 
    GLX_BLUE_SIZE, 4, 
    GLX_DEPTH_SIZE, 24,
    None };

void setupCamera()
{
    // Create camera
    PerspectiveCamera * pNCCamera = beginEdit( PerspectiveCamera::create() );

    // Make the (world space) window aspect ratio 
    // correspond to the (pixel space) viewport aspect ratio    
    pNCCamera->setName("ViewCamera");
    pNCCamera->setAspectRatio(renderArea->getAspectRatio());   

    // Setup the cameras' headlight
    PointLight * pPointLight = beginEdit( PointLight::create() );
    pPointLight->setAmbientColor(Vec3f(1.f, 1.f, 1.f));
    pPointLight->setDiffuseColor(Vec3f(1.0f, 1.0f, 1.0f));
    pPointLight->setSpecularColor(Vec3f(1.0f, 1.0f, 1.0f));
    pPointLight->setAttenuation(1.0f, 0.0f, 0.0f);
    pPointLight->setPosition( Vec3f( 5.0f, 5.0f, 0.0f ) );
    pPointLight->setOrientation( Quatf( 0.0f, 0.0f, 0.0f, 1.0f ) );
    pNCCamera->addHeadLight( endEdit( pPointLight ) );    

    // Zoom all
    pNCCamera->zoom( scene->getRootNode()->getBoundingSphere(), float(nvmath::PI_QUARTER));    

    // Create camera manipulator
    camManip = new TrackballCamManip;
    camManip->attach( endEdit(pNCCamera) );

    renderArea->setSceneData( camManip, scene, viewState );    
}

const Scene* loadScene(char* filename)
{
  printf("Loading %s...\n", filename);

  const Scene *pScene;

  // Set appropriate search paths for the 
  // loader dll and the sample file.
  string modulePath;
  string curDir;
  string dir;
  vector<string> searchPaths;
  
  nvutil::GetCurrentDir(curDir);  
  printf("Adding search path %s\n", curDir.c_str());
  searchPaths.push_back(curDir);

  nvutil::GetModulePath(modulePath);
  GetDirFromPath(modulePath, dir);
  printf("Adding search path %s\n", dir.c_str());
  searchPaths.push_back(dir);

  // also add the scene path    
  GetDirFromPath(filename, dir);
  printf("Adding search path %s\n", dir.c_str());
  searchPaths.push_back(dir);

  // Search for the right loader dll. 
  bool foundAppropriate = false;
  vector<nvutil::UPIID> piids;
  if ( nvutil::queryInterfaceType(searchPaths, PITID_SCENE_LOADER, piids) )
  {
    // extract file extension from filename
    string ext;
    nvutil::GetFileExtFromPath(filename, ext);

    nvutil::UPIID plugID;
    vector<nvutil::UPIID>::iterator it = piids.begin();
    for ( ; it != piids.end(); ++it )
    {
      if ( !_stricmp((*it).getPlugSpecificIDString(), ext.c_str()) )
      {
        // found, copy the ID
        plugID = *it;
        foundAppropriate = true;
        printf("Found %s loader\n", ext.c_str());
        break; // look no further
      }
    }

    // If there is a loader available, 
    // get the interface, load the file and 
    // release the interface when done.
    if ( foundAppropriate )
    {
      nvutil::PlugIn * plug; 
      if ( nvutil::getInterface(searchPaths, plugID, plug) )
      {
        nvsg::SceneLoader * loader = reinterpret_cast<SceneLoader*>(plug);
                
        string fname=filename;        
        pScene = loader->load( (const string)fname, searchPaths, viewState );        
        if( pScene ) 
        { 
          pScene->addRef();
        }        
        if(viewState)
        {
          viewState->addRef();
        }
        
        nvutil::releaseInterface(plugID);
      }
    }
  }

  if (!foundAppropriate)
  {
    printf("Error: No scene loader found!\n");
  }
  else
  {
    if(pScene)
    {
      return pScene;
    }
  }

  return NULL;
}

int main(int argc, char** argv)
{
  Display* dpy;
  int screen;
  Window wnd;
  XSetWindowAttributes attr;
  XEvent event;
  XVisualInfo* vi;
  int width, height;
  Atom wm_delete_window;

  // one-time initialization of global data required by the NVSG API
  nvsg::nvsgInitialize();
  

  if( argc > 1 )
  {
    // load a scene as specified on command line
    scene = loadScene(argv[1]);
    if( scene == NULL )
    {
      printf("Aborting...\n");
      return 1;
    }
  }
  else
  {  
    // create a simple default scene
    scene = SimpleScene::create();
    scene->addRef();
  }

  // create the render area
  renderArea = new GLXRenderArea;  

  // setup camera and manipulator
  setupCamera();

  dpy = XOpenDisplay(0);
  screen = DefaultScreen(dpy);
  vi = glXChooseVisual(dpy, screen, attrList);

  attr.border_pixel = 0;  
  attr.event_mask = ExposureMask | KeyPressMask | ButtonPressMask | ButtonReleaseMask | ButtonMotionMask | StructureNotifyMask;

  wnd = XCreateWindow(dpy, RootWindow(dpy, vi->screen),
            0, 0, 400, 400, 0, vi->depth, InputOutput, vi->visual,
            CWBorderPixel | CWEventMask, &attr);
        
  renderArea->attachWindow(dpy, wnd);
  renderArea->init(vi);
  renderArea->setViewportSize(400, 400);

  XMapRaised(dpy, wnd);  

  // register to receive a message, when the window is about to be destroyed
  wm_delete_window = XInternAtom(dpy, "WM_DELETE_WINDOW", False);
  XSetWMProtocols( dpy, wnd, &wm_delete_window, 1);

  // main event loop
  bool done = false;  
  while (!done)
  {    
    while (XPending(dpy) > 0)
    {
      XNextEvent(dpy, &event);             
      switch (event.type)
      {
        case Expose:
            if (event.xexpose.count != 0)
            {
              break;
            }                 
            renderArea->renderScene();
            break;

        case ConfigureNotify:
            // detect window-size change          
            if ((event.xconfigure.width != width) || 
                (event.xconfigure.height != height))
            {
              width=event.xconfigure.width;
              height=event.xconfigure.height;
              renderArea->setViewportSize(width,height);
              renderArea->triggerRedraw();
            }
            break;        

        case ButtonPress:                   
            switch(event.xbutton.button)
            {
            case Button1:
              renderArea->mouseButtonDown(CameraManipulator::_LEFT_MOUSE_BTN, event.xbutton.x, event.xbutton.y);
              break;
            case Button2:
              renderArea->mouseButtonDown(CameraManipulator::_MIDDLE_MOUSE_BTN, event.xbutton.x, event.xbutton.y);
              break;
            case Button3:
              break;
            default:
              break;
            }
            break;

        case ButtonRelease:                   
            switch(event.xbutton.button)
            {
            case Button1:
              renderArea->mouseButtonUp(CameraManipulator::_LEFT_MOUSE_BTN, event.xbutton.x, event.xbutton.y);
              break;
            case Button2:
              renderArea->mouseButtonUp(CameraManipulator::_MIDDLE_MOUSE_BTN, event.xbutton.x, event.xbutton.y);
              break;
            case Button3:
              break;
            default:
              break;
            }
            break;

        case MotionNotify:
            renderArea->mouseMotion(event.xmotion.x, event.xmotion.y);            
            if( camManip->getButtonState() != CameraManipulator::_NO_BTN_KEY
              || camManip->getKeyState() != CameraManipulator::_NO_BTN_KEY)
            {
              renderArea->triggerRedraw();
            }
            break;

        case KeyPress:
            if (XLookupKeysym(&event.xkey, 0) == XK_Escape)
            {
              done = true;
            }
            break;       

        case ClientMessage:
            if( event.xclient.data.l[0] == wm_delete_window )
            {
              done = true;
            }
            
            break;

        default:
            break;
      }
    }
  }

  renderArea->destroy();
  delete renderArea;

  XCloseDisplay(dpy);

  scene->removeRef();

  delete camManip;

  if(viewState)
  {
    viewState->removeRef();
    viewState = NULL;
  }
  
  // free global data required by the nvsg API
  nvsg::nvsgTerminate();

  return 0;
}

