/***************************************************************************
 *   Copyright (C) 2004 by dsievers@NVIDIA
 *   
 ***************************************************************************/
#include <wx/wx.h>
#include <wx/msgdlg.h>
#include <wx/cursor.h>

#include "nvsg/nvsgapi.h"
#include "nvutil/Assert.h" 

#include "vwrview.h"

#include "nvsg/ErrorHandling.h"
#include "nvsg/PointLight.h"
#include "nvsg/PerspectiveCamera.h"
#include "nvsg/Scene.h"

#include "nvutil/IOLayer.h"

#include "nvmath/Vec3f.h"
#include "nvmath/Quatf.h"

#include "vwrdocument.h"

#include "vwr.h"

// enable memory leak detection
#include "nvutil/DbgNew.h"

using namespace nvsg;
using namespace nvui;
using nvmath::Vec3f;
using nvmath::Quatf;

IMPLEMENT_DYNAMIC_CLASS(VwrView, wxView)

VwrView::VwrView()
 : wxView() 
 , m_pRA(NULL)
 , m_pCamera(NULL)
 , m_pCamManip(NULL)
{
}

VwrView::~VwrView()
{
 
}

/*
bool VwrView::OnCreate(wxDocument *doc, long flags)
{    

}
*/

void VwrView::OnDraw(wxDC *dc)
{
  if(!GetDocument())
  {
    return;
  }

}

void VwrView::OnUpdate(wxView *sender, wxObject *hint)
{  
}

void VwrView::cleanup()
{
  cleanupRenderArea();
  if(m_pCamManip)
  {
    delete(m_pCamManip);
    m_pCamManip=NULL;
  }
}

// called each time a new doc/view is created, but skipped
// on quitting the application
bool VwrView::OnClose(bool deleteWindow)
{    
  Activate(false);

  ((VwrFrame*)GetFrame())->enableMenuNavigation(false);
  ((VwrFrame*)GetFrame())->setCurrentView(NULL);

  if(deleteWindow)
  { 
    cleanup();
  }
  
  return true;
}

bool VwrView::init()
{
  SetFrame(theApp->GetFrame());  
  __ASSERT(m_viewFrame);

  Activate(true);
    
  setupRenderArea();    
  
  if(m_pRA)
  {      
    // Init camera and camera manipulator. The camera must 
    // be connected to a manipulator, so we can use/manipulate it.
    // In this case we use a trackball manipulator to control the 
    // camera movement.
    if ( ! m_pCamManip )
    {     
      __VERIFY( m_pCamManip = new TrackballCamManip );     
    }
  
    const Scene * pScene = ((VwrDocument*)m_viewDocument)->getScene();        
    const ViewState * pViewState = ((VwrDocument*)m_viewDocument)->getInitialViewState();        
    
    __ASSERT( pScene );
    if ( pScene )
    {   
      Camera * pCam = NULL; // use this for edit operations on the camera

      // ViewState's camera, if available,  has highest prio. 
      // If we can't get a camera from the ViewState, ask the Scene for a camera.
      // If this fails too, create a new camera from scratch
      if (  (pViewState && (m_pCamera=pViewState->getCamera()))
         || (m_pCamera=pScene->getCamera()) 
         )
      {
        __ASSERT(m_pCamera);
        m_pCamera->addRef();
      } 
      else 
      {
        // Create camera.
        m_pCamera = PerspectiveCamera::create();     
        m_pCamera->addRef();      

        pCam = beginEdit(m_pCamera);
        pCam->setName( "ViewCamera" );
        // Make scene fit into the viewport.
        pCam->zoom( pScene->getRootNode()->getBoundingSphere() );
        endEdit(pCam);
      }
      
      // Make the (world space) window aspect ratio 
      // correspond to the (pixel space) viewport aspect ratio
      __ASSERT( m_pCamera );
      pCam = beginEdit(m_pCamera);
      pCam->setAspectRatio( m_pRA->getAspectRatio() );
      
      // add a headlight to the camera if scene contains no light
      if((!pScene->containsLight()) && (m_pCamera->getNumberOfHeadLights()==0))
      {
        PointLight * pPointLight = beginEdit( PointLight::create() );
        pPointLight->setAmbientColor( Vec3f(0.2f, 0.2f, 0.2f) );
        pPointLight->setDiffuseColor( Vec3f(1.0f, 1.0f, 1.0f) );
        pPointLight->setSpecularColor( Vec3f(1.0f, 1.0f, 1.0f) );
        pPointLight->setAttenuation( 1.0f, 0.0f, 0.0f );
        pCam->addHeadLight( endEdit(pPointLight) );
      }
      endEdit(pCam);

      // define render area's scene data
      m_pCamManip->attach( m_pCamera );
      m_pRA->setSceneData( m_pCamManip, pScene, pViewState );

      // after hand-over the camera the view don't need to reference it any longer
      m_pCamera->removeRef();

      ((VwrFrame*)GetFrame())->enableMenuNavigation(true);
      ((VwrFrame*)GetFrame())->setCurrentView(this);

      return true;
    }
  }
 
  return false;
}

void VwrView::setupRenderArea()
{
  static int iAttributes[10];
  
  iAttributes[0] = WX_GL_RGBA;
  iAttributes[1] = WX_GL_DEPTH_SIZE;
  iAttributes[2] = 24;
  iAttributes[3] = WX_GL_STENCIL_SIZE;
  iAttributes[4] = 0;
  iAttributes[5] = WX_GL_DOUBLEBUFFER;
  iAttributes[6] = 0; 
  iAttributes[7] = 0; 
  iAttributes[8] = 0;
  iAttributes[9] = 0;
  
  // Throw away existing render area stuff.
  cleanupRenderArea();
    
  __ASSERT(GetFrame());  
    
  m_pRA = new GLwxRAWnd(GetFrame(),  iAttributes);
  
  if (m_pRA)
  {        
    if ( m_pRA->init() )
    {
      // Adjust size of render area.        
      int width, height;
      GetFrame()->GetClientSize(&width, &height);
      m_pRA->setViewportSize( width, height );     
    }
    else
    {
      wxMessageBox("Can't setup render area!", "Error", wxOK | wxICON_EXCLAMATION );
    
      delete m_pRA;
      m_pRA = NULL;
    } 
  } 
  else
  {
    wxMessageBox("Can't create render window!", "Error", wxOK | wxICON_EXCLAMATION );
  }
 
}

void VwrView::cleanupRenderArea()
{
  // Rernder area clean up.  
  if ( m_pRA ) 
  {         
    m_pRA->destroy();
    delete m_pRA;
    m_pRA = NULL;
  }
}

void VwrView::viewAll()
{  
  VwrDocument* pDoc = (VwrDocument*)GetDocument();  

  if( pDoc && m_pRA )
  {
    if(pDoc->getScene() && pDoc->getScene()->getRootNode())
    {
      if(m_pCamManip->getAttachedCamera())
      {
        m_pCamManip->getAttachedCamera()->zoom(pDoc->getScene()->getRootNode()->getBoundingSphere());
        m_pRA->triggerRedraw();
      }
    }
  }
}


