// glut.cpp : Defines the entry point for the console application.
//

#if defined(_WIN32)
# define WIN32_LEAN_AND_MEAN
#endif

#include "nvsg/nvsgapi.h" // NVSG storage-class defines
#include "glutRenderArea.h"
#include "SimpleScene.h"

#include "nvsg/PointLight.h"
#include "nvsg/PerspectiveCamera.h"
#include "nvsg/ViewState.h"
#include "nvui/TrackballCamManip.h"

#include <GL/glut.h>

#include "nvutil/DbgNew.h" // this must be the last include

using namespace nvsg;
using nvmath::Vec3f;
using nvmath::Quatf;
using namespace nvui;

glutRenderArea* renderArea;
const SimpleScene* scene;
TrackballCamManip * pCamManip;

void setupCamera()
{
    // Create camera and manipulator
    pCamManip = new TrackballCamManip;
    PerspectiveCamera * pNCCamera = beginEdit( PerspectiveCamera::create() );

    // Make the (world space) window aspect ratio 
    // correspond to the (pixel space) viewport aspect ratio    
    pNCCamera->setName("ViewCamera");
    pNCCamera->setAspectRatio(renderArea->getAspectRatio());   

    // Setup the cameras' headlight
    PointLight * pPointLight = beginEdit( PointLight::create() );
    pPointLight->setAmbientColor(Vec3f(1.f, 1.f, 1.f));
    pPointLight->setDiffuseColor(Vec3f(1.0f, 1.0f, 1.0f));
    pPointLight->setSpecularColor(Vec3f(1.0f, 1.0f, 1.0f));
    pPointLight->setAttenuation(1.0f, 0.0f, 0.0f);
    pPointLight->setPosition( Vec3f( 5.0f, 5.0f, 0.0f ) );
    pPointLight->setOrientation( Quatf( 0.0f, 0.0f, 0.0f, 1.0f ) );
    pNCCamera->addHeadLight( endEdit( pPointLight ) );    

    // Zoom all
    pNCCamera->zoom( scene->getRootNode()->getBoundingSphere(), float(nvmath::PI_QUARTER));    

    pCamManip->attach(endEdit( pNCCamera ));
    
    renderArea->setSceneData( pCamManip, scene, NULL );   
}

void display(void)
{
  renderArea->renderScene();
}

void reshape(int w, int h)
{
  renderArea->setViewportSize(w, h);
}

void mouseMove(int x, int y)
{
  renderArea->mouseMotion(x,y);  
  if( pCamManip->getButtonState() != CameraManipulator::_NO_BTN_KEY
      || pCamManip->getKeyState() != CameraManipulator::_NO_BTN_KEY)
  {
    renderArea->triggerRedraw();
  }
  
}

void mouseButton(int button, int state, int x, int y)
{
  UINT bn;

  switch(button)
  {
  case GLUT_LEFT_BUTTON:
    bn=CameraManipulator::_LEFT_MOUSE_BTN;
    break;
  case GLUT_MIDDLE_BUTTON:
    bn=CameraManipulator::_MIDDLE_MOUSE_BTN;
    break;
  default:
    return;
  }
  
  (state==GLUT_UP) ? renderArea->mouseButtonUp(bn, x, y) : renderArea->mouseButtonDown(bn, x, y);
}

int main(int argc, char** argv)
{
  // one-time initialization of global data required by the NVSG API
  nvsg::nvsgInitialize();
  
  // create the render area
  renderArea = new glutRenderArea;  

  // create the scene tree
  scene = SimpleScene::create();
  scene->addRef();

  // setup camera and manipulator
  setupCamera();

  glutInit(&argc, argv);
  glutInitDisplayMode (GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH);
  glutInitWindowSize(400, 400);
  glutCreateWindow("Sporting NVSG with glut...");  
  glutReshapeFunc(reshape);  
  glutDisplayFunc(display);
  glutMotionFunc(mouseMove);
  glutMouseFunc(mouseButton);
  glutMainLoop();
  
  // note: the cleanup code is unfortunately never called, because glutMainLoop() exits directly.

  renderArea->destroy();
  delete renderArea;

  scene->removeRef();  

  delete pCamManip;
  
  // free global data required by the nvsg API
  nvsg::nvsgTerminate();

  return 0;
}

